
#include <bspline2d.h>
#include <bspline.c>
#include <sparsevect.h>

double EqualStepBSpline2D::Value(const double& x, const double& y, bool zero_outside_range) const {
  if(zero_outside_range) {
    if(x<xmin) return 0;
    if(x>xmax) return 0;   
    if(y<ymin) return 0;
    if(y>ymax) return 0;   
  }
  double v=0;
  
  double rx = xscale*x+xoffset;
  size_t ixmin = (rx>2) ? rx-2 : 0;
  size_t ixmax = (ixmin+3<nx) ? ixmin+3 : nx;
  
  double ry = yscale*y+yoffset;
  size_t iymin = (ry>2) ? ry-2 : 0;
  size_t iymax = (iymin+3<ny) ? iymin+3 : ny;
  for(size_t ix=ixmin;ix<ixmax;ix++) {
    double b = Bspline3(rx,ix);
    for(size_t iy=iymin;iy<iymax;iy++) {
      v += b*Bspline3(ry,iy)*params(ix+nx*iy);
    }
  }
  return v;
}

bool EqualStepBSpline2D::Fit(size_t i_nx, const double*x, size_t i_ny, const double* y,  double** f, bool adapt_npars) {
  bool debug = true;
  if(debug) cout << "DEBUG EqualStepBSpline2D::Fit starting" << endl;

  SetRange(x[0],x[i_nx-1],y[0],y[i_ny-1]);
  if(adapt_npars)
    SetNPar(i_nx,i_ny); // sets nx and ny
  
  double rx,ry;
  //if(debug) cout << "DEBUG EqualStepBSpline2D::Fit compute x sparsevect" << endl;
  vector<SparseVect> xv;
  {
    for(size_t ixf=0;ixf<i_nx;ixf++,x++) { // func nodes 
      rx = xscale*(*x)+xoffset;  
      
      size_t ixmin = (rx>2) ? rx-2 : 0;
      size_t ixmax = (ixmin+3<nx) ? ixmin+3 : nx;
      SparseVect v;
      for(size_t ixs=ixmin;ixs<ixmax;ixs++)
	v[ixs]=Bspline3(rx,ixs);
      xv.push_back(v);
    }
  }

  //if(debug) cout << "DEBUG EqualStepBSpline2D::Fit compute y sparsevect" << endl;
  vector<SparseVect> yv;
  {
    for(size_t iyf=0;iyf<i_ny;iyf++,y++) { // func nodes 
      ry = yscale*(*y)+yoffset;

      size_t iymin = (ry>2) ? ry-2 : 0;
      size_t iymax = (iymin+3<ny) ? iymin+3 : ny; 
      SparseVect v;
      for(size_t iys=iymin;iys<iymax;iys++)
	v[iys]=Bspline3(ry,iys);
      yv.push_back(v);
    }
  }
  
  size_t npar = nx*ny;
  Mat A(npar,npar);
  Vect B(npar);
  Vect H(npar);
  if(debug) cout << "DEBUG EqualStepBSpline2D::Fit start filling matrix" << endl;
  for(size_t ixf=0;ixf<i_nx;ixf++) { // func nodes 
    SparseVect& ixv = xv[ixf];
    
    const double* fx = f[ixf];
    for(size_t iyf=0;iyf<i_ny;iyf++,fx++) { // func nodes
      SparseVect& iyv = yv[iyf];
      
      //if(debug) cout << "DEBUG EqualStepBSpline2D::Fit " << ixf << " " << iyf << " " << *fx << endl;
      H.Zero();
      for(SparseVectConstIterator itx = ixv.begin(); itx != ixv.end(); itx++)
	for(SparseVectConstIterator ity = iyv.begin(); ity != iyv.end(); ity++)
	  H(itx->first + ity->first * nx ) = itx->second * ity->second;
      
      const double *hj = H.Data();
      double *a = A.NonConstData();
      double *b = B.NonConstData();
      for(size_t j=0;j<npar;j++,hj++,b++) {
	if( ! *hj) {a+=npar; continue;}
	*b += (*fx)*(*hj);
	const double *hi = hj;
	a+=j;
	for(size_t i=j;i<npar;i++,hi++,a++)
	  *a += (*hi)*(*hj);//  Mat(H)*H.transposed();
      }
    }
  }
  
  if (debug) cout << "DEBUG EqualStepBSpline2D::Fit solving" << endl;
  
  int status = cholesky_solve(A,B,"L");
  if(status != 0) {
    return false;
  } 
  params=B;
  
  if (debug) cout << "DEBUG EqualStepBSpline2D::Fit done" << endl;
  
  return true;
}

bool EqualStepBSpline2D::Fit(Grid2DFunction &func) { // cannot be const 
  return Fit(func.NstepX(),func.xvals(),func.NstepY(),func.yvals(),func.fvals());
}
////
extern "C" {
  double divdifd_(const double *, const double *, const int *, const double *, const int*);
}
double BSpline2D::RemappedX(const double& x) const {
  int n = original_x_nodes.size();
  if(n<2) {
    cout <<" ERROR in BSpline2D::RemappedX n=" << n << endl;
    abort();
  }
  int deg = 2;
  return divdifd_(remapped_x_nodes.Data(), original_x_nodes.Data(), &n, &x, &deg);
}
double BSpline2D::RemappedY(const double& y) const {
  int n = original_y_nodes.size();
  if(n<2) {
    cout <<" ERROR in BSpline2D::RemappedY n=" << n << endl;
    abort();
  }
  int deg = 2;
  return divdifd_(remapped_y_nodes.Data(), original_y_nodes.Data(), &n, &y, &deg);
}

double BSpline2D::Value(const double& x, const double& y, bool zero_outside_range) const {
  return EqualStepBSpline2D::Value(RemappedX(x),RemappedY(y),zero_outside_range);
}


bool BSpline2D::Fit(size_t nx, const double*x, size_t ny, const double* y,  double** f, bool adapt_npar) {
  
  cout << "DEBUG BSpline2D::Fit" << endl;
  
  original_x_nodes.allocate(nx);
  remapped_x_nodes.allocate(nx);
  double *o = original_x_nodes.NonConstData();
  double *r = remapped_x_nodes.NonConstData();
  for(size_t i=0;i<nx;i++,x++,o++,r++) {
    *o = *x;
    *r = i;
  }
  original_y_nodes.allocate(ny);
  remapped_y_nodes.allocate(ny);
  o = original_y_nodes.NonConstData();
  r = remapped_y_nodes.NonConstData();
  for(size_t i=0;i<ny;i++,y++,o++,r++) {
    *o = *y;
    *r = i;
  }
  bool status = EqualStepBSpline2D::Fit(nx,remapped_x_nodes.Data(),
					ny,remapped_y_nodes.Data(),f);
  if( ! status) cout << "ERROR BSpline2D::Fit, EqualStepBSpline2D::Fit failed" << endl;
  return status;
}
bool BSpline2D::Fit(Grid2DFunction &func) {
  
  
  return Fit(func.NstepX(),func.xvals(),func.NstepY(),func.yvals(),func.fvals());
}
