#include <cmath>
#include "cosmology.h"
#include "fatalerror.h"



using namespace std;

// Reynald appelle ca l'integrale de Riemann:
template <class Function>  double integ(const Function &F, const double Xmin, const double Xmax, const int Nsteps = 100)
{
  double sum = 0;
  double xstep = (Xmax-Xmin)/double(Nsteps);
  double x = Xmin + 0.5*xstep;
  for (int i=0; i< Nsteps; ++i)
    { 
      sum += F(x);
      x+= xstep;
    }
  return sum*xstep;
}



ostream& operator << ( ostream &stream, const Cosmology &This) 
{ This.dump(stream); return stream;}

static double ident(const double x) { return x;}
static double one(const double x) { return 1;}
static double cub(double x) { return x*x*x;}
static double sq(double x) { return x*x;}

Cosmology::Cosmology(const double &Omegak) : omegak(Omegak)
{
  sqrt_abs_omegak = sqrt(fabs(omegak));
  if (omegak > 0.0001)
    {
      Sin = &sinh;
      Cos = &cosh;
    }
  else if (omegak < -0.0001)
    {
      Sin = &sin;
      Cos = &cos;
    }
  else
    {
      Sin = &ident;
      Cos = &one;
      sqrt_abs_omegak = 1.;
    }
 }



double Cosmology::Dm(const double z,int Nsteps) const
{
  return Sin(sqrt_abs_omegak*integ_hinv(0,z,Nsteps))/sqrt_abs_omegak; 
}

double Cosmology::Dl(const double z,int Nsteps) const
{
  return (1+z)*Dm(z,Nsteps);
}

double Cosmology::Da(const double z,int Nsteps) const
{
  return Dm(z,Nsteps)/(1+z);
}

double Cosmology::dVdz(const double z) const
{
  /* still to be checked against CPT 92 curves */
  /* checked "experimentally" however that this routine is the derivative
     of Volume(z) */
  double integral = integ_hinv(0,z);
  double dm = Sin(sqrt_abs_omegak*integral)/sqrt_abs_omegak; 
  double ddmdz = Cos(sqrt_abs_omegak*integral) / E(z);
  return dm*dm*ddmdz/sqrt(1+omegak*sq(dm));
}
  
//! from 0 to z. it is in fact H_0^3 V
double Cosmology::Volume(double z) const
{
  double integral = integ_hinv(0,z);
  double dm = Sin(sqrt_abs_omegak*integral)/sqrt_abs_omegak; 
  double x = omegak*sq(dm);
  /* I checked experimentaly the continuity over omegak = 0 */
  if (fabs(x) > 1e-3)
    {
      // use exact (however unstable) expression (CPT eq 27.)
      return 0.5*(dm*sqrt(1+x) - integral)/omegak;
    }
  else
    {
      // use taylor expansion
      return cub(dm)*(1 - 0.3*x + 9.*sq(x)/56.- 5.*cub(x)/48.)/3.;
    }
}
  

/*
> a:=int(x^2/sqrt(1+ok*x^2),x=0..z);
                           2 1/2   3/2        1/2              2 1/2
                z (1 + ok z )    ok    - ln(ok    z + (1 + ok z )   ) ok
       a := 1/2 --------------------------------------------------------
                                           5/2
                                         ok
 
(Equivalent to Carrol Press Turner eq 27)

> series(a,ok=0);
                3         5            7   2          9   3       4
           1/3 z  - 1/10 z  ok + 3/56 z  ok  - 5/144 z  ok  + O(ok )
 
>                                                                               
*/



void Cosmology::GradLogDl(const double z, double *grad, 
			  int Nsteps) const
{
  Cosmology *copy = Clone();
  double dl0 = Dl(z, Nsteps);
  double eps = 0.001;
  for (int i=0; i< NPar(); ++i)
    {
      if (dl0 == 0) {grad[i] = 0; continue;}
      double pval = GetParam(i);
      copy->SetParam(i, pval+eps);
      double dl1 = copy->Dl(z,Nsteps);
      copy->SetParam(i, pval);
      grad[i] = (dl1-dl0)/(eps*dl0);
    }
  delete copy;
}


GeneralCosmo::GeneralCosmo(const double OmegaM, const double OmegaX, 
			   const double W0, const double W1)
  : Cosmology(1-OmegaM- OmegaX), omegam(OmegaM), omegax (OmegaX), 
    w0(W0), w1(W1) 
{
}


class hinv {
public : 
  const GeneralCosmo &par;
  hinv(const GeneralCosmo &C) : par(C) {};
  double operator()(const double z) const
  {
    
    double h2 = par.omegam*cub(1.+z) + par.omegax*
    exp(3.*(par.w1*z + (1.+ par.w0-par.w1)*log(1.+z)))
      + sq(1+z)*par.omegak;
    return 1./sqrt(h2);
  }
};


double GeneralCosmo::integ_hinv(const double &ZMin, const double &ZMax, const int NStep) const
{
  hinv f(*this);
  return integ(f,ZMin,ZMax,NStep);
}
  

double GeneralCosmo::E(const double z) const
{
  return sqrt(omegam*cub(1.+z) + omegax*
	      exp(3.*(w1*z + (1.+ w0-w1)*log(1.+z)))
	      + sq(1+z)*omegak);
}

void GeneralCosmo::SetParam(const int i, const double Val)
{
  if (i==0) *this = GeneralCosmo(Val, Omegax(), W0(),  W1()      );
  else if (i==1) *this = GeneralCosmo(Omegam(), Val,  W0(), W1()  );
  else if (i==2) this->w0 = Val;
  else if (i==3) this->w1 = Val;
  else FatalError(" wrong param number in SetParam");
}

double GeneralCosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else if (i==1) Omegax();
  else if (i==2) return this->w0;
  else if (i==3) return this->w1;
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}



void GeneralCosmo::dump(ostream &stream) const
{
  stream << " om-ox-w0-w1 cosmo : (" << omegam << ',' << omegax << ',' << w0 << ',' << w1 << ')' << endl;
}


//=============================== OmegamOmegalCosmo =========================
  
OmegamOmegalCosmo::OmegamOmegalCosmo()
  : GeneralCosmo(0.28,0.72, -1, 0)
{}



OmegamOmegalCosmo::OmegamOmegalCosmo(const double omegam, const double omegal )
  : GeneralCosmo(omegam, omegal, -1, 0)
{}

void OmegamOmegalCosmo::dump(ostream &stream) const
{
  stream << " Om-Ol cosmo : (" << Omegam() << ',' << Omegax() << ')' << endl;
}


void OmegamOmegalCosmo::SetParam(const int i, const double Val)
{
  if (i==0) *this = OmegamOmegalCosmo(Val, Omegax()      );
  else if (i==1) *this = OmegamOmegalCosmo(Omegam(), Val);  
  else FatalError(" wrong param number in SetParam");
}

double OmegamOmegalCosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else if (i==1) return Omegax();
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}

//=============================== FlatLCDMCosmo =========================
  
FlatLCDMCosmo::FlatLCDMCosmo()
  : GeneralCosmo(0.28,0.72, -1, 0)
{}



FlatLCDMCosmo::FlatLCDMCosmo(const double omegam)
  : GeneralCosmo(omegam, 1-omegam, -1, 0)
{}

void FlatLCDMCosmo::dump(ostream &stream) const
{
  stream << " FLCDM cosmo : (" << Omegam() <<  ')' << endl;
}


void FlatLCDMCosmo::SetParam(const int i, const double Val)
{
  if (i==0) *this = FlatLCDMCosmo(Val);
  else FatalError(" wrong param number in SetParam");
}

double FlatLCDMCosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}



//================================ OmegamW0 ===========================

OmegamW0Cosmo::OmegamW0Cosmo(const double Omegam, const double W0) 
  :GeneralCosmo(Omegam, 1.-Omegam, W0, 0.0)
{}


void OmegamW0Cosmo::dump(ostream &stream) const
{
  stream << " Om-w0 cosmo : (" << Omegam() << ',' << W0() << ')' << endl;
}

void OmegamW0Cosmo::SetParam(const int i, const double Val)
{
  if (i==0) *this = OmegamW0Cosmo(Val,       W0());
  else if (i==1) *this = OmegamW0Cosmo(Omegam() , Val );
  else FatalError(" wrong param number in SetParam");
}

double OmegamW0Cosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else if (i==1) return W0();
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}



//================================ OmegamOmegaXW0 ===========================

OmegamOmegaXW0Cosmo::OmegamOmegaXW0Cosmo(const double Omegam, const double OmegaX, const double W0) 
  :GeneralCosmo(Omegam, OmegaX, W0, 0.0)
{}


void OmegamOmegaXW0Cosmo::dump(ostream &stream) const
{
  stream << " Om-Ox-w0 cosmo : (" << Omegam() << ',' << Omegax() << ',' << W0() << ')' << endl;
}

void OmegamOmegaXW0Cosmo::SetParam(const int i, const double Val)
{
  if (i==0) *this = OmegamOmegaXW0Cosmo(Val, Omegax(),       W0()      );
  else if (i==1) *this = OmegamOmegaXW0Cosmo(Omegam(), Val, W0()  );
  else if (i==2) *this = OmegamOmegaXW0Cosmo(Omegam(),       Omegax(),  Val);
  else FatalError(" wrong param number in SetParam");
}

double OmegamOmegaXW0Cosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else if (i==1) return Omegax();
  else if (i==2) return W0();
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}


//================================ OmegamW0W1Cosmo ===========================
OmegamW0W1Cosmo::OmegamW0W1Cosmo(const double Omegam, const double W0, const double W1)
  : GeneralCosmo(Omegam, 1-Omegam, W0, W1)
{
}

void OmegamW0W1Cosmo::dump(ostream &stream) const
{
  stream << " Om-w0-w1 cosmo : (" << Omegam() << ',' << ',' << W0() << ',' << W1() << ')' << endl;
}

void OmegamW0W1Cosmo::SetParam(const int i, const double Val)
{
  if (i==0) *this = OmegamW0W1Cosmo(Val, W0(), W1()      );
  else if (i==1) *this = OmegamW0W1Cosmo(Omegam(),  Val, W1()  );
  else if (i==2) *this = OmegamW0W1Cosmo(Omegam(),  W0(), Val);
  else FatalError("  wrong param number in SetParam");
}

double OmegamW0W1Cosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else if (i==1) return W0();
  else if (i==2) return W1();
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}


//================================ SNAPCosmo ===========================




SNAPCosmo::SNAPCosmo(const double OmegaM, const double OmegaX, const double W0, const double Wprime)
  : Cosmology(1-OmegaM- OmegaX), omegam(OmegaM), omegax (OmegaX), w0(W0), wprime(Wprime) 
{
}

class SNAPhinv {
public :
  const SNAPCosmo &par;
  SNAPhinv(const SNAPCosmo &C) : par(C) { } ;
  double operator()(const double z) const 
  {
  /* I don't understand what the following formula is supposed to describe... */
#ifdef STORAGE
    double h2 = par.omegam*cub(1.+z) 
      + par.omegax * exp(3.*(1.+par.w0+2.*par.wprime)*log(1.+z)-6.*par.wprime*(z/(1.+z)))
      + sq(1+z)*par.omegak;
    return 1./sqrt(h2);
#endif
    /* the following formula assumes that w(z) = w0+wprime*z/(1+z) */
    double h2 = par.omegam*cub(1.+z) 
      + par.omegax * exp(3.*(log(1+z)*(1+par.w0+par.wprime)-par.wprime*z/(1+z)))
      + sq(1+z)*par.omegak;
    return 1./sqrt(h2);
  }
};


double SNAPCosmo::integ_hinv(const double &ZMin, const double &ZMax, 
			     const int NStep) const
{
  SNAPhinv f(*this);
  return integ(f,ZMin,ZMax,NStep);
}
  


double SNAPCosmo::E(const double z) const
{
  SNAPhinv f(*this);
  return 1./f(z);
#ifdef STORAGE
  return sqrt(omegam*cub(1.+z) 
      + omegax * exp(3.*(1.+w0+2.*wprime)*log(1.+z)-
			 6.*wprime*(z/(1.+z)))
      + sq(1+z)*omegak);
#endif
}

void SNAPCosmo::SetParam(const int i, const double Val)
{
  if (i==0) *this = SNAPCosmo(Val, Omegax(), W0(),       wprime      );
  else if (i==1) *this = SNAPCosmo(Omegam(), Val,       W0(), wprime  );
  else if (i==2) *this = SNAPCosmo(Omegam(), Omegax(),Val, wprime  );
  else if (i==3) *this = SNAPCosmo(Omegam(), Omegax(),   W0(), Val);
  else FatalError(" ben merdealors");
}

double SNAPCosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else if (i==1) return Omegax();
  else if (i==2) return W0();
  else if (i==3) return wprime;
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}



void SNAPCosmo::dump(ostream &stream) const
{
  stream << " om-ox-w0-wp cosmo : (" << omegam << ',' << omegax << ',' << w0 << ',' << wprime << ')' << endl;
}


// ========= OmegamW0WprimeCosmo ====================

OmegamW0WprimeCosmo::OmegamW0WprimeCosmo(const double Omegam, const double W0, const double W1)
  : SNAPCosmo(Omegam, 1-Omegam, W0, W1)
{
}

void OmegamW0WprimeCosmo::dump(ostream &stream) const
{
  stream << " Om-w0-wp cosmo : (" << Omegam() << ',' << ',' << W0() << ',' << wprime << ')' << endl;
}

void OmegamW0WprimeCosmo::SetParam(const int i, const double Val)
{
  if (i==0) *this = OmegamW0WprimeCosmo(Val, W0(),       wprime      );
  else if (i==1) *this = OmegamW0WprimeCosmo(Omegam(),  Val, wprime  );
  else if (i==2) *this = OmegamW0WprimeCosmo(Omegam(),       W0(), Val);
  else FatalError(" wrong param number in SetParam");
}

double OmegamW0WprimeCosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else if (i==1) return W0();
  else if (i==2) return wprime;
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}


////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

//================================ HuCosmo =============================

HuCosmo::HuCosmo(const double OmegaM, const double OmegaX, const double Wn, const double Wa, const double Zn)
  : Cosmology(1-OmegaM-OmegaX) ,
    omegam(OmegaM), omegax (OmegaX), wn(Wn), wa(Wa), zn(Zn) {}

// Mise en place de la paramétrisation de Hu :
class Huhinv {
public :
  const HuCosmo &par;
  Huhinv(const HuCosmo &C) : par(C) { } ;
  double operator()(const double z) const
  { 
    double h2 = par.omegam*cub(1.+z)
      + par.omegax * exp ( -3.*par.wa*(z/(1.+z)) + 3.*(1.+par.wn + (1./(1.+par.zn))*par.wa)*log(1.+z) )
      + sq(1+z)*par.omegak;
    return 1./sqrt(h2);
  }
};

double HuCosmo::integ_hinv(const double &ZMin, const double &ZMax, 
			     const int NStep) const
{
  Huhinv f(*this);
  return integ(f,ZMin,ZMax,NStep);
}

double HuCosmo::E(const double z) const
{
  return sqrt(omegam*cub(1.+z) 
      + omegax * exp( -3.*wa*(z/(1.+z)) + 3.*(1. + wn + (1./(1.+zn))*wa )*log(1.+z) )
      + sq(1+z)*omegak);
}

void HuCosmo::SetParam(const int i, const double Val)
{
       if (i==0) *this = HuCosmo(Val,      Omegax(), Wn(), wa  , Zn() );
  else if (i==1) *this = HuCosmo(Omegam(), Val,      Wn(), wa  , Zn() );
  else if (i==2) *this = HuCosmo(Omegam(), Omegax(), Val,  wa  , Zn() );
  else if (i==3) *this = HuCosmo(Omegam(), Omegax(), Wn(), Val , Zn() );
  else FatalError("  wrong param number in SetParam");
}

double HuCosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else if (i==1) return Omegax();
  else if (i==2) return wa;
  else if (i==3) return Zn();
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}



void HuCosmo::dump(ostream &stream) const
{
  stream << " om-ox-wn-wa cosmo : (" << omegam << ',' << omegax << ',' << wn<< ',' << wa << ") developpement autour de " << zn << endl;
}


// ============= OmegamWnWaCosmo ====================

OmegamWnWaCosmo::OmegamWnWaCosmo(const double Omegam, const double Wn, const double Wa, const double Zn) 
  : HuCosmo(Omegam, 1-Omegam, Wn, Wa, Zn) {}
                                                                         
void OmegamWnWaCosmo::SetParam(const int i, const double Val)
{
       if (i==0) *this = OmegamWnWaCosmo(Val        ,Wn()    ,Wa()    ,Zn());
  else if (i==1) *this = OmegamWnWaCosmo(Omegam()   ,Val     ,Wa()    ,Zn());
  else if (i==2) *this = OmegamWnWaCosmo(Omegam()   ,Wn()    ,Val     ,Zn());
       //  else if (i==3) *this = OmegamWnWaCosmo(Omegam()   ,Wn()    ,Wa()    ,Zn());
       else FatalError (" Check routine OmegamWnWaCosmo::SetParam...") ;
}

double OmegamWnWaCosmo::GetParam(const int i) const
{
  if (i==0) return Omegam();
  else if (i==1) return Wn();
  else if (i==2) return Wa();
  else if (i==3) return Zn();
  else FatalError(" wrong param number in GetParam");
  return 42; // rever reached
}




void OmegamWnWaCosmo::dump(ostream &stream) const
{
  stream << " om-wn-wa cosmo : (" << Omegam() << ',' << Wn() << ',' << Wa() << ") developpement autour de " << Zn() << endl;
}


