
#include "filterwheel.h"
#include <iostream>
#include "instrument.h" /* for IndexParams and co... */
#include <cmath>

#include <cstring> // for strlen, strspn, ...

//#define DEBUG

/***************** FilterCollection ************/

std::ostream& operator << (std::ostream &stream, const FilterCollection &Collection)
{
  Collection.dump(stream); return stream;
}



/************************ FilterWheel *************************/

#include <stdio.h>
#include "fileutils.h"
#include "pathutils.h"
#include "snfitexception.h"

FilterWheel::FilterWheel(const string &FileName, const string &InstrumentName)
{
  //cout << FileName << endl;
  dirName = DirName(FileName);
  FILE *f = fopen(FileName.c_str(),"r");

  cout << "INFO FilterWheel::FilterWheel reading " << FileName << endl;


  if (!f) 
    {
      cerr << "ERROR FilterWheel::FilterWheel : cannot find " << FileName << endl;
      throw(SnfitException(" FilterWheel : cannot find "+FileName));
      //return;
    }
  char line[1024];
  //cout << " reading filters from : " << FileName << endl;
  while (fgets(line,1024,f))
    {
      char *p = line + strspn(line," \t");
      if (strlen(p) == 0) continue;
      if (p[0] == '#') continue;
      char name[128],tag[128],filename[128];
      int nread = sscanf(line,"%s %s %s",name, tag, filename);
      if (nread == 2) sscanf(line,"%s %s",name, filename);
      if (nread ==2 || nread == 3)
	{
	  string completeName  = AssembleFileName(dirName,filename);
#ifdef DEBUG
	  cout << " FilterWheel : reading filename =" << completeName 
	       << " filterName=" << name 
	       << " InstName=" << InstrumentName << endl;
#endif	  

	  Filter f(completeName.c_str(),name,InstrumentName);
	  push_back(f);
	}
      else 
	{
	  cerr << "ERROR FilterWheel::FilterWheel :  cannot decode : " << p << " in " << FileName << endl;
	  throw(SnfitException(" FilterWheel : cannot decode "+string(p)+" in "+FileName));
	}
    }
  fclose(f);
}

vector<string> FilterWheel::BandNames() const
{
  vector<string> names;
  for (list<Filter>::const_iterator it = begin(); it != end() ; ++it)
    names.push_back(it->Band());
  return names;
}


const Filter& FilterWheel::GetFilterByBand(const string &Band) const
{
  for (list<Filter>::const_iterator it = begin(); it != end() ; ++it)
    if (it->Band() == Band) return *it;
  cerr << " Requested filter tagged " << Band << " in filter wheel that contains: " << endl << *this;
  cerr << " expect a crash shortly " << endl;
  return *(Filter *) NULL; 
}

std::ostream& operator << (std::ostream &stream, const FilterWheel& W)
{
  for (FilterCIterator it = W.begin(); it != W.end() ; ++it)
    {
      stream << it->Name() << endl;
    }
  return stream;
}


/**************** NoParamFilterWheel ********************/



Filter NoParamFilterWheel::GetFilterByBand(const string &Band, const IndexParams* P) const
{
  if (P) {} ; // "unused parameter ..." warning killer
  return FilterWheel::GetFilterByBand(Band);
}


/*************** RadiallyVariableFilterWheel **************/


bool RadiallyVariableFilterWheel::has_band(const string &BandName) const
{
  for (BandsHanger::const_iterator i = filterScans.begin(); i!= filterScans.end(); ++i)
    if (i->band==BandName) return true;
  return false;
}

const RadiallyVariableFilterWheel::BandData& RadiallyVariableFilterWheel::find_band(const string &BandName) const
{
  for (BandsHanger::const_iterator i = filterScans.begin(); i!= filterScans.end(); ++i)
    if (i->band==BandName) return *i;
  throw(SnfitException("RadiallyVariableFilterWheel::find_band : no such band "+BandName));
}

RadiallyVariableFilterWheel::BandData& RadiallyVariableFilterWheel::find_band(const string &BandName)
{
  for ( BandsHanger::iterator i = filterScans.begin(); i!= filterScans.end(); ++i)
    if (i->band==BandName) return *i;
  throw(SnfitException("RadiallyVariableFilterWheel::find_band : no such band "+BandName));
}


#include <algorithm> // for sort

RadiallyVariableFilterWheel::RadiallyVariableFilterWheel(const string &FileName, const string &InstrumentName) : allFilters(FileName,InstrumentName)
{
  string dirName = DirName(FileName);
  for (FilterCIterator i=allFilters.begin(); i != allFilters.end(); ++i)
    {
      const Filter &f = *i;
      const string band = f.Band();
      if (!f.Globals().HasKey("MEASUREMENT_RADIUS"))
	throw(SnfitException("RadiallyVariableFilterWheel : no MEASUREMENT_RADIUS key in "+f.Name()));
      double radius = f.Globals().getDoubleValue("MEASUREMENT_RADIUS");
      if (!has_band(band)) filterScans.push_back(BandData(band));
      find_band(band).push_back(FilterScan(radius, f));
    }
  // sort each band by increasing radius;
  for (BandsHanger::iterator i = filterScans.begin(); i != filterScans.end(); ++i)
    {
      BandData &b = *i;
      sort(b.begin(), b.end());
      // ... and check that there are no duplicate entries.... because it is senseless.
      double prevRad = b[0].radius;
      for (size_t k=1; k<b.size(); ++k)
	{
	  if (b[k].radius == prevRad)
	    throw(SnfitException("RadiallyVariableFilterWheel : 2 entries at the same radius for band "+i->band));
	  prevRad = b[k].radius;
	}
    }
#ifdef DEBUG
  cout << "RadiallyVariableFilterWheel : contents of " << FileName << ':' << endl;
  cout << *this << endl;
#endif
}


Filter RadiallyVariableFilterWheel::GetFilterByBand(const string & Band, const IndexParams* P) const
{
  // get the "position" argument
  const FocalPlanePosition *pos = dynamic_cast<const FocalPlanePosition*>(P);
  if (P && !pos)
    {
      throw(SnfitException(" RadiallyVariableFilterWheel::GetFilterByBand was called with a second argument which is not a \"FocalPlanePosition *\""));
    }

  // cook up a default position (for e.g. drawing software). (This may be regarded as dangerous....) 
  FocalPlanePosition defPos(0,0);
  if (!P) 
    {
      cout << "RadiallyVariableFilter::GetFilterByBand called without a position argument. Using (0,0)." << endl;
      pos = &defPos;
    }
    
  // find the requested band
  const BandData &bandData = find_band(Band);

  double rad = sqrt(pos->x*pos->x+pos->y*pos->y);

  // bracket "rad" (no need for a binary search: resampling takes much more time)
  size_t nRadii = bandData.size();
  size_t k;
  for (k=0; k<nRadii; ++k) if (bandData[k].radius >= rad) break;
  if (k == nRadii)
    {
      char toto[256];
      sprintf(toto," x=%f, y=%f, rad=%f, radmax=%f",pos->x,pos->y,rad, bandData[nRadii-1].radius);
      throw(SnfitException(" RadiallyVariableFilterWheel::GetFilterByBand : provided radius is too large "
			   +string(toto)));
    }
  if (k==0) return Filter(*bandData[0].transmission);
  double rprev = bandData[k-1].radius;
  double rnext = bandData[k].radius;
  double f = (rad-rprev)/(rnext-rprev); // checked when loading that rprev!=rnext
  const General1DFunction &fprev =  *(bandData[k-1].transmission);
  const General1DFunction &fnext =  *(bandData[k].transmission);
  int nval = fprev.Nval();
  vector<double> xval(nval);
  vector<double> fval(nval);
  for (int i=0; i<nval ; ++i)
    {
      double lambda = fprev.Xval(i);
      xval[i] = lambda;
      fval[i] = (1-f)*fprev.Fval(i) + f*fnext.Value(lambda);
    }
  const string &instrumentName = bandData[k].transmission->InstrumentName();
  return Filter(General1DFunction(nval, &(fval[0]),&(xval[0])), 
		Band, instrumentName);
}

std::vector<std::string> RadiallyVariableFilterWheel::BandNames() const
{
  std::vector<std::string> bands;
  bands.reserve(filterScans.size());
  for (BandsHanger::const_iterator i = filterScans.begin(); i != filterScans.end(); ++i)
    {
      bands.push_back(i->band);
    }
  return bands;
}


void RadiallyVariableFilterWheel::dump(std::ostream &stream) const
{
  for (BandsHanger::const_iterator i = filterScans.begin(); i != filterScans.end(); ++i)
    {
      stream << " band : " << i->band << " rads = ";
      const BandData &b = *i;
      for (size_t k=0; k<b.size()-1; ++k) stream << b[k].radius << ',';
      stream << b.back().radius << endl;
    }  
}
  
vector<double> RadiallyVariableFilterWheel::MeasurementRadii(const string &Band) const
{
  const BandData &bandData = find_band(Band);
  vector<double> v;
  v.reserve(bandData.size());
  for (BandData::const_iterator j = bandData.begin(); j!= bandData.end(); ++j) 
    v.push_back(j->radius);
  return v;
}
  
const Filter& RadiallyVariableFilterWheel::MeasuredFilter(const std::string &Band, const double Radius) const
{
  const BandData &bandData = find_band(Band);
  for (BandData::const_iterator j = bandData.begin(); j!= bandData.end(); ++j)
    if (j->radius == Radius) return *(j->transmission);
  throw(SnfitException( " RadiallyVariableFilterWheel::MeasuredFilter : no such radius"));
}


