#ifndef FILTERWHEEL__H
#define FILTERWHEEL__H


#include "filter.h"
#include <list>

//! a hanger for a set of filters
class FilterWheel : public list<Filter>
{
 private :
  std::string dirName;

 public :
  //! the File should contain one line per filter. each line is a name, a short tag (e.g. UBVRI) and a filename that contains the actual filter data.
  FilterWheel(const string &FileName, const string &InstrumentName);
  FilterWheel() {};

  const Filter& GetFilterByBand(const string & Band) const;
  std::vector<std::string> BandNames() const;


};
typedef std::list<Filter>::const_iterator FilterCIterator;

std::ostream& operator << (std::ostream &stream, const FilterWheel&);

class IndexParams;


//! virtual interface class (to be used in Instrument)
class FilterCollection
{
 public :
  virtual Filter GetFilterByBand(const string &Band, const IndexParams *Index = NULL) const = 0;
  virtual std::vector<std::string> BandNames() const = 0;
  virtual void dump(std::ostream & s) const = 0;
};

std::ostream& operator << (std::ostream &stream, const FilterCollection & Collection);


//! implementation of  FilterCollection for a "standard" filterwheel.
class NoParamFilterWheel : public FilterCollection, protected FilterWheel {

 public :
  //! reads the File using FilterWheel constructor
  NoParamFilterWheel(const string &FileName, const string &InstrumentName) 
    : FilterWheel(FileName, InstrumentName) 
    {};
  //! Get Filter by Band
  Filter GetFilterByBand(const string & Band, const IndexParams* P = NULL) const;

  void dump(std::ostream &stream) const { stream << (const FilterWheel &)(*this);}

  // don't undestand why I have to do that....
  std::vector<std::string> BandNames() const {return FilterWheel::BandNames();} 

  virtual ~NoParamFilterWheel() {} ;
  
};

#include <vector>

//! implementation of  FilterCollection for a radially variable filterwheel.
class RadiallyVariableFilterWheel : public FilterCollection
{
 private :
  FilterWheel allFilters;
  
  struct FilterScan
  {
    double radius;
    const Filter *transmission;
    FilterScan(double R,const Filter &T) : radius(R),transmission(&T){};
    bool operator<(const FilterScan &Right) const // for sort
    { return (this->radius<Right.radius);} 
  };

  struct BandData : public std::vector<FilterScan>
  {
    std::string band;
    BandData(const std::string &B) : band(B) {};

  };
  
  typedef std::vector<BandData> BandsHanger;
  BandsHanger  filterScans;

  bool has_band(const std::string &Band) const;
  const BandData& find_band(const std::string& Band) const;
  BandData& find_band(const std::string &Band);

 public:
  
  RadiallyVariableFilterWheel(const std::string &FileName, 
			      const string &InstrumentName);

  Filter GetFilterByBand(const string & Band, const IndexParams* P) const;

  // Mandatory members for FilterCollection derived classes:

  void dump(std::ostream &stream) const;
  
  std::vector<std::string> BandNames() const;

  virtual ~RadiallyVariableFilterWheel() {};

  // Routine(s) specific to this class

  std::vector<double> MeasurementRadii(const std::string &Band) const;
  const Filter& MeasuredFilter(const std::string &Band, const double Radius) const;
  


};







#endif
