#include <iostream>
#include <string>
#include <fstream>
#include <vector>
#include <algorithm>

#define MyAssert(A) \
if (!(A)) \
  { \
    std::cerr << " assertion failed :" << std::endl << #A << std::endl \
              << "aborting" << std::endl; \
     abort();\
  }\

#include "fatalerror.h"
#include "fullfit.h"
#include "matvect.h"
#include "wavelength.h"

static double sqr(const double &x) { return x*x;}


//#define DEBUG

// =========================================================================
// =========================================================================



LcForFit::LcForFit(const size_t npoints, const Filter& filter_input) : filter(&filter_input) {
  lambda_mean = filter->Mean();
  MeasurementCovMat.allocate(npoints,npoints);
  Fluxes.allocate(npoints);
  Dates.allocate(npoints);
}

LcForFit::LcForFit() {
  filter=0;
}

SpecForFit::SpecForFit(const MeasuredSpectrum* input_spectrum, const size_t npoints) :
  spectrum(input_spectrum)
{
  Fluxes.allocate(npoints);
  Lambdas.allocate(npoints);
  Weights.allocate(npoints);
  DataVariances.allocate(npoints);
}



void SpecForFit::ComputeWeightsForFit(Model2Fit& model, bool use_model_errors) {

#ifdef DEBUG
  cout << "SpecForFit::ComputeWeightsForFit for date = " << spectrum->Date() << endl;
#endif


  for(size_t i = 0; i< DataVariances.size(); i++) {
    double var = DataVariances(i);
    if(use_model_errors) {
      var += sqr(model.SpectrumModelError(Lambdas(i),spectrum->Date(),true));
    }
    //cout << " var " << " " << Lambdas(i) << " " << Fluxes(i) << " " << var << endl;
    Weights(i) = 1./var;
  }
  
}

double LcForFit::relative_error_based_on_observer_phase(const double &phase)  {
  
  // completely empirical
  double res=0.01;
  if(phase<0.) res+=0.2*sqr(phase/15.);
  if(phase>0.) res+=0.2*sqr(phase/40.);
  return res;
}

void LcForFit::ComputeWeightMatForFit(Model2Fit& model, bool use_model_errors, 
				      bool use_kcorr_errors, double u_band_systematic_error,
				      bool use_observer_phase_weight, double phase_weight) {
  
  

  WeightMatForFit = MeasurementCovMat;
  size_t npoints = WeightMatForFit.SizeX();
  
  /*
  cout << "LcForFit::ComputeWeightMatForFit " 
       << use_model_errors << " " 
       << use_kcorr_errors << " " 
       << u_band_systematic_error << " " 
       << endl;
  */

  //cout <<"phase_weight #3 = " << phase_weight << endl;
  // on rajoute des trucs
  if(use_model_errors 
     || use_kcorr_errors 
     || (u_band_systematic_error>0 && lambda_mean<3700) 
     || use_observer_phase_weight || (phase_weight>0)      
     ) {
    
    double date_of_max = 0;
    if(use_observer_phase_weight || (phase_weight>0) )
      date_of_max = model.params.LocateParam("DayMax")->val;

    Vect model_flux(npoints);
    
    double fmax = model.Flux(filter,date_of_max);

    
    const double * d = Dates.Data();
    double * mf = model_flux.NonConstData();
    double * w  = WeightMatForFit.NonConstData();
    
    for(size_t c=0;c<npoints;c++,d++,mf++,w+=(npoints+1)) { // *w = WeightMatForFit(c,c)
      *mf = fabs(model.Flux(filter,*d));
      
      if(use_model_errors) {
	*w += sqr(model.ModelRelativeError(filter,*d,true)*(*mf));
      }

      if(use_observer_phase_weight) {
	// unchecked
	*w += sqr(relative_error_based_on_observer_phase(*d-date_of_max)*(*mf));
      }
      if(phase_weight>0) {
	*w += sqr(phase_weight*fmax*sqr(model.GetPhase(*d)/20.));
      }
    }
    

    if(use_kcorr_errors) {
#ifdef  NEW_ERRORS  
      cout << "apply NEW_ERRORS" << endl;
      for(size_t i=0;i<npoints;i++) {
	for(size_t j=0;j<=i;j++) {
	  double cov = model.ModelRelativeCovariance(filter,Dates(i),Dates(j))*model_flux(i)*model_flux(j);
	  WeightMatForFit(i,j) += cov;
	  if(i!=j) WeightMatForFit(j,i) += cov;
	}
      }
#else
      // now add kcorr errors
      double kvar = sqr(model.ModelKError(filter));
      WeightMatForFit += kvar*Mat(model_flux)*model_flux.transposed();
#endif
    }
    
    if(u_band_systematic_error>0 && lambda_mean<3700) {
      cout << "Assume systematic error for U-band of u_band_systematic_error " << u_band_systematic_error << endl;
      WeightMatForFit += sqr(u_band_systematic_error)*Mat(model_flux)*model_flux.transposed();
    }
  }
  
  // on inverse
  WeightMatForFit.CholeskyInvert("S");
}

void MeasuredDataForFit::ComputeWeightMatForFit(Model2Fit& model, bool use_model_errors, 
						bool use_kcorr_errors,  double u_band_systematic_error,
						bool use_observer_phase_weight, double phase_weight) {
  
  cout << "INFO MeasuredDataForFit::ComputeWeightMatForFit : use_model_errors = " << use_model_errors << endl;
  cout << "INFO MeasuredDataForFit::ComputeWeightMatForFit : use_kcorr_errors = " << use_kcorr_errors << endl;
  cout << "INFO MeasuredDataForFit::ComputeWeightMatForFit : u_band_systematic_error = " << u_band_systematic_error << endl;
  cout << "INFO MeasuredDataForFit::ComputeWeightMatForFit : use_observer_phase_weight = " << use_observer_phase_weight << endl;
  cout << "INFO MeasuredDataForFit::ComputeWeightMatForFit : phase_weight = " << phase_weight << endl;
  cout << "INFO MeasuredDataForFit::ComputeWeightMatForFit : " << LCs.size() << " LCs" << endl;
  cout << "INFO MeasuredDataForFit::ComputeWeightMatForFit : " << Spectra.size() << " spectra"  << endl;

  for(size_t i=0;i<LCs.size();i++)
    LCs[i].ComputeWeightMatForFit(model,use_model_errors,use_kcorr_errors,u_band_systematic_error,use_observer_phase_weight,phase_weight);
  
  for(size_t i=0;i<Spectra.size();i++)
    Spectra[i].ComputeWeightsForFit(model,use_model_errors);
  

}
  

void MeasuredDataForFit::Load(const double& date_min, const double& date_max, const double& w_min, const double& w_max, Model2Fit& model, size_t max_number_of_lightcurves_in_fit) {
  
  cout << "INFO MeasuredDataForFit::Load date_range       = " << date_min << " " << date_max  << endl;
  cout << "INFO MeasuredDataForFit::Load wavelength range = " << w_min << " " << w_max  << endl;
  
  LCs.clear();
  
  
  size_t nbands =  data.NBands();
  double min_wavelength = 0;
  double max_wavelength = 1000000000;
  
  // we need to choose the band to keep here
  // => define a range of wavelength
  if(max_number_of_lightcurves_in_fit && (nbands>max_number_of_lightcurves_in_fit)) {
    nbands=max_number_of_lightcurves_in_fit;
    double z = model.params.LocateParam("Redshift")->val;
    vector<double> restwavelengthdiff;
    for(size_t il=0; il<data.NBands(); il++) {
      const Filter& observerfilter = data.PointList(il).GetEffectiveFilterWithMilkyWay();
      if ( ! model.CanComputeFluxFor(observerfilter))
	continue;
      restwavelengthdiff.push_back(fabs(observerfilter.CentralWaveLength()/(1.+z)-B_WAVELENGTH));
    }
    std::sort(restwavelengthdiff.begin(),restwavelengthdiff.end());
    min_wavelength = (B_WAVELENGTH - restwavelengthdiff[nbands-1])*(1.+z) - 10.;
    max_wavelength = (B_WAVELENGTH + restwavelengthdiff[nbands-1])*(1.+z) + 10.;
  }
  
  LCs.reserve(nbands);
  
  for(size_t il=0; il<data.NBands(); il++) {
    const MeasuredPointList& plist = data.PointList(il);
    const Filter& f = plist.GetEffectiveFilterWithMilkyWay();
    double l = f.CentralWaveLength();
    if((l<min_wavelength) || (l>max_wavelength)) {
      cout << "INFO MeasuredDataForFit::Load Discarding LC " << f.Band() << " cause keep only " << nbands << " bands" << endl;
      continue;
    }
    if ( ! model.CanComputeFluxFor(f)) {
      cout << "INFO MeasuredDataForFit::Load Discarding LC " << f.Band() << " cause outside model range " << endl;
      continue;
    }
    // collect the valid measurements
    list<CountedRef<MeasuredPoint> > points_for_fit;
    for(MeasuredPointCIterator ip = plist.begin(); ip!=plist.end(); ++ip) {
      if( (*ip)->Date()>date_min && (*ip)->Date()<date_max ) 
	points_for_fit.push_back(*ip);
	else
	  cout << "INFO MeasuredDataForFit::Load Discarding for fit point " << (*ip)->Index() << " date = " << (*ip)->Date() << endl;

    }

    if (points_for_fit.size()==0) {
      cout << "WARNING MeasuredDataForFit::Load empty lc" << endl;
      continue;
    }


    LcForFit LC(points_for_fit.size(),f);
    bool data_has_covariance_matrix = (plist.CovMat.SizeX()>0);
#ifdef DEBUG
    cout << "data_has_covariance_matrix = " << data_has_covariance_matrix << endl;
#endif
    size_t i_counter = 0;
    for(MeasuredPointCIterator ip = points_for_fit.begin(); ip!=points_for_fit.end(); ++ip, ++i_counter) {
      LC.Fluxes(i_counter) = (*ip)->Flux();
      LC.Dates(i_counter) = (*ip)->Date();
      if(data_has_covariance_matrix) {
	size_t j_counter = i_counter;
	for(MeasuredPointCIterator jp = ip; jp!=points_for_fit.end(); ++jp, ++j_counter) {
	  LC.MeasurementCovMat(i_counter,j_counter) = plist.CovMat((*ip)->Index(),(*jp)->Index());
	  LC.MeasurementCovMat(j_counter,i_counter) = LC.MeasurementCovMat(i_counter,j_counter);
	}
      }else{
	LC.MeasurementCovMat(i_counter,i_counter) = sqr((*ip)->FluxErr());
      }
    }
#ifdef DEBUG
    cout << "==== band " << f.Band() << " ====" << endl;
    cout << "Dates  : " << LC.Dates << endl;
    cout << "Fluxes : " << LC.Fluxes << endl;
    cout << "MeasurementCovMat : " << LC.MeasurementCovMat << endl;
#endif
    LCs.push_back(LC);
  }
  
  // now spectra

  Spectra.clear();
  size_t nspectra =  data.NSpectra();
  Spectra.reserve(nspectra);
  
  double z = model.params.LocateParam("Redshift")->val;

  for(size_t is=0;is<nspectra;is++) {
    const MeasuredSpectrum &spec = data.Spectrum(is);
    if( (spec.Date() < date_min) || (spec.Date() > date_max) ) continue;
    
    // collect the valid measurements
    list<CountedRef<MeasuredSpectrumPoint> > points_for_fit;
    for(MeasuredSpectrum::const_iterator ip = spec.begin(); ip!=spec.end(); ++ip) {
      double rl = (*ip)->Wavelength()/(1.+z);
      if( (rl>=w_min) && (rl<=w_max) ) 
	points_for_fit.push_back(*ip);
    }
    if (points_for_fit.size()==0) {
      cout << "WARNING MeasuredDataForFit::Load empty spectrum" << endl;
      continue;
    }
    
    // now save all of this
    SpecForFit spec4fit(&spec,points_for_fit.size());
    int i=0;

#ifdef DEBUG
    cout << "MeasuredDataForFit::Load loading spectrum at date " << spec.Date() << " with npoints = " << points_for_fit.size() << endl;
#endif

    for(list<CountedRef<MeasuredSpectrumPoint> >::iterator ip = points_for_fit.begin(); ip!=points_for_fit.end(); ++ip,++i) {
      spec4fit.Fluxes(i)=(*ip)->Flux();
      spec4fit.Lambdas(i)=(*ip)->Wavelength();
      spec4fit.DataVariances(i)=sqr((*ip)->FluxErr());
    }
    Spectra.push_back(spec4fit);
  }

}

MeasuredDataForFit::MeasuredDataForFit(const MeasuredData& data_input) : data(data_input) {}




// =========================================================================
// =========================================================================









double FullFitPrior::Chi2(const FitParams& params) const {
  const LcParam* p = params.LocateParam(name);
  if(!p) return 0;
  return sqr((p->val-val)/err);
}

FullFit::FullFit(const MeasuredData &D, Model2Fit &M) 
  : measuredData(D),
    measuredDataForFit(D),
    model(M),
    phasemin(-15),
    phasemax(+40),
    lambda_min(2000),
    lambda_max(9000),
    recalibrate_spectra(false),
    use_model_errors(false),
    use_kcorr_errors(false),
    use_minos(false),
    use_observer_phase_weight(false),
    update_weight_in_fit(false),
    u_band_systematic_error(0),
    max_number_of_lightcurves_in_fit(0),
    phase_weight(0.)

{dump_mat=false;}




// Set parameters of the model into minuit

bool FullFit::InitFit(bool init_params)
{
  cout << "INFO FullFit::InitFit int parameters = " << init_params << endl;
  if(init_params) {
    model.InitParams(&measuredData);
    if(recalibrate_spectra)
      for(size_t s=0;s<measuredData.NSpectra();s++) {
	if(measuredData.Spectrum(s).NumberOfCalibrationParameters()>1) // for 1 param, donot explicit the fit
	  model.InitCalibrationParams(&measuredData.Spectrum(s),measuredData.Spectrum(s).NumberOfCalibrationParameters());
      }
  }
  
  // get range of dates
  const LcParam* daymaxpar = model.params.LocateParam("DayMax");
  if(!daymaxpar) {
    cerr << "???? in lcfit, there is no parameter daymax in the model !!!!" << endl;
    abort();
  }
  const LcParam* zpar = model.params.LocateParam("Redshift");
  if(!zpar) {
    cerr << "???? in lcfit, there is no parameter Redshift in the model !!!!" << endl;
    abort();
  }
  
  double day_of_max = daymaxpar->val;
  double z        = zpar->val;
  double date_min = day_of_max+phasemin*(1.+z);
  double date_max = day_of_max+phasemax*(1.+z);
  
  // load data
  double w_min,w_max;
  model.GetWavelengthRange(w_min,w_max);
  measuredDataForFit.Load(date_min,date_max,w_min,w_max,model,max_number_of_lightcurves_in_fit);

  if (measuredDataForFit.LCs.size()>0 || measuredDataForFit.Spectra.size()>0) {
    measuredDataForFit.ComputeWeightMatForFit(model,use_model_errors,use_kcorr_errors,
					      u_band_systematic_error,use_observer_phase_weight,phase_weight);
    //cout <<"phase_weight #1 = " << phase_weight << endl;
  }else{
    cerr << "Fatal error in FullFit::InitFit : no data for fit" << endl;
    cerr << "with these parameters:" << endl;
    cerr << model.params << endl;
    exit(-12);
  }
  
  return true;
} 
 

// minuit stuff :

typedef double (*_Model)();

static FullFit* TheFullFit = NULL;

static void MinuitFcn(int *npar, double *grad, double *fval,
		 double *parp, int *iflag, _Model Model) 
{
  // warning killer
  if (grad || iflag || Model) {}

  TheFullFit->GetModel()->SetParams(parp, *npar);
  *fval = TheFullFit->Chi2();
#ifdef DEBUG2
  int total_npar =  TheFullFit->GetModel()->params.size();
  cout << " values ";
  for (int i=0; i < total_npar; ++i) cout << parp[i] << ' ' ;
  cout << *fval << endl;
#endif /* DEBUG */
}


#include <minuit.h>


void FullFit::Dump() {
  
  cout << "==============================================" << endl;
  model.params.dump(cout, "Values retrieved from the fit : ");
  cout << "==============================================" << endl;
  
  dump_mat = true;
  
  // dump chi2
  // ==================================================
  double chi2_lcs = 0;
  int ndata_lcs = 0;
  double chi2_spectra = 0;
  int ndata_spectra = 0;
  
  int nparams_tot = 0;
  int nparams_spectra = 0;
  int nparams_lcs;
  int npar = model.params.size();
  for (int i=0; i<npar; i++) {
    if(! model.params[i]->isfixed)
      nparams_tot++;
  }
  
  nparams_spectra = 0;
  if(recalibrate_spectra)
    for(size_t s = 0; s < measuredDataForFit.Spectra.size(); ++s) {
      nparams_spectra += measuredDataForFit.Spectra[s].spectrum->NumberOfCalibrationParameters();
    }
  
  nparams_lcs = nparams_tot - nparams_spectra;
  
  for (size_t l = 0; l < measuredDataForFit.LCs.size(); ++l) {
    const LcForFit& LC = measuredDataForFit.LCs[l];
    double thisChi2 = Chi2_of_lightcurve(LC);
    int ndataband = LC.Fluxes.size();
    std::cout << "INFO FullFit : Single LC " << LC.filter->Band() << " " 
	      << thisChi2 << " " 
	      << ndataband << " " 
	      << thisChi2/ndataband 
	      << std::endl;
    
    chi2_lcs += thisChi2;
    ndata_lcs += ndataband;
  }
  
  for (size_t s = 0; s < measuredDataForFit.Spectra.size(); ++s) {
    const SpecForFit& spec = measuredDataForFit.Spectra[s];
    double thisChi2 =  Chi2_of_sn_spectrum(spec);
    int ndataspec = spec.Fluxes.size();
    std::cout << "INFO FullFit : Single SPEC " << spec.Date() << " " 
	      << thisChi2 << " " 
	      << ndataspec << " " 
	      << thisChi2/ndataspec 
	      << std::endl;
    
    chi2_spectra += thisChi2;
    ndata_spectra += ndataspec;
  }
  
  double chi2_priors = 0;
  for (size_t p = 0; p < priors.size(); ++p) {
    const FullFitPrior& prior = priors[p];
    double chi2 = prior.Chi2(model.params);
    std::cout << "INFO FullFit : Single PRIOR " << prior.name << " (" << prior.val << "+-" << prior.err << ") " << chi2 << endl;
  }
  if((ndata_lcs+ndata_spectra+priors.size() - nparams_tot) > 0)
    std::cout << "INFO FullFit : TOT chi2/dof " << (chi2_lcs+chi2_spectra+chi2_priors)/(ndata_lcs+ndata_spectra+priors.size() - nparams_tot) << " (NDF:" << (ndata_lcs+ndata_spectra+priors.size() - nparams_tot) << ")" << std::endl;
  else
    std::cout << "INFO FullFit : TOT chi2/dof " << -1 << " (NDF:" << (ndata_lcs+ndata_spectra+priors.size() - nparams_tot) << ")" << std::endl;
  if(ndata_lcs) {
    if((ndata_lcs - nparams_lcs)>0)
      std::cout << "INFO FullFit : All LC chi2/dof " << (chi2_lcs)/(ndata_lcs - nparams_lcs) << " (NDF:" << (ndata_lcs - nparams_lcs) << ")" << std::endl;
    else
      std::cout << "INFO FullFit : All LC chi2/dof " << -1 << " (NDF:" << (ndata_lcs - nparams_lcs) << ")" << std::endl;
  }
  if(ndata_spectra) {
    if((ndata_spectra - nparams_spectra)>0)
      std::cout << "INFO FullFit : All SPEC chi2/dof " << (chi2_spectra)/(ndata_spectra - nparams_spectra) << " (NDF:" << (ndata_spectra - nparams_spectra) << ")" << std::endl;
    else
      std::cout << "INFO FullFit : All SPEC chi2/dof " << -1 << " (NDF:" << (ndata_spectra - nparams_spectra) << ")" << std::endl;
  cout << "==============================================" << endl;
  }
}


bool FullFit::Minos() {
  
  // recompute model errors
  // =============================
  measuredDataForFit.ComputeWeightMatForFit(model,use_model_errors,use_kcorr_errors,u_band_systematic_error,phase_weight);
 
  std::cout << "==================== MINOS =======================" << std::endl;
  int minim_ierr;
  char comd[80];
  strcpy(comd,"MINIMIZE");
  MNCOMD(MinuitFcn, comd,minim_ierr,NULL);  
  strcpy(comd,"MINOS");
  MNCOMD(MinuitFcn, comd,minim_ierr,NULL);  
  return true;
}

double FullFit::Scan(const string& parameter_name, const double &valmin, const double &valmax, const int n, bool write_res) {
  ofstream *stream = 0;
  
  double min_chi2 = 1.e32;
  double val_at_min = 0;
  double chi2 = 0;
  if(write_res) {
    stream = new ofstream(("scan_"+parameter_name+".list").c_str());
    *stream << "# p : " << parameter_name << endl;
    *stream << "# chi2 : " << endl;
    *stream << "#end " << endl;
  }
  
  
  bool one_successful_minimization = false;
  double step = (valmax-valmin)/(n);
  for(double v=valmin;v<valmax+step*0.1;v+=step) {
    model.FixParameter(parameter_name,v);
    
    //measuredDataForFit.ComputeWeightMatForFit(model,use_model_errors,use_kcorr_errors,u_band_systematic_error);
 
    
    if( ! one_successful_minimization )
      if( ! Minimize(true) ) continue; // with initialization of parameters

    for(int i=0;i<2;i++) // need to do it twice cause model errors may have changed   
      if( ! Minimize(false)) continue;
    
    one_successful_minimization = true;

    
    chi2 = Chi2();
    if(chi2<min_chi2) {
      val_at_min = v;
      min_chi2 = chi2;
    }
    if(write_res) *stream << v << " " << Chi2() << endl;
  }
  
  if(write_res) {
    stream->close();
    cout << "Wrote scan of parameter " << parameter_name << " in scan_" << parameter_name << ".list" << endl;
  }
  model.params.LocateParam(parameter_name)->isfixed=false;
  return val_at_min;
}

// Just calls flush. Specific to GNU fortran
// Necessary to use it because gfortran buffers output before writing
// to stdout, and the fortran and C prints run out of sync.
extern "C" {
  void fortranflush_(const int &unit); 
};


bool FullFit::Minimize(bool init_params) {
  
  // initialization of the fitter
  // =====================================================
  if ( ! InitFit(init_params) )
    return false;
    
#ifdef DEBUG
  model.params.dump(cout, " initial param values :");
#endif
  // use minuit
  // first init :
  MNINIT(5,6,7);
  int npar = model.params.size();
  for (int i=0; i<npar; i++)
    {
      LcParam& param = *(model.params[i]);
      // To give char* to dammed Minuit
      if(param.Name().size()>10) {
	cerr << "Fatal error with parameter named '" << param.Name() << "' :" << endl;
	cerr << "name of parameter cannot be larger than 10 characters, this is a limitation of Minuit" << endl;
	exit(12);
      }
      char name[1500];
      sprintf(name,"%s",param.Name().c_str());
      int ierr;
      
      /*
      // debug
      cout << "DEBUG FullFit::Minimize param " << i << " " 
	   << param.Name()  << " " 
	   << param.val     << " "  
	   << param.step    << " " 
	   << param.minv    << " " 
	   << param.maxv    << endl;
      */
      
      MNPARM(i+1, name, param.val, param.step, param.minv, param.maxv, ierr);
      // no need to check ierr??
      
      if(param.isfixed) {
	// fixe un parametre
	char cmd[100];
	sprintf(cmd, "FIX %i",i+1);
	MNCOMD(MinuitFcn,cmd,ierr,NULL);
      }
    }  

  TheFullFit = this;

  
  // minimization
  // =====================================================
  int minim_ierr;
  char comd[80];
  strcpy(comd,"MINIMIZE");
  MNCOMD(MinuitFcn, comd,minim_ierr,NULL);  
  // flush fortran unit 6 to stdout, to get fortran and C printouts in the historical order
  fortranflush_(6); 
  
  std::cout << "INFO FullFit::Minimize : status from MINIMIZE " << minim_ierr << std::endl;
  
  

  // push back values in the right place(s)
  // ==================================================

  
  vector<string> varyingParams; // in order to address properly the cov matrix (indexed by varying params)
  for (int i=0; i<npar; i++)
    {
      char name[128];
      double val, error, minv, maxv;
      int isVariable;
      MNPOUT(i+1,name,val,error,minv,maxv, isVariable);

      // cout << "DEBUG FullFit::Minimize " << name << " " << isVariable << endl;
      
      LcParam &p = *model.params.LocateParam(name);
      MyAssert(string(name)== p.Name());// in case something gets mad
      p.val = val;
      p.err = error;
      p.minv = minv;
      p.maxv = maxv;
      if (isVariable > 0 ) // == 0 : fixed, <0 undefined
	varyingParams.push_back(p.name);	  
    }
  
  
  // catch the covariance matrix
  // ==================================================
  Mat cov(npar,npar);
  MNEMAT(cov(0,0),npar);
  double c;
  for (size_t i=0; i<varyingParams.size(); ++i)
    for (size_t j=i; j <varyingParams.size(); ++j) {
      c = cov(i,j);
      if(i==j && c<0) {
	cerr << "WARNING negative diagonal term " << i << " in covariance matrix" << endl;
	c = -c;
      }
      model.params.SetCovariance(varyingParams[i], varyingParams[j], c);
    }
  // ==================================================

  
  
  return true;
}
// end of minuit stuff 



double FullFit::Chi2_of_lightcurve(const LcForFit& LC) const
{
  size_t npoints = LC.Fluxes.size();
  Vect models(npoints);
  for(size_t c=0; c<npoints;c++) {
    models(c)=model.Flux(LC.filter,LC.Dates(c));
  }
  
  double chi2 = 0;
  
  Vect residuals  = LC.Fluxes-models;
   
  // recompute weight mat ??
  if(update_weight_in_fit)
    const_cast<LcForFit& >(LC).ComputeWeightMatForFit(model,use_model_errors,use_kcorr_errors,u_band_systematic_error,use_observer_phase_weight,phase_weight);

  chi2 +=  scalar_product(residuals,LC.WeightMatForFit,residuals);
  
  return chi2;
}

double FullFit::Chi2_of_sn_spectrum(const SpecForFit& spec) const
{
  double chi2 = 0;
  size_t npoints = spec.Fluxes.size();
  double mflux;
  bool gal = model.CanFitGalaxySpectrum();
  const double* l = spec.Lambdas.Data();
  const double* f = spec.Fluxes.Data();
  const double* w = spec.Weights.Data();
  const double& d = spec.Date();
  
  if( (!recalibrate_spectra) || spec.spectrum->NumberOfCalibrationParameters()==0) {
    for(size_t c=0; c<npoints;c++,l++,f++,w++) {
      mflux = model.SpectrumFlux(*l,d); // flux de la SN
      if(gal) mflux += model.GalaxySpectrumFlux(*l); // + flux de la contamination galactique
      chi2 += (*w)*sqr(*f-mflux);
    }
  }else{
    if( spec.spectrum->NumberOfCalibrationParameters() == 1) {
      if(! gal) {
      
	// do not explicit the recalibration
	/*
	  voir cahier 2009 p 46
	 */
	double sigma_calib = spec.spectrum->SigmaCalibration();
	double a = 0;
	double b = 0;
	for(size_t c=0; c<npoints;c++,l++,f++,w++) {
	  mflux = model.SpectrumFlux(*l,d); // flux de la SN
	  chi2 += (*w)*sqr(*f-mflux);
	  a += (*w)*mflux*(*f-mflux);
	  b += (*w)*sqr(mflux);
	}
	chi2 -= sqr(a)/(1./sqr(sigma_calib) + b ); 
      
      } else {
      // independent of lambda
      double scale = model.SpectrumCalibration(*l,*spec.spectrum); // recalibration
      for(size_t c=0; c<npoints;c++,l++,f++,w++) {
	mflux = scale*model.SpectrumFlux(*l,d);
	if(gal) mflux += model.GalaxySpectrumFlux(*l);
	chi2 += (*w)*sqr(*f-mflux);
      }
      }
    }else{
      for(size_t c=0; c<npoints;c++,l++,f++,w++) {
	mflux = model.SpectrumFlux(*l,d)*model.SpectrumCalibration(*l,*spec.spectrum);
	if(gal) mflux += model.GalaxySpectrumFlux(*l);
	chi2 += (*w)*sqr(*f-mflux);
      }
    }
  }
  
  return chi2;
}

double FullFit::Chi2_of_galaxy_spectrum(const MeasuredSpectrum& s) const
{
  if( ! s.IsAGalaxy() ) {
    cerr << "Fatal error FullFit::Chi2_of_galaxy_spectrum, spectrum is not that of a galaxy" << endl;
    exit(-12);
  }
  
  double chi2 = 0;
  for (MeasuredSpectrum::const_iterator i= s.begin(); i!= s.end(); ++i)
    {
      const MeasuredSpectrumPoint *p = *i;
      if(p->FluxErr()<=0) {
	cerr << "in SpectrumFit::Chi2 , FluxErr = " << p->FluxErr() << endl;
	continue;
      }
      chi2+= sqr( (p->Flux() - model.GalaxySpectrumFlux(p->Wavelength(),true))/p->FluxErr());
    }
  return chi2;
}

double FullFit::Chi2() 
{
  
  double chi2 = 0;

  // lightcurves
  for (size_t l = 0; l < measuredDataForFit.LCs.size(); ++l) {
    chi2 += Chi2_of_lightcurve(measuredDataForFit.LCs[l]);
  }
  
  // sn spectra
  for(size_t s = 0 ; s < measuredDataForFit.Spectra.size(); s++) {
    chi2 += Chi2_of_sn_spectrum(measuredDataForFit.Spectra[s]);
  }
  
  // galaxy spectrum
  for(size_t s = 0 ; s < measuredData.NSpectra(); s++) {
    if(measuredData.Spectrum(s).Galaxy() && model.CanFitGalaxySpectrum())
      chi2 += Chi2_of_galaxy_spectrum(*(measuredData.Spectrum(s).Galaxy()));
  }

  // priors
  for(size_t p=0; p<priors.size(); p++) {
    chi2 += priors[p].Chi2(model.params);
  }
  return chi2;
}

	      
