#ifndef FULLFIT__H
#define FULLFIT__H

#include <iostream>
#include <string>
#include <map>
#include <vector>

#include "measureddata.h"
#include "model2fit.h"
#include "param.h"
#include "countedref.h"

class FullFitPrior {
public :
  string name;
  double val;
  double err;
  FullFitPrior(const string& in, const double &iv, const double &ie) : name(in), val(iv), err(ie) {}
  FullFitPrior() : name("noname"), val(0.), err(0.) {}
  double Chi2(const FitParams& params) const;
};


class LcForFit {

private :
  
  double lambda_mean;
 
public :

  const Filter* filter;
  
  Mat MeasurementCovMat;
  Mat WeightMatForFit;
  Vect Fluxes;
  Vect Dates;
  const double& MeanLambda() const {return lambda_mean;}
  LcForFit();
  LcForFit(const size_t npoints, const Filter& filter_input);

  double relative_error_based_on_observer_phase(const double &phase) ;
  
  void ComputeWeightMatForFit( Model2Fit& model, bool use_model_errors=false, 
			       bool use_kcorr_errors=false, double u_band_systematic_error=0.,			       
			       bool use_observer_phase_weight=false, double phase_weight=0);
};

class SpecForFit {

 
  public :
    CountedRef<MeasuredSpectrum> spectrum;
    Vect Fluxes;
  Vect Lambdas;
  Vect DataVariances;
  Vect Weights;
  SpecForFit(const MeasuredSpectrum* input_spectrum, const size_t npoints);
  double Date() const {return spectrum->Date();};
  string Name() const {return spectrum->Name();};
  void ComputeWeightsForFit(Model2Fit& model, bool use_model_errors=false);
};

class MeasuredDataForFit {
  
private :
  const MeasuredData& data;

public :

  vector<LcForFit> LCs;
 vector<SpecForFit> Spectra;
 
 void ComputeWeightMatForFit( Model2Fit& model, bool use_model_errors=false, 
			      bool use_kcorr_errors=false,  double u_band_systematic_error=0., 
			      bool use_observer_phase_weight=false, double phase_weight=0);
  void Load(const double& date_min, const double& date_max,  const double& w_min, const double& w_max,  Model2Fit& model, size_t max_number_of_lightcurves_in_fit=0);

  
  
  MeasuredDataForFit(const MeasuredData& data_input);
};




/*! a class for fitting light curve models to data with Minuit 
 *
 * all pertinent info is in measuredData and model (which is a virtual class)
 * contains all the minuit handling
 */
class FullFit
{
  private :
    const MeasuredData& measuredData;
    MeasuredDataForFit measuredDataForFit;
    bool dump_mat;
  
    Model2Fit &model;
    
    double chi2Val;
    double chi2Valpdf;
    double phasemin;
    double phasemax;
    double lambda_min;
    double lambda_max;
    bool   recalibrate_spectra;
    bool   use_model_errors;
    bool   use_kcorr_errors;
    bool   use_minos;
    bool   use_observer_phase_weight; // for photoz scan
    bool   update_weight_in_fit; // if true, minuit will minimize accounting for the variation of weight due to model errors


    double u_band_systematic_error;
    size_t    max_number_of_lightcurves_in_fit;
    double phase_weight;
 
    map<string,Mat> weight_matrices;
    void compute_weight_matrices();
    
    

    public :
      
      vector<FullFitPrior> priors;

      FullFit(const MeasuredData &Data, Model2Fit& Model);
    
    Model2Fit* GetModel() {return &model;};
    const MeasuredDataForFit& GetMeasuredDataForFit() const {return measuredDataForFit;};
    int NPar () const {return model.params.size();}
    
    void RecalibrateSpectra(bool yesOrno = true) { recalibrate_spectra = yesOrno;};
    
    double Chi2_of_lightcurve(const LcForFit& LC) const;
    double Chi2_of_sn_spectrum(const SpecForFit& s) const;
    double Chi2_of_galaxy_spectrum(const MeasuredSpectrum& s) const;
    
    double Scan(const string& parameter_name, const double &valmin, const double &valmax, const int n=10, bool write_res=true);
    double Chi2();
    bool Minos();
    void Dump();
    bool InitFit(bool init_params=true);
    
    void UseModelErrors(bool YoN = true) {use_model_errors = YoN;};
    bool DoUseModelErrors() const {return use_model_errors;};
    void UseKcorrErrors(bool YoN = true) {use_kcorr_errors = YoN;};
    bool DoUseKCorrErrors() const {return use_kcorr_errors;};
    void UseObserverPhaseWeight(bool YoN = true) {use_observer_phase_weight = YoN;};
    bool DoUseObserverPhaseWeight() const {return use_observer_phase_weight;};
    void SetPhaseWeight(double w) {phase_weight = w;};
    bool DoUpdateWeightInFit() const {return update_weight_in_fit;};
    void SetUpdateWeightInFit(bool YoN = true) {update_weight_in_fit = YoN;};
    double GetPhaseWeight() const {return phase_weight;};
    
    void UseMinos(bool YoN = true) {use_minos = YoN;};
    bool DoUseMinos() const {return use_minos;};
    void SetUbandSystematicError(const double& v) {u_band_systematic_error = v;};
    double UbandSystematicError() const {return u_band_systematic_error;};
    void SetMaxNumberOfLightcurvesInFit(const size_t& n) {max_number_of_lightcurves_in_fit=n;};
    size_t MaxNumberOfLightcurvesInFit() const {return max_number_of_lightcurves_in_fit;};
    
    
    bool Minimize(bool init_params=true);
    
    void dump(std::ostream & stream = std::cout) const{
      stream << model.ModelName() << endl;
      model.params.dump(stream);
    }
    friend ostream& operator << (ostream &stream, const FullFit &lc)
      { lc.dump(stream); return stream;}
    

    virtual ~FullFit() {}
    void SetPhaseRange(double pmin, double pmax) {phasemin=pmin;phasemax=pmax;};
    void SetWavelengthRange(const double& lmin, const double& lmax) {lambda_min = lmin;lambda_max = lmax;};
    
};



#endif /* FULLFIT__H */
