#include <iostream>
#include <math.h>
#include "hostabsorption.h" 

using namespace std;

// Extinction low (Cardelli, 1989 ApJ.345.245C) : 
// delta_mag = A(lambda) = [a(x)Rv+b(x)]E(B-V)

// For optical/NIR: 1.1<= x <= 3.3, x in mum**(-1)
double a_opt(const double x)
{
 double y = x - 1.82;
 // Horner's scheme for polynomials
 double a = 1 + y*(0.17699 + y*(-0.50447 + y*(-0.02427
	  + y*(0.72085 + y*(0.01979 + y*(-0.77530 + 0.32999*y))))));

 return a;
}

double b_opt(const double x)
{
  double y = x - 1.82;
  double b = y*(1.41338 + y*(2.28305 + y*(1.07233 +y*(-5.38434
	   + y*(-0.62251 + y*(5.30260 - 2.09002*y))))));

  return b;
} 

//For UV : 3.3 <= x <= 8, x in mum**(-1)
double a_uv(const double x)
{
 double Fa;
  if((5.9 <= x) && (x <= 8)) 
    Fa = -(x-5.9)*(x-5.9) * (0.04473 + 0.009779*(x-5.9));
  if(x < 5.9)
    Fa=0;

  double a = 1.752 - 0.316*x - 0.104/((x-4.67)*(x-4.67) + 0.341) + Fa;

  return a;
}

double b_uv(const double x)
{
 double Fb;
  if((5.9 <= x) && (x <= 8)) 
    Fb = (x-5.9)*(x-5.9) * (0.213 + 0.1207*(x-5.9));
  if(x < 5.9)
    Fb=0;

  double b = -3.09 + 1.825*x + 1.206/((x-4.62)*(x-4.62) + 0.263) + Fb;

  return b;
}

//! lambda in Ang.
double FluxExtinctionLin(const double &lambda, const void* Data)
{
  double EBV = ((double *) Data)[0];
  return exp((lambda-10000.)*EBV/10000.);
}


//! lambda in Ang.

double A_ExtinctionCardelli(const double &lambda, const void* Data)
{
  double EBV = ((double *) Data)[0];
  double Rv = ((double *) Data)[1];
  double x = 1./(lambda/1E4); //lambda in angstroms, x in mum**(-1)

  double a,b;
  
  if(x < 1.1) { // Infrared (0.3<x<1.1) (9091<l<33333.)
    double y = pow(x,1.61);
    a = 0.574*y;
    b = -0.527*y;
  }else{
    if(x<3.3) { // Optical/NIR (1.1<x<3.3) (3030<l<9091)
      a = a_opt(x);
      b = b_opt(x);
    }else{ // UV (3.3<x<8.) (1250<l<3030)
      a = a_uv(x);
      b = b_uv(x);
    }
  }

  double A_l = (a*Rv + b)*EBV;
  return(A_l);
}

double FluxExtinctionCardelli(const double &lambda, const void* Data)
{
  double A_l = (A_ExtinctionCardelli(lambda,Data))/2.5;
  return pow(10., -A_l);
}

//-------------------------------------------------------------------------
// Extinction Calzetti  et al. (2000) estimate R_V = 4.05 +/- 0.80 from optical
//             -IR observations of 4 starbursts.
// k (lambda) = 2.659 (-1.857 + 1.040/lambda) + R'_V  0.63 micro-m < lambda <  2.20 micro-m [6300 A ; 22200 A]
// k (lambda) = 2.659 (-2.156 + 1.509/lambda -0.198/lambda^2 + 0.011/lambda^3 ) + R'_V 
// pour 0.12 micro-m < lambda <  0.63 micro-m soit [1200 A ; 6300 A]

double A_ExtinctionCalzetti(const double &lambda, const double ebv)
{
  double RV = 4.05 ;
  double x = 1./(lambda/1E4); //lambda in angstroms, x in mum**(-1) 
  //double lambda_min = 1200 ;
  double lambda_min = 912 ;
  double lambda_max = 22200 ;
  
  double A_l = 0. ;
  if ( (lambda < lambda_min)  || (lambda > lambda_max  ))
    {
      return(A_l);
    }
  double k ;
  if (lambda > 6300 )
    k = 2.659*(-1.857 + 1.040*x) + RV ; 
  else
    k = 2.659*(-2.156 + 1.509*x -0.198*x*x + 0.011*x*x*x ) + RV ; 
  A_l = k*ebv ;
  return(A_l) ;
}


double FluxExtinctionCalzetti(const double &lambda,  const double ebv)
{
   double A_l = (A_ExtinctionCalzetti(lambda,ebv))/2.5;

  return pow(10., -A_l);
}
