#include <iostream>
#include <string>
#include <fstream>

#include "fittedsn.h"
#include "hubblefitter.h"
//#include "hubblemagmodel.h"
#include "hubblemagmodel2.h"
#include "brent.h" // for estimation of the intrinsic dispersion
#include "cosmology_prior.h"
#include <wmap5cosmo.h>
#include <wmap5cosmoprior.h>

void usage (const char *pg) {
  std::cerr << pg << " <list1> <list2> ... " << std::endl;
  std::cerr << " options : -c <what> : what cosmology is fitted (LCDM,FLCDM,FWCDM,FWWPCDM) default is FLCDM" << endl;
  std::cerr << "           -f parameter value  : fix parameter (among Omegam,Alpha,Beta, ... etc)" << endl;
  std::cerr << "           -e <intrinsic_disp> : fix parameter" << endl;
  std::cerr << "           -z <min value of redshift>  " << endl;
  std::cerr << "           -Z <max value of redshift>  " << endl;
  std::cerr << "           -B use as a prior baryon wiggles constraints from Eisenstein 2005" << endl;
  std::cerr << "           -W use WMAP5 prior" << endl;
  std::cerr << "           -g different set of parameters for galaxy types (as many as types in files)" << endl; 
  std::cerr << "           -t # : use only SNe from this type of galaxies" << endl; 
  std::cerr << "           -b : correct for Malquist bias" << endl;
  std::cerr << "           -C # : iterative clipping at # sigma" << endl;
  std::cerr << "           -H : remove gal wo host " << endl;
  
  exit(EXIT_FAILURE);
}


int main (int argc, char ** argv) {
  
  if(argc < 2) usage(argv[0]);
  
  vector<string> dictfilenames;
  string cosmo2fit = "FLCDM";
  bool eisenstein = false;
  bool use_wmap5_prior = false;
  double intrinsic_dispersion=0.15;
  bool   intrinsic_dispersion_is_fixed = false;
  double zmin=0;
  double zmax=100;
  map<string,double> fixed_parameter_values;
  bool split_by_galaxy_type = false;
  bool remove_sn_wo_host  = false; 
  int only_this_type_of_galaxy = -1;
  bool malmquist_bias_correction = false;
  float nsig_clipping = -1;

  for (int i=1; i<argc; ++i) {
    char *arg = argv[i];
    if (arg[0] != '-') {
      dictfilenames.push_back(arg);
      continue;
    }else{
      switch(arg[1]) {
      case 'C': nsig_clipping = atof(argv[++i]); break;
      case 'c': cosmo2fit = argv[++i]; break;
      case 'b': malmquist_bias_correction = true; break;
      case 'f': {
	string pname = argv[++i];
	double value = atof(argv[++i]);
	fixed_parameter_values[pname]=value; break;
      }
      
      case 'e': 
	intrinsic_dispersion = atof(argv[++i]); 
	intrinsic_dispersion_is_fixed = true;
	break;	
      case 'z': zmin=atof(argv[++i]); break;
      case 'Z': zmax=atof(argv[++i]); break;  	
      case 'B': eisenstein = true ; break;
      case 'W': use_wmap5_prior = true ; break;
      case 'g': split_by_galaxy_type  = true ; break;
      case 'H':  remove_sn_wo_host = true ; break;
      case 't': only_this_type_of_galaxy=atoi(argv[++i]); break;  	
      default:
	cerr << "unknown option " << arg << endl;
	usage(argv[0]);
      }
    }
  }
  if(dictfilenames.size()==0) {
    cerr << "a list must be given" << endl;
    usage(argv[0]);
  }
  


  

  
  // read data
  // ==========================================
  AllSn data;
  for(size_t i=0;i<dictfilenames.size();++i) 
    data.ReadDictFile(dictfilenames[i]);
  
  // rm data with z cutoff
  {
    SnIterator it = data.begin();
    while(it != data.end()) {
      if ( (*it)->GetZ() >= zmin && (*it)->GetZ() < zmax )
	++it;
      else {
	cout << "erasing " << (*it)->GetName() << " z=" << (*it)->GetZ() << endl;
	it=data.erase(it);
      }
    }
  }

  int ngaltypes=1;


  // rm SNe that are not in specific type of galaxies if required
  if(only_this_type_of_galaxy>=0) {
    SnIterator it = data.begin();
    while(it != data.end()) {
      if ( (*it)->GetGType() == only_this_type_of_galaxy )
	++it;
      else {
	cout << "erasing " << (*it)->GetName() << " z=" << (*it)->GetZ() << endl;
	it=data.erase(it);
      }
    }
  }

  // removing SN with no host type
  if( (split_by_galaxy_type) || (remove_sn_wo_host) )
    {
      SnIterator it = data.begin();
      while(it != data.end()) {
	if ( (*it)->GetGType()>=0 )
	  ++it;
	else {
	  cout << "erasing " << (*it)->GetName() << " z=" << (*it)->GetZ() << endl;
	  it=data.erase(it);
	}
      }
    }

  if (split_by_galaxy_type)
    {
      // let's count the # of SN / type
      for(int i = 0 ; i < ngaltypes ; i++)
	{
	  int ni=0 ;
	  for(SnIterator it = data.begin(); it != data.end() ; it++)
	    {
	      if ((*it)->GetGType() == i)
		ni++;
	    }
	  cerr <<"Number of SN for galaxy_type = " << i << " : " << ni << endl ;
	}
  
      ngaltypes=int(data.NumberOfGalaxyTypes());
      cout << "Use " << ngaltypes << " different galaxy types" << endl; 
 
    }
      



#ifdef WITH_HOST_EXTINCTION 
  if(data.HasExtinctionMeasurement()) {
    cout << "Data has Cardelli extinction measurement" << endl; 
  }
#endif

  // which fitter is used
 
  HubbleMagModel2 *model = new HubbleMagModel2(ngaltypes);
#ifdef WITH_UV_FIT
  model->has_uv_fit = data.HasUVFit();
#endif
#ifdef WITH_HOST_EXTINCTION 
  model->has_cardelli_extinction = data.HasExtinctionMeasurement();
#endif
  model->SetNumberOfSALT2Components(2);  
  
  
  // which cosmology is fitted
  if(! model->SetCosmo2Fit(cosmo2fit)) {
    cerr << "What does " << cosmo2fit << " means ??" << endl;
    usage(argv[0]);
  }
  
  
  // create parameters before fixing them
  model->InitParams();
  model->CorrectMalmquistBias(malmquist_bias_correction);

  // fix some parameters if any
  for(map<string,double>::const_iterator it = fixed_parameter_values.begin(); 
      it != fixed_parameter_values.end(); ++it) {
    model->FixParam(it->first,it->second);
  }
#ifdef DALPHABETADZ
  model->FixParam("dAlphadz",0.);
  model->FixParam("dBetadz",0.);
#endif
  
  int ndata = data.size();
  int nbparams = model->NbParams();
  int ndf = ndata-nbparams;
  cout << "(ndata,nparams,ndf) = " << "(" << ndata << ","  << nbparams << ","  << ndf << ")" << endl; 
  
  
  HubbleFitter hubfit(data,*model);
  if(eisenstein)
    hubfit.AddPrior(new Eisenstein2005_Prior());
  if(use_wmap5_prior)
    hubfit.AddPrior(new Wmap5CosmoPrior());
  
  // =====================================================
  // loop if sigma clipping
  // =====================================================
  
  double rms,rmserror;
   
  int noutliers = 0;
  int noutliers_for_iteration = 0;
  do {
  //  =====================================================
  // loop on minuit minimization
  // because intrinsic dispersion may be reevaluated
  // =====================================================
  intrinsic_dispersion = hubfit.MinimizeTheChi2_WithIntrisicDispersion(intrinsic_dispersion, intrinsic_dispersion_is_fixed);
  
  // compute RMS
  ndata=0;
  double res;
  rms=0;
  
  for (SnIterator it = data.begin() ; it != data.end(); ++it) {
    res = model->Res(*it);
    if ( (!(res>0)) && (!(res<=0)) ) {
      cout << "there is an nan for " << (*it)->GetName() <<endl;
      exit(EXIT_FAILURE);
    }
    rms += res*res;
    ndata +=1;
  }
  rms = sqrt(rms/(ndata-nbparams));
  
  cout << "ndata " << ndata << " " << data.size() << endl;
  
  rmserror = rms/sqrt(2.*( ndata-nbparams));
  
  
  
  if(nsig_clipping > 0 ) {
    noutliers_for_iteration = 0;
    SnIterator it = data.begin();
    while(it != data.end()) {
      res = model->Res(*it);
      if(fabs(res)>nsig_clipping*rms) {
	cout << "removing outlier " << (*it)->GetName() << " z=" << (*it)->GetZ() << " res=" << res << " rms=" << rms << endl;
	it=data.erase(it);
	noutliers_for_iteration++;
	
      }else{
	//cout << "keeping " << (*it)->GetName() << " z=" << (*it)->GetZ() << " res=" << res << " rms=" << rms << endl;
	++it;
      }
    }
  }
  noutliers += noutliers_for_iteration;

}while(noutliers_for_iteration>0);
  printf("RMS = %3.3f +- %3.3f    for %d SNe and %d parameters, with %d outliers\n",rms,rmserror,ndata,nbparams,noutliers);
  
  ofstream stream("result_cosmology.dat");
  model->DumpModel(stream);
  stream.close();
  

  delete model;

  return EXIT_SUCCESS;
}

