#include <iostream>
#include <string>
#include <fstream>

#include <algorithm>

#define MyAssert(A) \
if (!(A)) \
  { \
    std::cerr << " assertion failed :" << std::endl << #A << std::endl \
              << "aborting" << std::endl; \
     abort();\
  }\

#include <minuit.h>

#include "brent.h"
#include "cosmology_prior.h"
#include "hubblefitter.h"
#include "cosmology.h"
#include "matvect.h"




//===============================================================



HubbleFitter::HubbleFitter(const AllSn &AllSn, HubbleModel &Model)
  : allsn(AllSn),
    hubblemodel(Model)
{
}

void HubbleFitter::InitTheFit()
{
  hubblemodel.InitParams();
}

typedef double (*_Model)();

static HubbleFitter* TheHubbleFitter = NULL;

static void Function(int *npar, double *grad, double *fval,
		     double *parp, int *iflag, _Model Model)
{
  // warning killer
  if (grad || iflag || Model) {}

  TheHubbleFitter->GetHubbleModel()->SetParams(parp, *npar);
  *fval = TheHubbleFitter->TheChi2();
}



Mat HubbleFitter::GetCovMat() {
  int npar = hubblemodel.hubbleparams.size();
  Mat cov(npar,npar);
  MNEMAT(cov(0,0),npar);
  return cov;
}

static double sqr(const double &x) { return x*x;}


void HubbleFitter::MinimizeTheChi2(bool initfit)
{
  if(initfit)
    InitTheFit();

  MNINIT(5,6,7);

  int npar = hubblemodel.hubbleparams.size();

  for (int i = 0; i < npar; ++i)
    {
      LcParam& param = *(hubblemodel.hubbleparams[i]);
      // To give char* to dawmed Minuit
      char name[12];
      sprintf(name,"%s",param.Name().c_str());
      int ierr;
      MNPARM(i+1, name, param.val, param.step, param.minv, param.maxv, ierr);

      if(param.isfixed){
	//fix a parameter
	char cmd[100];
	sprintf(cmd, "FIX %i", i+1);
	MNCOMD(Function, cmd, ierr, NULL);
      }
    }

  TheHubbleFitter = this;

  int minim_ierr;
  char comd[80];
  strcpy(comd,"MINIMIZE");
  MNCOMD(Function, comd,minim_ierr,NULL);

  std::cout << " status from MINIMIZE " << minim_ierr << std::endl;
  
  // get the chi2 and params
  strcpy(comd,"SHOW FCN");
  int ierr;
  
  MNCOMD(Function, comd,ierr,NULL);  
  
 // push back values in the right place(s)
  vector<string> varyingParams; // in order to address properly the cov matrix (indexed by varying params)
  for (int i=0; i<npar; i++)
    {
      char name[128];
      double val, error, minv, maxv;
      int isVariable;
      MNPOUT(i+1,name,val,error,minv,maxv, isVariable);
      LcParam &p = *(hubblemodel.hubbleparams.LocateParam(name));
      MyAssert(string(name)== p.Name());// in case something gets mad
      p.val = val;
      p.err = error;
      p.minv = minv;
      p.maxv = maxv;
   
      if (isVariable > 0 ) // == 0 : fixed, <0 undefined
	varyingParams.push_back(p.name);	  
    }
  
  
  //catch the covariance matrix
  Mat cov(npar,npar);
  MNEMAT(cov(0,0),npar);
  for (size_t i=0; i<varyingParams.size(); ++i)
    for (size_t j=i; j <varyingParams.size(); ++j)
      hubblemodel.hubbleparams.SetCovariance(varyingParams[i], varyingParams[j], cov(i,j));
  
  hubblemodel.hubbleparams.dump(cout, "Values of the fit : ");

//   cout << "*********" << endl;
//   int nbdata = allsn.size();
//   int nbparams = hubblemodel.hubbleparams.size();

//   cout << "Chi2 = " << TheChi2() << endl;
//   cout << "Nb of sne = " << nbdata << endl;
//   cout << "Nb params = " << nbparams << endl;
//   cout << "ndf = " << nbdata-nbparams << endl;
//   cout << "chi2/dof = " << TheChi2()/(nbdata-nbparams) << endl;

  if ( hubblemodel.GetCosmo2Fit() == HubbleModel::LCDM ) {
    const LcParam* om_p = hubblemodel.hubbleparams.LocateParam("Omegam");
    const LcParam* ol_p = hubblemodel.hubbleparams.LocateParam("Omegal");
    
    double ot = om_p->val+ol_p->val;
    double err = sqrt( sqr(om_p->err) 
		      + sqr(ol_p->err) 
		       + 2.*hubblemodel.hubbleparams.GetCovariance("Omegam","Omegal"));
    cout << "Omegat " << ot << " " << err << endl;

  }

}

// end of minuit stuff



int HubbleFitter::Ndeg_of_Freedom()
{
 int ndata = allsn.size(); 
 int nbparams = hubblemodel.NbParams();
 return(ndata-nbparams);

}

double HubbleFitter::MinimizeTheChi2_WithIntrisicDispersion(const double& intrinsic_dispersion, const bool intrinsic_dispersion_is_fixed)
{
  int ndf = Ndeg_of_Freedom();
  bool first_time = true;
  
  double disp = intrinsic_dispersion;
  // Set intrinsic dispersion value
  hubblemodel.IntrinsicDispersion = intrinsic_dispersion;
  
  while (true) {
    
    // Do the minimization of the chi2
    
    MinimizeTheChi2(first_time);
    first_time = false;

    if(intrinsic_dispersion_is_fixed)
      break;
    
    // Re-evaluate intrinsic dispersion (this alters model->IntrinsicDispersion)
    int status = 0;
    box b; 
    b.hubblefitter = this ;
    b.ndf = ndf;
    double v;
    disp = hubblemodel.IntrinsicDispersion;
    // recalcule intrisic disp jusqu'a ce que le chi2 soit egal a 1
    hubblemodel.IntrinsicDispersion 
      = brent((AnalyticFunction *)chi2_for_intrinsic_dispersion,0.,disp,1.,0.001,&b,v,status);
    if(status) {
      cerr << "Brent est pas content, status = " << status << endl;
      return 12;
    }
    
    if(fabs(disp - hubblemodel.IntrinsicDispersion)<0.001)
      break;
    
  }
  
  // attention priors
  cout << "Intrinsic dispersion = " << hubblemodel.IntrinsicDispersion << " (for chi2dof = " << TheChi2()/ndf  << " )" << endl;
  printf("FINAL_CHI2 %12.10f\n",TheChi2());
  return(disp);
}

double HubbleFitter::TheChi2() 
{
  double chi2 = 0;
  for (SnCIterator it = allsn.begin() ; it != allsn.end(); ++it)
    {
      chi2 += hubblemodel.Chi2(*it);
    }

  if(Priors.size()>0) {
    Cosmology *cosmo = hubblemodel.GetCosmology();
    for(size_t i=0;i<Priors.size(); i++) {
      chi2 += Priors[i]->Chi2(cosmo);
    }
  }

#ifdef STORAGE

  if(bEisenstein2005Prior) {
    // add to chi2 prior on baryon wiggles;
#ifdef TAYLOR_EXPANSION
    double om = hubblemodel.GetOmegam();
    double ol = 1-om;
    double w0 = -1;
    if(hubblemodel.doFitW())
      w0 = hubblemodel.GetOmegal(); // this is not a bug, w0 is stored in omegal parameter
    else
      ol = hubblemodel.GetOmegal();      
    chi2 += sqr( (om-( 0.273+0.123*(1+w0)-0.137*(1-om-ol)))/0.025 );
#else
    Cosmology *cosmo = hubblemodel.GetCosmology();
    chi2 += EisensteinChi2(*cosmo);
    delete cosmo;
#endif
  }
#endif
  
  return chi2;
}


// SMALL APPLICATIONS USEFUL ELSEWHERE



double chi2_for_intrinsic_dispersion(const double& intrinsic_dispersion, const void* context) {
  
  box* b = (box*)(context);
  b->hubblefitter->GetHubbleModel()->IntrinsicDispersion = intrinsic_dispersion;
  double r = b->hubblefitter->TheChi2()/b->ndf-1.;
  return r*r;
} 


// to maginalize
