#include <iostream>
#include <string>
#include <fstream>
#include <algorithm>
#include <cstdio>

#include "pathutils.h"
#include "fileutils.h"

#include "hubblemagmodel2.h"
#include "fittedsn.h"
#include "cosmology.h"

#include "matvect.h"
#include "malmquist_bias.h"


#ifndef Pi
#define Pi               3.14159265358979323846  /* pi */
#endif

//#define DEBUG

#define Z_PIVOT 0.5

static double sqr(double x) {return x*x;}

HubbleMagModel2::HubbleMagModel2(int input_number_of_galaxy_types) :
  HubbleModel()
{
  number_of_galaxy_types = input_number_of_galaxy_types ;
  
  // saved in hubbleparams in init params
  alpha = new LcParam*[number_of_galaxy_types] ;// ( Those are LcParams* table)
  beta  = new LcParam*[number_of_galaxy_types] ;
  Mprime = new LcParam*[number_of_galaxy_types] ;  
  for (int ii = 0; ii < number_of_galaxy_types ; ii++)
    {alpha[ii]=0;beta[ii]=0;Mprime[ii]=0;}

  

#ifdef DALPHABETADZ
  d_alpha_dz = new LcParam*[number_of_galaxy_types] ;// ( Those are LcParams* table)
  d_beta_dz  = new LcParam*[number_of_galaxy_types] ;
  for (int ii = 0; ii < number_of_galaxy_types ; ii++)
    {d_alpha_dz[ii]=0;d_beta_dz[ii]=0;}
#endif



  lc_model=SALT2;
  
#ifdef WITH_HOST_EXTINCTION 
  Rb = 0;
  has_cardelli_extinction = false;
#endif
#ifdef WITH_UV_FIT
  beta_uv=0;
  has_uv_fit = false;
#endif
}

double HubbleMagModel2::Mag(const FittedSn* sn) const
{
#ifdef DEBUG
  cout << sn->mag << " " << sn->shape << " " << sn->color << endl;
#endif
  int gtype = sn->GetGType() ;
  // when the number of galaxy types doesn't match the type
  // ex : when fitting wo type
  if (number_of_galaxy_types == 1)
    gtype = 0 ;

  double mag = sn->mag->val + alpha[gtype]->val*sn->shape->val - beta[gtype]->val*sn->color->val;

#ifdef DALPHABETADZ
   mag +=  (sn->GetZ()-Z_PIVOT)*(d_alpha_dz[gtype]->val*sn->shape->val - d_beta_dz[gtype]->val*sn->color->val);
#endif
   
#ifdef WITH_HOST_EXTINCTION  
  if(has_cardelli_extinction) {
    mag -= (Rb->val)*sn->ebv->val;
  }
#endif
#ifdef WITH_UV_FIT  
  if(has_uv_fit) {
    mag -= (beta_uv->val)*sn->uv->val;
  }
#endif
  
  if(malmquist_bias_correction)
    mag -= malmquist_bias(sn->GetZ(),sn->GetSet());

  return mag;
}

double HubbleMagModel2::Mprime_(const FittedSn* sn) const{
  int gtype = sn->GetGType() ;
  if  (number_of_galaxy_types == 1)
    gtype = 0 ;
  return Mprime[gtype]->val ;
}


double HubbleMagModel2::Res(const FittedSn* sn) const
{
  // get distance
  Cosmology *cosmomodel=GetCosmology(); // this is a function of HubbleModel
  double dl = cosmomodel->Dl(sn->GetZ());
  delete cosmomodel;
  
  int gtype = sn->GetGType() ;
  if (number_of_galaxy_types == 1)
    gtype = 0 ;
  
  return Mag(sn) - Mprime[gtype]->val - 5*log10(dl) -43.16; 
}



double HubbleMagModel2::Variance(const FittedSn* sn) const
{

  int gtype = sn->GetGType() ;
  if (number_of_galaxy_types == 1)
    gtype =0;

  int nparams = 3;

#ifdef WITH_HOST_EXTINCTION
  int index_cardelli=nparams;
  if(has_cardelli_extinction)
    nparams++;
#endif
#ifdef WITH_UV_FIT  
  int index_uv=nparams;
  if(has_uv_fit)
    nparams++;
#endif

  Vect V(nparams); //
  V(0) = 1;
  V(1) = alpha[gtype]->val; // X1
  V(2) = -beta[gtype]->val; // color  

#ifdef DALPHABETADZ
  V(1) += (sn->GetZ()-Z_PIVOT)*d_alpha_dz[gtype]->val;
  V(2) -= (sn->GetZ()-Z_PIVOT)*d_beta_dz[gtype]->val;  
#endif
  
#ifdef WITH_HOST_EXTINCTION
  if(has_cardelli_extinction)
    V(index_cardelli) = -Rb->val; // ebv
#endif
#ifdef WITH_UV_FIT 
  if(has_uv_fit)
    V(index_uv) = -beta_uv->val; // ebv
#endif


  Mat K(nparams,nparams);
  K(0,0) = sqr(sn->mag->err)
    + sqr(2.1714724*0.001/sn->GetZ()) // peculiar velocity (300 km/s), 2.1714724=5/ln(10)
    + sqr(IntrinsicDispersion); // intrinsic dispersion
  K(1,1) = sqr(sn->shape->err);
  K(2,2) = sqr(sn->color->err);

  K(1,0) = K(0,1) = sn->Params()->GetCovariance("mag","shape");
  K(2,0) = K(0,2) = sn->Params()->GetCovariance("mag","color");
  K(2,1) = K(1,2) = sn->Params()->GetCovariance("shape","color");
  
  
#ifdef WITH_HOST_EXTINCTION
  if(has_cardelli_extinction) {
    K(index_cardelli,index_cardelli) = sqr(sn->ebv->err);
    K(index_cardelli,0) = K(0,index_cardelli) = sn->Params()->GetCovariance("mag","EBV");
    K(index_cardelli,1) = K(1,index_cardelli) = sn->Params()->GetCovariance("shape","EBV");
    K(index_cardelli,2) = K(2,index_cardelli) = sn->Params()->GetCovariance("color","EBV");
  }
#endif
#ifdef WITH_UV_FIT 
  if(has_uv_fit) {
    K(index_uv,index_uv) = sqr(sn->p_uv->err);
    K(index_uv,0) = K(0,index_uv) = sn->Params()->GetCovariance("mag","UV");
    K(index_uv,1) = K(1,index_uv) = sn->Params()->GetCovariance("shape","UV");
    K(index_uv,2) = K(2,index_uv) = sn->Params()->GetCovariance("color","UV");
  }
#endif

  return scalar_product(V, K*V);
}



void HubbleMagModel2::InitParams() {
  // parameters for cosmo
  HubbleModel::InitParams();
  
  // parameters for SNe

  // delete pre-existing parameters
  if(number_of_galaxy_types>1) {
    hubbleparams.DeleteParamIfExists("Alpha");
    hubbleparams.DeleteParamIfExists("Beta");
    hubbleparams.DeleteParamIfExists("Gamma");
    hubbleparams.DeleteParamIfExists("Mprime");
#ifdef DALPHABETADZ
    hubbleparams.DeleteParamIfExists("dAlphadz");
    hubbleparams.DeleteParamIfExists("dBetadz");
#endif
  }

#ifdef WITH_HOST_EXTINCTION  
  if( has_cardelli_extinction ) {
    Rb = hubbleparams.LocateOrAddNonConstParam("Rb");
    if(!Rb->isfixed){
	Rb->val = 0;
	Rb->step = 0.1;
	Rb->minv = -100.;
	Rb->maxv = 100;
      }
  }
#endif
#ifdef WITH_UV_FIT
  if( has_uv_fit ) {
    beta_uv = hubbleparams.LocateOrAddNonConstParam("BetaUV");
    if(!beta_uv->isfixed){
	beta_uv->val = 0;
	beta_uv->step = 0.1;
	beta_uv->minv = -100.;
	beta_uv->maxv = 100;
      }
  }
#endif
  
  // brighter-slower corr. parameter
  for(int i = 0 ; i < number_of_galaxy_types ; i++)
    {
      char snum[10];
      sprintf(snum, "%d", i);
      string ssnum = snum ;
      if (number_of_galaxy_types == 1)
	ssnum = "" ;
      //#define ONLY_ADDITIONAL_M (same alpha and beta for all fits)       
#ifdef ONLY_ADDITIONAL_M 
      if(i>0) { // only one other Mprime (no alpha beta split)
	alpha[i] = alpha[0];
	beta[i] = beta[0];
#ifdef DALPHABETADZ
	d_alpha_dz[i] = d_alpha_dz[0];
	d_beta_dz[i] = d_beta_dz[0];
#endif	
	
      }else{
#endif

      string alpha_i = "Alpha"+ssnum;       
      alpha[i] = hubbleparams.LocateOrAddNonConstParam(alpha_i); 
      if(!alpha[i]->isfixed){
	alpha[i]->val = 0.1;
	alpha[i]->step = 0.01;
	alpha[i]->minv = -100.;
	alpha[i]->maxv = 100;
      }
  
      // brighter-bluer corr. parameter
      string beta_i = "Beta"+ssnum;
      beta[i] = hubbleparams.LocateOrAddNonConstParam(beta_i);
      if(!beta[i]->isfixed){
	beta[i]->val = 2;
	beta[i]->step = 0.1;
	beta[i]->minv = -100.;
	beta[i]->maxv = 100;
      }
      
#ifdef DALPHABETADZ
      string d_alpha_dz_i = "dAlphadz"+ssnum;       
      d_alpha_dz[i] = hubbleparams.LocateOrAddNonConstParam(d_alpha_dz_i); 
      if(!d_alpha_dz[i]->isfixed){
	d_alpha_dz[i]->val = 0.;
	d_alpha_dz[i]->step = 0.01;
	d_alpha_dz[i]->minv = -100.;
	d_alpha_dz[i]->maxv = 100;
      }
  
      // brighter-bluer corr. parameter
      string d_beta_dz_i = "dBetadz"+ssnum;
      d_beta_dz[i] = hubbleparams.LocateOrAddNonConstParam(d_beta_dz_i);
      if(!d_beta_dz[i]->isfixed){
	d_beta_dz[i]->val = 2;
	d_beta_dz[i]->step = 0.02;
	d_beta_dz[i]->minv = -100.;
	d_beta_dz[i]->maxv = 100;
      }
#endif
#ifdef ONLY_ADDITIONAL_M    
 }
#endif
      
      // absolute magnitude
      string mprime_i = "Mprime"+ssnum;
      Mprime[i] = hubbleparams.LocateOrAddNonConstParam(mprime_i);
      if(!Mprime[i]->isfixed){
	Mprime[i]->val = -19.3;
	Mprime[i]->step = 0.2;
	Mprime[i]->minv = -1000.;
	Mprime[i]->maxv = 1000;
      }
    }
}

