#include "instrument.h"
#include "filterwheel.h"
#include "datacards.h"
#include "pathutils.h"
#include "fileutils.h"

/* 1 Angst */
#define WAVELENGTH_STEP 1.

/* The whole thing should be redesigned to allow for a more abstract
   definition of what may vary in an Instrument transmission.
   As it is now, it only provides variations of the filters as a function
   of something. One may need at some point that e.g. the atmospheric 
   transmission depends on something.
   We'll think about it when it becomes necessary....
*/


//#define DEBUG

static Sampled1DFunction *read_trans(const DataCards& cards,
				     const string & TagName,
				     const string &PathToMainFile)
{
#ifdef DEBUG
  cout << "in read_trans :" << endl;
  cout << "  TagName = " << TagName << endl;
  cout << "  PathToMainFile = " << PathToMainFile << endl;
#endif
  
  if (!cards.HasKey(TagName) || cards.SParam(TagName) == "1")
    {
      return new Constant1DFunction(1.);
    }
  else
    {
    string s = AssembleFileName(PathToMainFile,
				cards.SParam(TagName));
    return new  General1DFunction(s);
    }
}



Instrument::Instrument(const string & Name) 
{
  optics_transmission = NULL;
  mirror_reflectivity = NULL;
  qe = NULL;
  atmospheric_transmission = NULL;
  filters = NULL;
  filename = LocateFileFromCard(Name)+"/instrument.cards";

#ifdef DEBUG
  cout << " Reading '" << filename << "' for instrument " << Name 
       << std::endl;
#endif
  
  DataCards cards(filename);
  string pathToMainFile = DirName(filename);
  instrumentName = BaseName(pathToMainFile);


  optics_transmission =      read_trans(cards, "OPTICS_TRANS", pathToMainFile);
  mirror_reflectivity =      read_trans(cards, "MIRROR_REFLECTIVITY", pathToMainFile);
  qe =                       read_trans(cards, "QE", pathToMainFile);
  atmospheric_transmission = read_trans(cards, "ATMOSPHERIC_TRANS", pathToMainFile);
  
  


  if (cards.HasKey("MIRROR_AREA"))
    {
      mirror_area = cards.DParam("MIRROR_AREA");
#ifdef DEBUG    
      cout << " mirror area (cm^2) " << mirror_area << endl;
#endif
    }
  else mirror_area = 1;

  if (cards.HasKey("PIX_SIZE"))
    {
      pix_size = cards.DParam("PIX_SIZE");
#ifdef DEBUG      
      cout << " pixel size(\") " << pix_size << endl;
#endif
    }
  else pix_size = 0;
  if (cards.HasKey("RADIALLY_VARIABLE_FILTERS"))
    {
      filters = new RadiallyVariableFilterWheel(AssembleFileName(pathToMainFile,
								 cards.SParam("RADIALLY_VARIABLE_FILTERS")),Name); 
    }
  else if (cards.HasKey("FILTERS"))
    filters = new NoParamFilterWheel(AssembleFileName(pathToMainFile, 
						      cards.SParam("FILTERS")),
				     Name);

  else 
    cout << "WARNING Instrument::Instrument : NO filterwheel found in " << filename  << endl;
  
  if (cards.HasKey("EFF_FILTERS")) {
    cout << "WARNING Instrument::Instrument ignoring EFF_FILTERS in file " << filename << endl
	 << "WARNING Instrument::Instrument Please remove this deprecated line" << endl;
  }

  if (cards.HasKey("PSF_CORRECTION_FILTERS")) {
    PSFCorrections = FilterWheel(AssembleFileName(pathToMainFile, 
						  cards.SParam("PSF_CORRECTION_FILTERS")),Name);
    cout << "INFO Instrument::Instrument found PSF correction filters " << cards.SParam("PSF_CORRECTION_FILTERS") << endl;
  }


#ifdef DEBUG
  cout << " for instrument " << Name << endl;
  cout << " filterwheel content " << endl << *filters << endl;
#endif

}


Filter Instrument::EffectiveFilterByBand(const string &Band, 
					 const IndexParams* P) const
{
  Filter filter = filters->GetFilterByBand(Band, P);
  /* overwrite the data part of the filter we just got (we actually own it!)
     with the result of the product
  */
  (General1DFunction &) filter = Product(WAVELENGTH_STEP,5, 
					 mirror_reflectivity, optics_transmission, 
					 &filter, qe, atmospheric_transmission);
  filter *= mirror_area;
  return filter;
}


FilterWheel Instrument::EffectiveFilters(const IndexParams*P) const
{
  std::vector<std::string> names = filters->BandNames();
  FilterWheel wheel;
  for (std::vector<std::string>::const_iterator i=names.begin(); i != names.end(); ++i)
    {
      wheel.push_back(EffectiveFilterByBand(*i, P));
    }
  return wheel;
}

bool Instrument::HasPSFCorrectionForBand(const std::string &Band) const 
{ 
  for (list<Filter>::const_iterator it = PSFCorrections.begin(); it != PSFCorrections.end() ; ++it)
    if (it->Band() == Band) return true;
  return false;
}

const Filter& Instrument::PSFCorrectionForBand(const std::string &Band) const
{
  return PSFCorrections.GetFilterByBand(Band);
}


#ifdef STORAGE /* I think it is not used */
//! if you have an effective filter in hand, returns the corresponding raw one 
const Filter Instrument::GetRawFilter(const Filter &EffectiveFilter)  const
{
  int pos=1;
  bool found = false;
  for (FilterCIterator it = effective_filters.begin(); it != effective_filters.end(); ++it)
    {
      const Filter &eff_filter = *it;
      if (eff_filter.Name() == EffectiveFilter.Name()) {found = true; break;}
      pos++;
    }
  if (!found)
    {
      cerr << "Instrument::GetRawFilter called with an argument which is not an effective filter : " 
	   << EffectiveFilter.Name() << endl;
      cerr << "crashing now " << endl;
      abort();
    }
  return filters.GetFilter(pos);
}
#endif /* STORAGE */
