#include <string>
#include <cstdio>
#include <iostream>

#include <legendre.h>
#include <fatalerror.h>



//! base of (not-normalized) Legendre polynomials
static double LegendrePol(const int Degree, const double &X)
{
  switch (Degree)
    {
    case 0: return 1; break;
    case 1: return X; break;
    case 2: return 0.5*(3*X*X-1.); break;
    case 3: return 0.5*(5*X*X-3)*X; break;
    case 4: return 0.125*((35*X*X-30)*X*X+3); break;
    case 5: return 0.125*(((63*X*X-70)*X*X+15)*X); break;
    case 6: return (((231*X*X-315)*X*X+105)*X*X-5)/16.; break;
    default : 
      return ((2*Degree-1)*X*LegendrePol(Degree-1, X) - 
	      (Degree-1)*LegendrePol(Degree-2, X))/double(Degree);
      /*
      FatalError(" consider implementing Legendre pol of abitrary degree");
      return 1; 
      */
      break;
    }
}

#include <cstdlib> // for abort()

LegPolSum3::LegPolSum3(const string &FileName)
{
  if(!Read(FileName)) {
    abort();
  }
}

LegPolSum3::LegPolSum3(int Dx, int Dy, int Dz,
	       double XMin, double XMax, 
	       double YMin, double YMax, 
	       double ZMin, double ZMax)
    : degx(Dx), degy(Dy), degz(Dz)
{
  coeffSize = (degx+1)*(degy+1)*(degz+1);
  coeffs.reserve(CoeffSize());
  ax = 2/(XMax-XMin); bx = -ax*XMin-1;
  ay = 2/(YMax-YMin); by = -ay*YMin-1;
  az = 2/(ZMax-ZMin); bz = -az*ZMin-1;
}


bool LegPolSum3::Write(const string &FileName)
{
  FILE *f = fopen(FileName.c_str(),"w");
  if (!f) 
    {
      cerr << " cannot open " << FileName << endl;
      return false;
    }
  fprintf(f,"%d %d %d\n",degx,degy,degz);
  fprintf(f,"%12g %12g %12g %12g %12g %12g \n",ax,bx,ay,by,az,bz);
  for (int i=0; i<coeffSize; ++i)
    fprintf(f,"%12g \n",coeffs[i]);
  fclose(f);
  return true;
}

void LegPolSum3::SetCoeffs(const double *Val)
{
  for (int i=0; i < coeffSize; ++i) coeffs[i] = Val[i];
}

bool LegPolSum3::Read(const string &FileName)
{
  FILE *f = fopen(FileName.c_str(),"r");
  if (!f) 
    {
      cerr << " cannot open " << FileName << endl;
      return false;
    }
  if (fscanf(f,"%d %d %d\n",&degx,&degy,&degz)!=3) 
    FatalError("error reading degrees in"+FileName);
  coeffSize = (degx+1)*(degy+1)*(degz+1);
  coeffs.reserve(CoeffSize());

  if (fscanf(f,"%lg %lg %lg %lg %lg %lg \n",&ax,&bx,&ay,&by,&az,&bz) != 6)
    FatalError("error reading scales in"+FileName);
  for (int i=0; i<coeffSize; ++i)
    fscanf(f,"%lg \n",&coeffs[i]);
  fclose(f);
  
  return true;
}


	




int LegPolSum3::CoeffSize() const
{
  return coeffSize;
}

void LegPolSum3::Monomials(const double &X, const double &Y, 
			   const double &Z, double *Monom) const
{
  double xred = ax*X+bx;
  double yred = ay*Y+by;
  double zred = az*Z+bz;
  int index = 0;
  double pol1, pol2, pol3;
  for (int ix=0; ix<=degx; ++ix)
    {
      pol1 = LegendrePol(ix,xred);
      for (int iy=0; iy<= degy; ++iy)
	{
	  pol2 = LegendrePol(iy, yred);
	  for (int iz =0; iz <= degz;++iz)
	    {
	      pol3 = LegendrePol(iz, zred);
	      Monom[index++]= pol1*pol2*pol3;
	    }
	}
    }
}




double  LegPolSum3::Value(const double &X, const double &Y, 
			  const double &Z) const
{
  double h[1000];// no dynamic allocation...
  Monomials(X,Y,Z,h);
  double val =0;
  for (int i=0; i< coeffSize; ++i) val += h[i]*coeffs[i];
  return val;
}




