#include <string>
#include <cstdio>
#include <iostream>
#include <iomanip>
#include <cmath>
#include <legsplit.h>
#include <fatalerror.h>



//! base of (not-normalized) Legendre polynomials
static double LegendrePol(const int Degree, const double &X)
{
  switch (Degree)
    {
    case 0: return 1; break;
    case 1: return X; break;
    case 2: return 0.5*(3*X*X-1.); break;
    case 3: return 0.5*(5*X*X-3)*X; break;
    case 4: return 0.125*((35*X*X-30)*X*X+3); break;
    case 5: return 0.125*(((63*X*X-70)*X*X+15)*X); break;
    case 6: return (((231*X*X-315)*X*X+105)*X*X-5)/16.; break;
    default : 
      return ((2*Degree-1)*X*LegendrePol(Degree-1, X) - 
	      (Degree-1)*LegendrePol(Degree-2, X))/double(Degree);
      /*
      FatalError(" consider implementing Legendre pol of abitrary degree");
      return 1; 
      */
      break;
    }
}


static double LegendrePolDer(const int Degree, const double &X)
{
  if (Degree<0) return 0;
  switch (Degree)
    {
    case 0: return 0; break;
    case 1: return 1; break;
    case 2: return 3*X; break;
    case 3: return 0.5*(15*X*X-3); break;
    case 4: return 0.5*(35*X*X-15)*X; break;
    case 5: return 0.125*((315*X*X-210)*X*X+15); break;
    default :   
      /* P'(n,x) = (2n-1)*P(n-1,x) + P'(n-2,x) */
      return (2*Degree-1)*LegendrePol(Degree-1,X) + LegendrePolDer(Degree-2,X);
      break;
    }
}


LegSplit3::LegSplit3(int Dx, int Dy, int Dz,
	       double XMin, double XMax,
		     double XCut,
	       double YMin, double YMax, 
	       double ZMin, double ZMax)
    : degx(Dx), degy(Dy), degz(Dz)
{
  coeffSize = (degx+1)*(degy+1)*(degz+1);
  coeffs.reserve(CoeffSize());
  ax = 2/(XMax-XMin); bx = -ax*XMin-1;
  ax2 = 2/(XCut-XMin); bx2 = -ax2*XMin-1;
  ay = 2/(YMax-YMin); by = -ay*YMin-1;
  az = 2/(ZMax-ZMin); bz = -az*ZMin-1;
}

LegSplit3::LegSplit3(const LegSplit3& Right) {
  degx = Right.degx;
  degy = Right.degy;
  degz = Right.degz;
  coeffSize = (degx+1)*(degy+1)*(degz+1);
  coeffs.reserve(coeffSize);
  ax = Right.ax; bx = Right.bx; 
  ax2 = Right.ax2; bx2 = Right.bx2;
  ay = Right.ay; by = Right.by;
  az = Right.az; bz = Right.bz;
    
  for (int i=0; i < coeffSize; ++i) coeffs[i] = Right.coeffs[i];
}


bool LegSplit3::Write(const string &FileName)
{
  FILE *f = fopen(FileName.c_str(),"w");
  if (!f) 
    {
      cerr << " cannot open " << FileName << endl;
      return false;
    }
  fprintf(f,"3\n%d %d %d\n",degx,degy,degz);
  fprintf(f,"%12g %12g\n",ax2,bx2);
  fprintf(f,"%12g %12g %12g %12g %12g %12g \n",ax,bx,ay,by,az,bz);
  for (int i=0; i<coeffSize; ++i)
    fprintf(f,"%12g \n",coeffs[i]);
  fclose(f);
  return true;
}

void LegSplit3::SetCoeffs(const double *Val)
{
  for (int i=0; i < coeffSize; ++i) coeffs[i] = Val[i];
}

bool LegSplit3::Read(const string &FileName)
{
  FILE *f = fopen(FileName.c_str(),"r");
  if (!f) 
    {
      cerr << " cannot open " << FileName << endl;
       return false;
    }
  int dim;
  if (fscanf(f,"%d\n",&dim)!=1)
    FatalError("error reading dimension in "+FileName);
  if(dim!=3)
    FatalError("error reading dimension in "+FileName);
  
  if (fscanf(f,"%d %d %d\n",&degx,&degy,&degz)!=3) 
    FatalError("error reading degrees in "+FileName);
  coeffSize = (degx+1)*(degy+1)*(degz+1);
  coeffs.reserve(CoeffSize());
  if (fscanf(f,"%lg %lg\n",&ax2,&bx2)!=2)
    FatalError("error reading scales in "+FileName);
  
  if (fscanf(f,"%lg %lg\n%lg %lg\n%lg %lg\n",&ax,&bx,&ay,&by,&az,&bz) != 6)
    FatalError("error reading scales in "+FileName);
  for (int i=0; i<coeffSize; ++i)
    fscanf(f,"%lg \n",&coeffs[i]);
  fclose(f);
  
  return true;
}


int LegSplit3::CoeffSize() const
{
  return coeffSize;
}

void LegSplit3::Monomials(const double &X, const double &Y, 
			   const double &Z, double *Monom) const
{
  double xred = ax*X+bx;
  double xred2 = ax2*X+bx2;
  double yred = ay*Y+by;
  double zred = az*Z+bz;
  int index = 0;
  double pol1, pol2, pol3;
  for (int ix=0; ix<=degx; ++ix)
    {
      if (ix<2) pol1 = LegendrePol(ix,xred);
      else
	{
	  if (xred2<1) pol1 = (xred2-1)*(xred2-1)*LegendrePol(ix-2,xred2);
	  else pol1 = 0;
	}
      for (int iy=0; iy<= degy; ++iy)
	{
	  pol2 = LegendrePol(iy, yred);
	  for (int iz =0; iz <= degz;++iz)
	    {
	      pol3 = LegendrePol(iz, zred);
	      Monom[index++]= pol1*pol2*pol3;
	    }
	}
    }

//   for(int i=0;i<10;i++) 
//     cout << coeffs[i] << "=>" << Monom[i] << " ";
//   cout << endl;
}




//! Compute the derivative of the monomials 
void LegSplit3::DMonomials(const double &X,
			   const double &Y,
			   const double &Z,
			   double* DMonomDx,
			   double* DMonomDy,
			   double* DMonomDz) const
{
    double xred  = ax*X+bx;
    double xred2 = ax2*X+bx2;
    double yred  = ay*Y+by;
    double zred  = az*Z+bz;
    
    int ix, iy, iz, index=0;
    double polx, poly, polz;
    double dpolx, dpoly, dpolz;
    
    for(ix=0;ix<=degx;ix++) {
	if(ix<2) {
	    polx = LegendrePol(ix,xred);
	    dpolx = ax * LegendrePolDer(ix,xred);
	}
	else {
	    if(xred2<1) {
		polx = (xred2-1)*(xred2-1)*LegendrePol(ix-2,xred2);
		dpolx = 2*(xred2-1)*ax2*LegendrePol(ix-2,xred2) +
		        ax2*(xred2-1)*(xred2-1)*LegendrePolDer(ix-2,xred2);
	    }
	    else {
		polx = 0;
		dpolx = 0;
	    }
	}
	for(iy=0;iy<=degy;iy++) {
	    // compute the poly and dpoly
	    poly = LegendrePol(iy,yred);
	    dpoly = ay * LegendrePolDer(iy,yred);
	    
	    for(iz=0;iz<=degz;iz++) {
		// compute the polz and dpolz
		polz = LegendrePol(iz,zred);
		dpolz = az * LegendrePolDer(iz,zred);
		
		DMonomDx[index] = dpolx *  poly *  polz;
		DMonomDy[index] =  polx * dpoly *  polz;
		DMonomDz[index] =  polx *  poly * dpolz;
		cout << " i=" << setw(3) << index
		     << " x: " << setw(4) << xred
		     << " y: " << setw(4) << yred 
		     << " z: " << setw(4) << zred
		     <<  " ddx: " << setw(5) << DMonomDx[index]
		  //		     <<  " ddy: " << setw(5) << DMonomDy[index]
		  //		     <<  " ddz: " << setw(5) << DMonomDz[index] 
		     << endl;
		index++;
	    } // end iy loop
	} // end iz loop
    } // end ix loop
}



double  LegSplit3::Value(const double &X, const double &Y, 
			  const double &Z) const
{
  double h[1000];// no dynamic allocation...
  Monomials(X,Y,Z,h);
  double val =0;
  for (int i=0; i< coeffSize; ++i) val += h[i]*coeffs[i];
  return val;
}



// LegSplit3 & LegSplit3::operator =(const LegSplit3& Right) {
  

//   return (*this);
// }

LegSplit3 LegSplit3::operator +(const LegSplit3& Right) const {
  LegSplit3 res(*this);
  res += Right;
  return res;

}

#include <cstdlib> // for abort()

void LegSplit3::operator +=(const LegSplit3& Right) {
  if(!SameDegrees(Right)) {
    cerr << "Try to add LegSplit3 with not the same degrees, I quit" << endl;
    abort();
  }
  for(int i=0;i<coeffSize;i++) {
    coeffs[i] += Right.coeffs[i];
  }
}


bool LegSplit3::SameDegrees(const LegSplit3& Right) const {
  return (degx==Right.degx && degy==Right.degy && degz==Right.degz);
}


// LegSplitN
//=============================================================================

LegSplitN::LegSplitN() {
  Zero();
}

LegSplitN::LegSplitN(const string &FileName) {
  Zero();
  Read(FileName);
}

LegSplitN::~LegSplitN() {
  Free();
}

void LegSplitN::Scale(double s) {
  for(int i=0;i<coeffSize;i++) {
    coeffs[i]*=s;
  }
}

void LegSplitN::Zero() {
 dimension = 0;
 degrees = 0;
 a = 0;
 b = 0; 
 buffer = 0;
}
void LegSplitN::Free() {
  if(degrees) delete [] degrees;
  if(a) delete [] a;
  if(b) delete [] b;
  if(buffer) delete [] buffer;
}

void LegSplitN::Allocate(int dim,int *degrees_i) {
  if(dim<=0) {
    cerr << "Allocate for negative or null dimension !!" << endl;
    abort();
  }
  
  dimension = dim;
  
  Free();
  
  degrees = new int[dimension];
  a = new double[dimension];   
  b = new double[dimension]; 
  
  coeffSize=1;
  for(int d=0;d<dimension;d++) {
    degrees[d] = degrees_i[d];
    coeffSize*=(degrees[d]+1);
  }
  coeffs.reserve(coeffSize);
  buffer = new double[coeffSize];
}

LegSplitN::LegSplitN(int dimension_i, int* degrees_i,
	  double* XMin, double* XMax, double X1Cut) {
  
  if(dimension_i<=0) {
    cout << "LegSplitN FATAL ERROR null or negative dimension" << endl;
    abort();
  }
  
  Zero();
  Allocate(dimension_i,degrees_i);
  
  for(int d=0;d<dimension;d++) { 
    a[d] = 2./(XMax[d]-XMin[d]); b[d] = -a[d]*XMin[d]-1; 
  }
  ax2 = 2/(X1Cut-XMin[0]); bx2 = -ax2*XMin[0]-1;
}
 

LegSplitN::LegSplitN(const LegSplitN& Right) {
  
  Zero();
  Allocate(Right.dimension,Right.degrees);
  
  for(int d=0;d<dimension;d++) { 
    a[d] = Right.a[d];
    b[d] = Right.b[d];
  }
  ax2 = Right.ax2;
  bx2 = Right.bx2;  
  for (int i=0; i < coeffSize; ++i) coeffs[i] = Right.coeffs[i];
}

int LegSplitN::CoeffSize() const
{
  return coeffSize;
} 


void LegSplitN::SetCoeffs(const double *Val)
{
  for (int i=0; i < coeffSize; ++i) coeffs[i] = Val[i];
}

// Monomials[i_0+(degree_0+1)*(i_1+(degree_1+1)*(i_2+(degree_2+2)*(+i_3 + ... )))]= Produit_{0 <= d < dimension} LegendrePol(i_d,xred=a[d]*X[d]+b[d])

#define USEBUFFER

void LegSplitN::Monomials(const double* X, double *Monom) const {
  

  double *xred = new double[dimension];
  
  for(int dim=0;dim<dimension;dim++)
    xred[dim] = a[dim]*X[dim]+b[dim];
  double xred2 = ax2*X[0]+bx2;
  
  int ncoefs,id,indexbis;
  for(int index = 0; index<coeffSize; index++) {
    Monom[index]=1;
    ncoefs = coeffSize;
    id=0;
    indexbis = 0;
    for(int dim=0;dim<dimension;dim++){
      
      // calcul de l'indice id selon la direction d, pour une valeur de index
      ncoefs/=(degrees[dim]+1);
      id = (index-indexbis)/ncoefs;
      indexbis += (id*ncoefs);

      //cout << index << " " << d << " " << id << " " << ncoefs << endl;
      
      // multiplication par le pol de legendre pour la dimension d 
      if(dim>0 || id < 2)
	Monom[index] *= LegendrePol(id, xred[dim]);
      else {
	if(xred2<1) 
	  Monom[index] *= (xred2-1)*(xred2-1)*LegendrePol(id-2, xred2);
	else 
	  Monom[index] = 0;
      }      
    }
  }  

  delete [] xred;
  
}

//! Compute the monomials of derivative along axis N ( 0<N<dimension)
void LegSplitN::DMonomialsDXN(int N, const double* X, double * DMonomDxN) const {
  
  // coordonnees reduites
  double *xred = new double[dimension];
  for(int d=0;d<dimension;d++)
    xred[d] = a[d]*X[d]+b[d];
  double xred2 = ax2*X[0]+bx2;
  
  int ncoefs,id,indexbis;
  for(int index = 0; index<coeffSize; index++) {
    DMonomDxN[index]=1;
    ncoefs = coeffSize;
    id=0;
    indexbis = 0;
    for(int d=0;d<dimension;d++){ 
      
      // calcul de l'indice id selon la direction d, pour une valeur de index
      ncoefs/=(degrees[d]+1);
      id = (index-indexbis)/ncoefs;
      indexbis += (id*ncoefs); // (sert juste a calculer id)

            
      // multiplication par la derivee
      if(d==N) {
	if(d>0 || id < 2) {
	  DMonomDxN[index]*= a[d] * LegendrePolDer(id,xred[d]);
	}else{
	  if(xred2<1) {
	    DMonomDxN[index]*= ( 2*(xred2-1)*ax2*LegendrePol(id-2,xred2) +
				 ax2*(xred2-1)*(xred2-1)*LegendrePolDer(id-2,xred2) );
	  }else{
	    DMonomDxN[index]=0;
	  }
	}
      }else{ // multiplication par le polynome 
	if(d>0 || id < 2)
	  DMonomDxN[index] *= LegendrePol(id, xred[d]);
	else {
	  if(xred2<1) 
	    DMonomDxN[index] *= (xred2-1)*(xred2-1)*LegendrePol(id-2, xred2);
	  else 
	    DMonomDxN[index] = 0;
	}
      }
    }
  }

  delete [] xred;
}

int LegSplitN::GetDegree(int d)  const {
  if(d>=0 && d<dimension) 
    return degrees[d]; 
  else 
    return 0;
}

bool LegSplitN::IsValidValue(int d,double val) const {
  if(d<0 || d>=dimension) return false;
  return (fabs(a[d]*val+b[d])<1);
}

//! Compute the derivative of the monomials 
void LegSplitN::DMonomials(const double* X, double ** DMonomDx) const {
  
  for(int d=0;d<dimension;d++) {
    DMonomialsDXN(d,X,DMonomDx[d]);
  }
}

bool LegSplitN::SameDegrees(const LegSplitN& Right) const {
  for(int d = 0;d < dimension ; d++) {
    if ( degrees[d] != Right.degrees[d] )
      return false;
  }
  return true;
}

LegSplitN LegSplitN::operator +(const LegSplitN& Right) const {
  LegSplitN res(*this);
  res += Right;
  return res;

}

void LegSplitN::operator +=(const LegSplitN& Right) {
  if(!SameDegrees(Right)) {
    cerr << "Try to add LegSplitN with not the same degrees, I quit" << endl;
    abort();
  }
  for(int i=0;i<coeffSize;i++) {
    coeffs[i] += Right.coeffs[i];
  }
}

double  LegSplitN::Value(const double *X) const {
 
#ifdef USEBUFFER
  double *h =  buffer;
#else
  double *h = new double[coeffSize];
#endif

  Monomials(X,h);
  double val =0;
  for (int i=0; i< coeffSize; ++i) val += h[i]*coeffs[i];

#ifndef USEBUFFER
  delete [] h;
#endif
  return val;

}

double  LegSplitN::Derivative(int d, const double *X) const {
 
#ifdef USEBUFFER
  double *h =  buffer;
#else
  double *h = new double[coeffSize];
#endif

  DMonomialsDXN(d,X,h);
  double val =0;
  for (int i=0; i< coeffSize; ++i) val += h[i]*coeffs[i];

#ifndef USEBUFFER
  delete [] h;
#endif
  return val;

}



double  LegSplitN::Value(const double X0, 
			 const double X1, 
			 const double X2, 
			 const double X3) const 
			 
{
  if(dimension>4) {
    cerr << "Cannot use this function cause dimension= " << dimension << ">4" << endl;
    abort();
  }

  double X[4];
  X[0]=X0;
  X[1]=X1;
  X[2]=X2;
  X[3]=X3;
  
  
 

  return Value(X);
}

bool LegSplitN::Write(const string &FileName) const
{
  FILE *f = fopen(FileName.c_str(),"w");
  if (!f) 
    {
      cerr << " cannot open " << FileName << endl;
      return false;
    }
  fprintf(f,"%d\n",dimension);
  for(int d=0;d<dimension;d++) {
    fprintf(f,"%d ",degrees[d]);
  }
  fprintf(f,"\n");
  fprintf(f,"%12.12g %12.12g\n",ax2,bx2);
  for(int d=0;d<dimension;d++) {
    fprintf(f,"%12.12g %12.12g\n",a[d],b[d]);
  }
  for (int i=0; i<coeffSize; ++i)
    fprintf(f,"%12.12g \n",coeffs[i]);
  fclose(f);
  return true;
}


bool LegSplitN::Read(const string &FileName)
{
  FILE *f = fopen(FileName.c_str(),"r");
  if (!f) 
    {
      cerr << " cannot open " << FileName << endl;
      return false;
    }
  int dim;
  if (fscanf(f,"%d\n",&dim)!=1)
    FatalError("error reading dimension in "+FileName);
  
  int *degs = new int[dim];
  for(int d=0;d<dim;d++) {
    if (fscanf(f,"%d ",&degs[d])!=1)
      FatalError("error reading a degree in "+FileName);
  }
  Allocate(dim,degs);
  delete [] degs;
  
  if (fscanf(f,"\n%lg %lg \n",&ax2,&bx2) != 2)
    FatalError("error reading scales in"+FileName);
  for(int d = 0; d<dimension;d++) {
    if (fscanf(f,"%lg %lg \n",&a[d],&b[d]) != 2)
      FatalError("error reading scales in"+FileName);
  }
  for (int i=0; i<coeffSize; ++i) {
    fscanf(f,"%lg \n",&coeffs[i]);
    //printf("coef %d %f\n",i,coeffs[i]);
  }
  fclose(f);
  
  return true;
}

