#include "magsys.h"


#include <map>
#include <string>
#include <cmath>

#include "pathutils.h"
#include "integflux.h"
#include "snfitexception.h"
#include "sampledfunction.h"
#include "filter.h"
#include "datafile.h"
#include "globalval.h"

using namespace std;


static double AB_ref(const double &x, const void* p)
{
  if (p) {}; // avoids a warning
  return 0.1085/(x*x);
}


static std::string mag_tag(const string &InstName, const string &BandName)
{
  return InstName+"::"+BandName;
}


/* 3 shemes implemented:
   - Mags are extracted only from a spectrum : used for the theoretical AB
     This is triggered by the "AB" SysName

   - Mags of an object of know spectrum are provided for defined filters
      This is used for the SNLS Landolt-BD17 based calibration.
      Note that the magnitudes in the MAgacam filters are arbitrary, or
      more precisely, bound to specific catalogs (Regnault et al 2009)
       This is the default behaviour of the constructor below: it expects
      @SPECTRUM my_stuff.dat
      STANDARD U 9.724
      ....
      MEGACAM g 9.691 

   - Mags of an object of known spectrum are provided at specific
     wavelengths. This is the wat used for Vega mags (where the magnitudes
     of Vega are interpolated at the central wavelength of every filter)
     This is what was used for Astier et al (2006). The previous scheme
     provides a much more accurate conversion to fluxes.
     This is triggered by putting in the driver file:
     @SPECTRUM vega_bohlin.dat
     @MAGS_AT_WAVELENGTHS
     3600 0.03
     4600 0.27
     3600 0.03
     ....
*/

MagSys::MagSys(const string &SysName) : name(SysName)
{
  offsetFunction = NULL;
  spectrum = NULL;
  if (SysName == "AB")
    {
      cout << "INFO MagSys::MagSys pure AB system " << endl;
      spectrum = new Computed1DFunction(AB_ref);
      mags[mag_tag("*","*")] = 0.;
      return;
    }
  
  fileName = LocateFileFromCard(SysName);
  
  cout << "INFO MagSys::MagSys reading " << fileName << endl;

  DataFile dataFile(fileName);
  const GlobalVal &glob = dataFile.Globals();
  
  if (!glob.HasKey("SPECTRUM")) throw(SnfitException("No key SPECTRUM in "+fileName));
  string spectrumKey =glob.getStringValue("SPECTRUM");
  string refSpectrumFileName = LocateFile(spectrumKey);
  spectrum = new General1DFunction(refSpectrumFileName);
  

  if (glob.HasKey("MAGS_AT_WAVELENGTHS"))
    {
      offsetFunction = new General1DFunction(fileName);
      return;
    }

  char line[4096];
  const char *p;
  while ((p=dataFile.next_line(line,4096)))
    {
      char instName[1000];
      char bandName[1000];
      double val;
      if (sscanf(p,"%s %s %lf",instName,bandName,&val)!=3)
	throw(SnfitException("MagSys::MagSys() : do not understand : \""
			     +string(line)+"\" in file : "+fileName ));
      string tag=mag_tag(instName,bandName);
      if (mags.find(tag) != mags.end()) throw(SnfitException("MagSys::MagSys() : at least two entries for : "+tag));
      mags[tag] = val;
    }
}

MagSys::~MagSys() 
{ delete spectrum; delete offsetFunction;}

	 
double MagSys::Mag(const Sampled1DFunction& ObjSpectrum,
		   const Filter &ObserverFilter,
		   const double &z, bool do_check_range) const
{
  double flux = IntegFlux(ObjSpectrum,ObserverFilter, z, do_check_range);
  if (flux<=0) return 99;
  return -2.5*log10(flux)+ZeroPoint(ObserverFilter);
}

double MagSys::Mag(const  Sampled1DFunction& ObjSpectrum,
		   const Filter &ObserverFilter,
		   const Sampled1DFunction &RestframeFilter,
		   const double &z, bool do_check_range) const

{
  double flux = IntegFlux(ObjSpectrum,ObserverFilter, RestframeFilter,
			  z, do_check_range);
  if (flux<=0) return 99;
  return -2.5*log10(flux)+ZeroPoint(ObserverFilter);
}

double MagSys::ZeroPoint(const Filter &ObserverFilter) const
{
  double flux = IntegFlux(*spectrum,ObserverFilter,0,true);
  if (offsetFunction == NULL) // case with mags provided for instruments::filters
    {
      string tag=mag_tag(ObserverFilter.InstrumentName(), ObserverFilter.Band());
      map<string,double>::const_iterator i = mags.find(tag);
      if (i == mags.end()) i = mags.find(mag_tag("*","*"));
      if (i == mags.end()) throw(SnfitException("MagSys::ZeroPoint : no entry for filter " + tag + " in magsys " + name+" \n please edit file "+fileName));
      double mag = i->second;
      return 2.5*log10(flux)+mag;
    }
  else // MAGS_AT_WAVELENGTHS case
    {
      double mag = offsetFunction->Value(ObserverFilter.Mean());
      return 2.5*log10(flux)+mag;
    }
}
			     

using namespace std;

typedef map<std::string, const MagSys*> TMagSysMap;
static TMagSysMap   MagSysMap;


const MagSys* GetMagSys(const string &Name)
{
  TMagSysMap::iterator i = MagSysMap.find(Name);
  if (i != MagSysMap.end()) return i->second;
  MagSys *m = new  MagSys(Name);
  MagSysMap[Name] = m;
  return m;
}




