#include "measureddata.h"
#include "dictfile.h"
#include "fatalerror.h"
#include <cmath>
#include <fstream>
#include <fstream>
#ifdef USECFITSIO
#include <fitsio.h>
#endif

#include "fileutils.h" // DecomposeString
#include "instrument.h"
#include "sampledfunction.h"
#include "pathutils.h"
#include "hostabsorption.h"
#include "datafile.h"
#include "magsys.h"

#ifndef  WAVELENGTH_STEP
/* 10 Angst */
#define WAVELENGTH_STEP 10 
#endif

// c'est a cause du fait que minuit admet un nombre limite de characteres pour
// les noms des parametres et que j'utilise le nom du spectre pour
// identifier ces parametres de recalibration
#define MAX_NUMBER_OF_CHARACTERS_FOR_NAME_OF_SPECTRUM 6

//#define DEBUG

static double sqr(double x) {return x*x;}

double MeasuredPointList::GetSimulationZP() const{
  
  return GetMagSys(magSystem)->ZeroPoint(EffectiveFilter);
}




//#ifdef STORAGE
// errorfloor = errorfloordval times the peak flux, see jha thesis page 177
void MeasuredPointList::ApplyErrorFloor(double errorfloordval) {

  // cerr << "WARNING MeasuredPointList::ApplyErrorFloor is disabled" << endl; return;
  cout << "applying error floor : " << errorfloordval << endl;
#ifdef DEBUG
  cout << "apply error floor " << endl;
#endif
  
  if(errorfloordval==0)
    errorfloordval = 0.01;
  bool has_covmat = CovMat.SizeX()>0;
  double fluxmax = -1000;
  for(MeasuredPointIterator i= begin(); i!= end(); ++i) {
    MeasuredPoint *p = *i;
    if(fluxmax < p->flux) {
      fluxmax = p->flux;
    } 
  }
  
  double errorfloor = fluxmax * errorfloordval ;
  int count = 0;
  for(MeasuredPointIterator i= begin(); i!= end(); ++i, ++count) {
    MeasuredPoint *p = *i;
    p->fluxerr = sqrt(sqr(errorfloor)+sqr(p->fluxerr));
    if(has_covmat)
      CovMat(count,count) += sqr(errorfloor);
  }
  
}
//#endif


void MeasuredPointList::ApplyScaling() {
#ifdef DEBUG
  cout << "apply scaling " << endl;
#endif
  // flux scaling
  // ================================================
  // we want to scale those fluxes in order to match with the simulation 
  // zero point.
  // flux -> flux* 10^{ [-ZP_{data} + ZP_{simul}(instrument,EffectiveFilter,magSystem)]/2.5 }
  
  // if WeightMat exists, we have to scale it a the same time

  // compute simulation zp
  double simulation_zeropoint = GetSimulationZP();
  
  double scaling=1;
  Vect scale(size()); // a vector of scalings
  int count = 0;
  for(MeasuredPointIterator i= begin(); i!= end(); ++i) {
    MeasuredPoint *p = *i;
    scaling =  pow(10.,0.4*(-p->zp+simulation_zeropoint));
    
#ifdef DEBUG
    cout << "ZP   : " << p->zp << " --> " << simulation_zeropoint << endl;
    cout << "Flux : " << p->flux << " --> " << p->flux * scaling << endl;
#endif
    
    p->flux *= scaling;
    p->fluxerr *= scaling;
    p->zp = simulation_zeropoint;
    scale(count) = scaling;
    count++;
  }
  
  if(CovMat.SizeX()>0) {
#ifdef DEBUG
    cout << "Apply scaling to covariance matrix" << endl;
    cout << "Before : " << CovMat << endl;
#endif
    for(size_t i = 0 ; i<CovMat.SizeX() ; i++)
      for(size_t j = 0 ; j<CovMat.SizeX() ; j++)
	CovMat(i,j)*=(scale(i)*scale(j));
  }
#ifdef DEBUG
    cout << "After : " << CovMat << endl;
#endif
}

MeasuredPointList::MeasuredPointList(const string &FileName,bool applyscaling, double MWEBV,const string& filter2shift,double wavelengthshift)
{
  DictFile file(FileName);
  instrumentName = file.GlobalValue("INSTRUMENT");
  string bandName = file.GlobalValue("BAND");
  magSystem = file.GlobalValue("MAGSYS");
  
  

  if(file.HasGlobalKey("WEIGHTMAT")) {
    WeightMatFileName = file.GlobalValue("WEIGHTMAT");
    string filename =  AddSlash(DirName(FullFileName(FileName)))+WeightMatFileName;
    if(FileExists(filename)) {
      ifstream stream(filename.c_str());
      Mat WeightMat;
      WeightMat.readASCII(stream);
      stream.close();
      WeightMat.Symmetrize("U"); // cause may be small differences when reading the file
      CovMat = WeightMat;
      CovMat.SymMatInvert();
      for(size_t i=0;i<WeightMat.SizeX();i++) {
	if(WeightMat(i,i)<1.e-20) {
	  cerr << "in MeasuredPointList::MeasuredPointList, Null diagonal value in weight map" << endl;
	  abort();
	}
      }
    }else{
      FatalError("cannot open matrix ASCII file "+filename,true);
    }
  }else{
    WeightMatFileName="";
  }
  
  // // MWEBV=0; cout << "WARNING force MWEBV to 0 " << endl;
  
  Instrument instrument(instrumentName);
  
  if(file.HasGlobalKey("X_FOCAL_PLANE")) {
    double x = atof(file.GlobalValue("X_FOCAL_PLANE").c_str());
    double y = atof(file.GlobalValue("Y_FOCAL_PLANE").c_str());
    
    cout << "Has a focal plane position " <<  x << " " << y << endl;
    FocalPlanePosition P(x,y);
    EffectiveFilter = instrument.EffectiveFilterByBand(bandName,&P);
  }else{
    EffectiveFilter = instrument.EffectiveFilterByBand(bandName);
  }
  
  if(bandName==filter2shift) {
    cout << "!!!!!!!! WARNING SHIFTING EFFECTIVE FILTER " << bandName <<  " OF " << wavelengthshift << " A" << endl;
    EffectiveFilter.ShiftX(wavelengthshift);
  }

  EffectiveFilterWithMilkyWay = EffectiveFilter;
  vector<Sampled1DFunction*> terms;
  terms.push_back(&EffectiveFilter);
  
  if(fabs(MWEBV)>1.e-10) {
    double data[2];
    data[0] = MWEBV;
    data[1] = 3.1; 
    // mean Rv for the Galaxy (see E.L. Fitzpatrick astro-ph/9809387 , J.A. Cardelli et al., ApJ 425, 245 (1989) )
    terms.push_back(new Computed1DFunction(&FluxExtinctionCardelli,data));
  }
  
  // add here PSF wavelength dependent correction if exists
  if(instrument.HasPSFCorrectionForBand(bandName)) {
    cout << "INFO MeasuredPointList::MeasuredPointList PSF correction for " << bandName << endl;
    terms.push_back(new Filter(instrument.PSFCorrectionForBand(bandName)));
    
  }
  if( terms.size()>1) { // already EffectiveFilter in there
    
    (General1DFunction &) EffectiveFilterWithMilkyWay = Product(WAVELENGTH_STEP,terms);
    
    for(size_t i=1;i<terms.size();++i) // start at 1 cause dont want to delete EffectiveFilter
      delete terms[i];
    
  }
  
  
  int index = 0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line, ++index)
    {
      MeasuredPoint *m = new MeasuredPoint(EffectiveFilterWithMilkyWay,index);
      if (line->HasKey("Flux"))
	{
	  m->flux =  line->Value("Flux");
	  m->fluxerr =  line->Value("Fluxerr");
	  if (m->fluxerr <= 0)
	    FatalError(" negative or null error in "+FileName);
	  m->zp = line->Value("ZP");
	}
      else if (line->HasKey("Mag"))
	{
	  if (line->HasKey("Flux"))
	    FatalError("Cannot mix Flux And Mag in file"+FileName);
	  // have to compute the reference flux for mags
	  double mag = line->Value("Mag");
	  double errMag = line->Value("Magerr");
	  if(errMag==0)
	    errMag = 0.001;
	  m->zp = 0;
	  m->flux = pow(10.,0.4*(m->zp-mag));
	  // 0.921 = 0.4*log(10), with log(e) = 1.
	  m->fluxerr = m->flux*0.921*errMag;
	}
      m->date = line->Value("Date");
      
      push_back(m);
    }
  
  if(applyscaling) 
    ApplyScaling();
  

}


MeasuredPointList::iterator MeasuredPointList::erase(MeasuredPointList::iterator toremove) {
  if(CovMat.SizeX()==0 || size()==1) {
    return  ((list<CountedRef<MeasuredPoint> >*) this)->erase(toremove);
  }
  
  size_t index=0;
  MeasuredPointIterator  iter = begin();
  while(iter!=toremove) {
    if(iter==end()) {
      cout << "what a strange error, this iterator does not match any in the list" << endl;
      return toremove;
    }
    index++;
    ++iter;
  }
  //cout << "remove line and column at index " << index << endl;
  CovMat = CovMat.WithoutRows(index,index);
  CovMat = CovMat.WithoutColumns(index,index);
  
  return ((list<CountedRef<MeasuredPoint> >*) this)->erase(toremove);
}


#include <cstdio>




MeasuredData::MeasuredData(const vector<string> &LcNames,bool applyscaling,const string& filter2shift,double wavelengthshift)
{
 
  if (LcNames.size() == 0)
    {
      //std::cerr << " Warning in MeasuredData::MeasuredData : no lightcurve given" 
      //	<< std::endl;
      name = "unkown"; 
      mwEbv = 0;
      z_cmb = 0;
      z_heliocentric = 0;
      ccd  = "unkown"; 
      field = "unkown";
      set = "unkown";
      return;
    }
  snpath=DirName(LcNames[0]);
  
  string lightFileName = snpath+"/lightfile";
  FILE *file = fopen(lightFileName.c_str(), "r");
  mwEbv = 0;
  name=BaseName(snpath);
  if(name=="" || name==".")
    name = "unknown";
  ccd  = "unkown"; 
  field = "unkown";
  set = "unkown";
  
  z_cmb = -3;
  z_heliocentric = -3;
  float z_default = -3;
  if(file) {
    char line[1024];
    
    while (fgets(line,1024,file))
      {
	char *start = line;
	start = start+strspn(start," \t");
	if (*start == '#') continue;
	char *cr = strrchr(start,'\n'); if (cr) *cr = '\0';
	vector<string> words;
	DecomposeString(words,start," ");
	//      if (words.size() != 2)
	//	FatalError(" expect 'Tag value' pairs in "+lightFileName);
	if (words.size() < 2)  continue;
	if (words[0] == "NAME") name = words[1];
	if (StringToUpper(words[0]) == "REDSHIFT") z_default = atof(words[1].c_str());
	if (words[0] == "z_heliocentric") z_heliocentric  = atof(words[1].c_str());
	if (words[0] == "z_cmb") z_cmb  = atof(words[1].c_str());
	if (words[0] == "MWEBV") mwEbv = atof(words[1].c_str());
	if (words[0] == "CCD") ccd = words[1];
	if (words[0] == "Field") field = words[1];
	if (words[0] == "SET") set = words[1];
      }
    fclose(file);
    if(z_heliocentric<-2 && z_default>-2)
      z_heliocentric = z_default;
    if(z_cmb<-2 && z_default>-2)
      z_cmb = z_default;
    
  }else{
    cout << "WARNING in MeasuredData:MeasuredData, could not read " << lightFileName << endl;
    cout << "        DO NOT FIT THIS SN WITHOUT FIXING THIS" << endl; 
  }
#ifdef DEBUG
  cout << " name = " << name << endl;
  cout << " z_heliocentric = " << z_heliocentric << endl;
  cout << " z_cmb = " << z_cmb << endl;
  cout << " mwEbv = " << mwEbv << endl;
#endif

  for (size_t i=0 ; i < LcNames.size(); ++i)
    {
      pointLists.push_back(new MeasuredPointList(LcNames[i],applyscaling,mwEbv,filter2shift,wavelengthshift));
    }

}


ostream &operator << (ostream &s, const MeasuredPointList &L)
{
  // ecrire ici les trucs:
  s <<"#Date :" <<endl;
  s <<"#Flux :" <<endl;
  s <<"#Fluxerr :"<<endl;
  s <<"#ZP :"<<endl;
  s <<"#end :" <<endl;
  s <<"@INSTRUMENT "<< L.GetInstrumentName() <<endl;
  s <<"@BAND "<< L.GetEffectiveFilter().Band() <<endl;
  s <<"@MAGSYS "<< L.GetMagSystem() <<endl;
  if(L.GetWeightMatFileName()!="")
     s <<"@WEIGHTMAT "<< L.GetWeightMatFileName() <<endl;
  for (MeasuredPointCIterator i = L.begin(); i != L.end(); ++i)
    s << *(*i);
  return s;
}

ostream &operator << (ostream &s, const MeasuredPoint &M)
{
  //s << " day = "  M.Day() << " flux = " << M.Flux() << " +/- " << M.FluxErr();
  s << M.Day() << " " << M.Flux() << " " << M.FluxErr() << " " << M.Zp()<< endl;
  return s;
}
  

void MeasuredData::Print(const int Level) const
{
  std::cout << "Name  : " <<  Name() << std::endl
	    << "z_heliocentric     : " << Zhelio() << std::endl 
	    << "z_cmb     : " << Zcmb() << std::endl
	    << "MWEBV: " << MWEBV() << std::endl
	    << "used bands : " << BandNames() << endl;

  if (Level >=2)
    {
      for (size_t i=0; i <NBands() ; ++i)
	{
	  const MeasuredPointList &pointList = PointList(i);
	  std::cout << " data in band " << pointList.BandName() << std::endl;
	  cout << pointList;
	}
    }
}


std::vector<std::string> MeasuredData::BandNames() const
{
  std::vector<std::string> result;
  for (size_t i =0; i < pointLists.size(); ++i)
    result.push_back(BandName(i));
  return result;
}


const MeasuredPointList * MeasuredData::GetMeasuredPointListByBand(const string& band) const
{
  for (size_t i =0; i < pointLists.size(); ++i) {
    if(pointLists[i]->GetEffectiveFilter().Band()==band) {
      return pointLists[i];
    }
  }
  return 0;
}

ostream &operator << (ostream &s, const std::vector<std::string> V)
{
  for (size_t i = 0; i < V.size(); ++i)
    s << V[i] << ' ';
  return s;
}

#ifdef STORAGE
void MeasuredData::RemovePointList(const size_t N)
{
  if(N>=pointLists.size()) {
    cout << "try to remove pointlist past the end of datalist" << endl;
    abort();
    return;
  }
  //we need iterators in here
  vector< CountedRef<MeasuredPointList> >::iterator it = pointLists.begin();
  for(size_t i=0;i<N;i++) // N steps forward 
    it++;
  // now remove entry
  pointLists.erase(it);
}


void MeasuredData::KeepOnlyPointsBetween(double observer_datemin,double observer_datemax) const {
  // this routine is const cause we want to use it with lcfit, however we are going to use a
  // non const function for each measuredpointlist, so we have to use const_cast
  

  for(size_t i=0;i<NBands();i++) {
    MeasuredPointList* plist = const_cast<MeasuredPointList*>(&*(pointLists[i]));
    if(!plist->isDiscarded())
      plist->KeepOnlyPointsBetween(observer_datemin,observer_datemax);
  }

}
#endif

#ifdef USECFITSIO
static void fitserror(int status) {
  if(status) {
    cerr << " we got these successive errors:" << endl;
    fits_report_error(stderr, status);
    exit(status);
  }
}

static void read_a_fits_array(const string &FitsName,Vect &X, Vect &Y) {
  

  int status = 0;
  fitsfile *fptr = 0;
  fits_open_file(&fptr,FitsName.c_str(),0,&status); 
  fitserror(status); 
  
  char value[256];

  // first get the size of the spectrum
  fits_read_key(fptr, TSTRING, "NAXIS1", &value, NULL, &status); fitserror(status); 
  int ndata = atoi(value);
  X.allocate(ndata);
  Y.allocate(ndata);
  

  // read fluxes
  {
    float nullval = 0;
    int anynull;
    fits_read_img(fptr, TDOUBLE, 1, ndata, &nullval,  Y.NonConstData() , &anynull, &status); fitserror(status); 
  }
  
  // read scale
  fits_read_key(fptr, TSTRING, "CRPIX1", &value, NULL, &status); fitserror(status); 
  int ref_pix = atoi(value);
  fits_read_key(fptr, TSTRING, "CRVAL1", &value, NULL, &status); fitserror(status); 
  double ref_wavelength = atof(value);
  fits_read_key(fptr, TSTRING, "CDELT1", &value, NULL, &status); fitserror(status); 
  double wavelength_step = atof(value);
  
  for(int i=0;i<ndata;i++) {
    X(i) = ref_wavelength+(i+1-ref_pix)*wavelength_step;
  }
  fits_close_file(fptr, &status); fitserror(status);
}
#endif //  USECFITSIO

static void read_an_ascii_array(const string &FileName,Vect &X, Vect &Y,int Ycol) {
  DataFile dfile(FileName);
  int ndat = dfile.NDat();
  //cout << "ndata in fluxes = " << ndat << endl;
  X.allocate(ndat);
  Y.allocate(ndat);
  dfile.ReadCols(X.NonConstData(),Y.NonConstData(),ndat,Ycol);
}

static string fullpath(const string& SpectrumDescriptor, const string& filename) {
  return AddSlash(DirName(FullFileName(SpectrumDescriptor)))+filename;
}

void MeasuredSpectrum::ReadIUE(const DictFile& file) {
  
  size_t size = file.size();
  
  Vect Wavelength(size);
  Vect Flux(size);
  Vect Error(size);
    
  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); 
       ++line, ++i) {
    Wavelength(i) = line->Value("WAVELENGTH");
    Flux(i) = line->Value("FLUX");
    Error(i) = line->Value("SIGMA");
  }
  
  LoadData(Wavelength,Flux,Error);  
 
}

void MeasuredSpectrum::ReadCFA(const DictFile& file) {
  
  size_t size = file.size();
  
  Vect Wavelength(size);
  Vect Flux(size);
  Vect Error;
  bool has_errors = file.begin()->HasKey("ERR");
  if(has_errors)
    Error.allocate(size);

  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); 
       ++line, ++i) {
    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("FLUX");
    if(has_errors)
      Error(i) = line->Value("ERR");
  }
  if(has_errors)
    LoadData(Wavelength,Flux,Error);
  else
    LoadData(Wavelength,Flux);  
}

void MeasuredSpectrum::ReadVLT1(const DictFile& file) {
  
  size_t size = file.size();
  
  // Count valid points : MASK = 0 :
  int validsz = 0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line ) 
  {
    // ! S.Baumont modif : copied by C. Balland
    //
    // if( line->Value("MASK") != 1 ) validsz += 1; 
    if( line->Value("MASK") != 1 && line->Value("WAVE") > 4250. && line->Value("WAVE") < 10000. ) validsz += 1;
  }

  Vect Wavelength(validsz);
  Vect Flux(validsz);
  Vect Error(validsz);
  Vect GalaxyFlux(validsz);
  Vect GalaxyError(validsz);

  

  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line)
    {
    //   ++line, ++i) {
    //if( line->Value("MASK") != 1 ){
    // ! S.Baumont modif : copied by C.Balland
    //
    if( line->Value("MASK") != 1 && line->Value("WAVE") > 4250. && line->Value("WAVE") < 10000. ){



    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("SN_SPEC");
    Error(i) = line->Value("SN_ERR");
    GalaxyFlux(i) = line->Value("GAL_SPEC");
    GalaxyError(i) = line->Value("GAL_ERR");  
    ++i;
    }
  }

   if( 0 ) {
    cout << "VLT spectra loaded : " << i << "/" << size << endl;
  }

  LoadData(Wavelength,Flux,Error);
  
  galaxy = new MeasuredSpectrum();
  galaxy->redshift = redshift;
  galaxy->is_a_galaxy = true;
  galaxy->LoadData(Wavelength,GalaxyFlux,GalaxyError);
  cout << "galaxy size = " << galaxy->size() << endl;
  
}

void MeasuredSpectrum::ReadVLT0(const DictFile& file) {
  
  size_t size = file.size();
  
  Vect Wavelength(size);
  Vect Flux(size);
  Vect Error(size);
  Vect GalaxyFlux(size);
  Vect GalaxyError(size);
  
  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); 
       ++line, ++i) {
    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("SN_SPEC");
    Error(i) = line->Value("SN_ERR");
    GalaxyFlux(i) = line->Value("GAL_SPEC");
    GalaxyError(i) = line->Value("GAL_ERR");  
  }
  
  LoadData(Wavelength,Flux,Error);
  
  galaxy = new MeasuredSpectrum();
  galaxy->redshift = redshift;
  galaxy->is_a_galaxy = true;
  galaxy->LoadData(Wavelength,GalaxyFlux,GalaxyError);
  cout << "galaxy size = " << galaxy->size() << endl;
  
}



//==================== For Gemini Data (CB 18/04/07) ===========

void MeasuredSpectrum::ReadGEMINI(const DictFile& file) {
  
  size_t size = file.size();
  
  // Count valid points : MASK = 0 :
  int validsz = 0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line ) 
  {
    // if( line->Value("MASK") != 1 ) validsz += 1; 
    if( line->Value("MASK") != 1 && line->Value("WAVE") > 5300. && line->Value("WAVE") < 10000. ) validsz += 1;
  }

  Vect Wavelength(validsz);
  Vect Flux(validsz);
  Vect Error(validsz);
  Vect GalaxyFlux(validsz);
  Vect GalaxyError(validsz);

  

  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line)
    {
    //   ++line, ++i) {
    //if( line->Value("MASK") != 1 ){
    // ! S.Baumont modif : copied by C.Balland
    //
    if( line->Value("MASK") != 1 && line->Value("WAVE") > 5300. && line->Value("WAVE") < 10000. ){



    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("SN_SPEC");
    Error(i) = line->Value("SN_ERR");
    GalaxyFlux(i) = line->Value("GAL_SPEC");
    GalaxyError(i) = line->Value("GAL_ERR");  
    ++i;
    }
  }

   if( 0 ) {
    cout << "GEMINI spectra loaded : " << i << "/" << size << endl;
  }

  LoadData(Wavelength,Flux,Error);
  
  galaxy = new MeasuredSpectrum();
  galaxy->redshift = redshift;
  galaxy->is_a_galaxy = true;
  galaxy->LoadData(Wavelength,GalaxyFlux,GalaxyError);
  cout << "galaxy size = " << galaxy->size() << endl;
  
}


//==================== For KECK Data (CB 21/05/07) ===========

void MeasuredSpectrum::ReadKECK(const DictFile& file) {
  
  unsigned int size = file.size();
  
  // Count valid points : MASK = 0 :
  int validsz = 0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line ) 
  {
    // if( line->Value("MASK") != 1 ) validsz += 1; 
    if( line->Value("MASK") != 1 && line->Value("WAVE") > 3400. && line->Value("WAVE") < 10000. ) validsz += 1;
  }

  Vect Wavelength(validsz);
  Vect Flux(validsz);
  Vect Error(validsz);
  Vect GalaxyFlux(validsz);
  Vect GalaxyError(validsz);

  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line)
    {
    //   ++line, ++i) {
    //if( line->Value("MASK") != 1 ){
    // ! S.Baumont modif : copied by C.Balland
    //
    if( line->Value("MASK") != 1 && line->Value("WAVE") > 3400. && line->Value("WAVE") < 10000. ){



    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("SN_SPEC");
    Error(i) = line->Value("SN_ERR");
    GalaxyFlux(i) = line->Value("GAL_SPEC");
    GalaxyError(i) = line->Value("GAL_ERR");  
    ++i;
    }
  }

   if( 0 ) {
    cout << "KECK spectra loaded : " << i << "/" << size << endl;
  }

  LoadData(Wavelength,Flux,Error);
  
  galaxy = new MeasuredSpectrum();
  galaxy->redshift = redshift;
  galaxy->is_a_galaxy = true;
  galaxy->LoadData(Wavelength,GalaxyFlux,GalaxyError);
  cout << "galaxy size = " << galaxy->size() << endl;
  
}




void MeasuredSpectrum::ApplyMask(const DictFile& file) {
  
  bool verbose = true;
  if(verbose) cout << "applying mask to spectrum " << endl;

  // get max flux for error floor
  //====================================
  double maxflux = 0;
  MeasuredSpectrum::iterator it=begin();
  while(it!=end()) {
    if(maxflux<(*it)->Flux())
      maxflux = (*it)->Flux();
    it++;
  }
  

  // lower weight in regions of atmospheric absoption lines
  // =============================================
  if(true){
    double lmin[] = {6850,7580};
    double lmax[] = {6970,7690};
    double min_error = maxflux/5;
    for(int m=0;m<2;m++) {
      MeasuredSpectrum::iterator it=begin();
      while(it!=end()) {
	const double& lambda = (*it)->Wavelength();
	if(lambda>lmin[m] && lambda<lmax[m]) {
	  double err = (*it)->FluxErr();
	  if(err<min_error)
	    (*it)->SetFluxErr(min_error);
	}
	it++;
      }
    }
  }
  
  // now read masks to remove points
  // =============================================
  char key[100];
  for(int maskindex=0;maskindex<100;maskindex++) {
    sprintf(key,"Mask%03d_min",maskindex);
    if(!file.HasGlobalKey(key)) {
      //cout << "no global key " << key << endl;
      continue;
    }

    double lmin = atof(file.GlobalValue(key).c_str());
    sprintf(key,"Mask%03d_max",maskindex);
    double lmax = atof(file.GlobalValue(key).c_str());

    if(verbose) cout << "removing points in mask " << maskindex <<" between " << lmin << " and " << lmax <<  endl;

    MeasuredSpectrum::iterator it=begin();
    while(it!=end()) {
      const double& lambda = (*it)->Wavelength();
      if(lambda>lmin && lambda<lmax) {
	it = erase(it);
	cout << "erasing point at lambda = " << lambda << endl;
	continue;
      }
      it++;
    }
  }
}


void MeasuredSpectrum::LoadData(const Vect& wavelength, const Vect& flux, const Vect& error) {
  
  size_t size = wavelength.size();
  if(size != flux.size()) { 
    cerr << "fatal error in MeasuredSpectrum::LoadData" << endl;
    exit(-12);
  }
  if(size != error.size()) { 
    cerr << "fatal error in MeasuredSpectrum::LoadData" << endl;
    exit(-12);
  }
  
  min_lambda =  wavelength(0);
  max_lambda =  wavelength(size-1);
  
  for(size_t i=0;i<size;i++) {
    MeasuredSpectrumPoint *p = new MeasuredSpectrumPoint();
    
    if(isnan(wavelength(i)) || isnan(flux(i)) || isnan(error(i)) ) {
      cerr << "WARNING nan in MeasuredSpectrum::LoadData" << endl;
      cout << "WARNING nan in MeasuredSpectrum::LoadData" << endl;
      continue;
    }
	


    p->SetWavelength(wavelength(i));
    p->SetFlux(flux(i));
    if(error(i)<=0) {
      cerr << "negative errors !!" << endl;
      p->SetFluxErr(flux(i));
    }else {
      p->SetFluxErr(error(i));
    }
    push_back(p);
  }
}
// tuned with spectrum calib of 1981B/spectrum_+00.dat
static double cooked_error(const double& flux, const double& reference_flux) {
  double min_error = reference_flux*0.06;
  double error = flux*0.02;
  if(error<min_error)
    error = min_error;
  return error;
}

void MeasuredSpectrum::LoadData(const Vect& wavelength, const Vect& flux) {
  cout << "cooking errors" << endl;
  size_t size = wavelength.size();
  if(size != flux.size()) { 
    cerr << "fatal error in MeasuredSpectrum::LoadData" << endl;
    exit(-12);
  }
  
  min_lambda =  wavelength(0);
  max_lambda =  wavelength(size-1);

  // very stupid estimate of errors
  double average_flux = 0;
  int iok = 0;
  for(size_t i=0;i<flux.size();i++)
    if(flux(i)>0) {
      average_flux += flux(i);
      iok++;
    }
  average_flux/=iok;
  
  for(size_t i=0;i<size;i++) {
    MeasuredSpectrumPoint *p = new MeasuredSpectrumPoint();
    p->SetWavelength(wavelength(i));
    p->SetFlux(flux(i));
    p->SetFluxErr(cooked_error(flux(i),average_flux));
    push_back(p);
  }
}

void MeasuredSpectrum::MilkyWayAbsorptionCorrection(const double& MWEBV) {
  if(milky_way_absorption_is_corrected) return;
  double data[2];
  data[0] = MWEBV;
  data[1] = 3.1; 
  double transmission = 1;
  // erase points if needed
  MeasuredSpectrum::iterator it=begin();
  while(it !=end()) {
    transmission = FluxExtinctionCardelli((*it)->Wavelength(),data);
    if(transmission<0.1) {
      it = erase(it); // huge assumption otherwise
    }else{
      double flux = (*it)->Flux()/transmission;
      double error = (*it)->FluxErr()/transmission;
      (*it)->SetFlux(flux);
      (*it)->SetFluxErr(error); 
      it++;
    }
  }
  milky_way_absorption_is_corrected = true;
}

void MeasuredSpectrum::SetName(const string& n) {
  name = n;
  if(name.size()> MAX_NUMBER_OF_CHARACTERS_FOR_NAME_OF_SPECTRUM) {
    cerr <<"Warning in MeasuredData, keeping only " << MAX_NUMBER_OF_CHARACTERS_FOR_NAME_OF_SPECTRUM << " characters of name " << name << endl;
    name=name.substr(0,MAX_NUMBER_OF_CHARACTERS_FOR_NAME_OF_SPECTRUM);  
  }
}

MeasuredSpectrum::MeasuredSpectrum(const string &FileName, bool mask) {
  
  if (! FileExists(FileName)) {
    cerr << "cannot find " << FileName << endl;
    exit(-12);
  }
  cout << "Loading spectrum " << FileName << endl;
  is_a_galaxy = false;
  from_space = false;
  milky_way_absorption_is_corrected = false;
  galaxy=0;
  number_of_calibration_parameters=-1;
  name="";
  //min_calibration_error = 1.e-8; // 
  //max_calibration_error = 1000.; //

  // read dictfile
  DictFile file(FileName);

  //if(file.HasGlobalKey("MINCALIBRATIONERROR")) {
  //  min_calibration_error = atof(file.GlobalValue("MINCALIBRATIONERROR").c_str());
  //}
  //if(file.HasGlobalKey("MAXCALIBRATIONERROR")) {
  //  max_calibration_error = atof(file.GlobalValue("MAXCALIBRATIONERROR").c_str());
  //}
  
  
  redshift  = atof(file.GlobalValue("Redshift").c_str());
  date      = atof(file.GlobalValue("Date").c_str());
  if(file.HasGlobalKey("Name")) {
    SetName(file.GlobalValue("Name"));
  }
  
  
  // is valid for pca
  valid4pca = false;
  if(file.HasGlobalKey("VALID4PCA")) {
    int val = atoi(file.GlobalValue("VALID4PCA").c_str());
    valid4pca = (val!=0);
    if(valid4pca) {
      cout << "valid for PCA" << endl;
    }else{
      cout << "not valid for PCA" << endl;
    }
  }
  
  if(file.HasGlobalKey("Phase")) {
    // cout << "has a phase info" << endl;
    phase = atof(file.GlobalValue("Phase").c_str());
  }else{
    phase     = 0;
  }
  
  
  

  // check format
  if(file.HasGlobalKey("FROMSPACE")) from_space=true; 
  
  if(file.HasGlobalKey("FORMAT")) {
    string format = file.GlobalValue("FORMAT");
    bool ok = false;
    if(format=="VLT0") { ReadVLT0(file); ok=true; } 
    if(format=="VLT1") { ReadVLT1(file); ok=true; mask=false; } // Modif S. Baumont 25/01/07 : mask=0
    if(format=="GEM") {ReadGEMINI(file); ok=true; mask=false; } //CB 18/04/2007
    if(format=="KECK") {ReadKECK(file); ok=true; mask=false; }  //CB 21/05/2007
    if(format=="IUE") { ReadIUE(file); ok=true; from_space=true;} 
    
    if(format=="CFA" || format=="FAKE") { ReadCFA(file); ok=true;} 
    if(!ok){
      cerr << "unknown FORMAT " << format << endl;
      exit(-12);
    }
    
  }else{
    
    // check cfitsio
    // ====================================
#ifndef USECFITSIO
    if(file.HasGlobalKey("Fits_Fluxes") || file.HasGlobalKey("Fits_Errors")) {
      cerr << "FATAL error, current installation of software cannot deal with fits files" << endl;
      exit(-12);
    }
#endif
    
    // data temporary container
    Vect Wavelength,Wavelength_for_error,Flux,Error;
    
    // read fluxes
    // ====================================
    if(file.HasGlobalKey("ASCII_Fluxes"))
      read_an_ascii_array(fullpath(FileName,file.GlobalValue("ASCII_Fluxes")),Wavelength,Flux,2);
#ifdef USECFITSIO
    if(file.HasGlobalKey("Fits_Fluxes"))
      read_a_fits_array(fullpath(FileName,file.GlobalValue("Fits_Fluxes")),Wavelength,Flux);
#endif
    if(file.HasGlobalKey("ASCII_Fluxes_and_Errors"))
      read_an_ascii_array(fullpath(FileName,file.GlobalValue("ASCII_Fluxes_and_Errors")),Wavelength,Flux,2); 
    if(Flux.Size()==0) {
      cerr << "No filename for fluxes given in " << FileName << endl;
      exit(-12);
    }
    
    // read errors
    // ====================================
    if(file.HasGlobalKey("ASCII_Errors"))
      read_an_ascii_array(fullpath(FileName,file.GlobalValue("ASCII_Errors")),Wavelength_for_error,Error,2);
#ifdef USECFITSIO
    if(file.HasGlobalKey("Fits_Errors"))
      read_a_fits_array(fullpath(FileName,file.GlobalValue("Fits_Errors")),Wavelength_for_error,Error);
#endif
    if(file.HasGlobalKey("ASCII_Fluxes_and_Errors"))
      read_an_ascii_array(fullpath(FileName,file.GlobalValue("ASCII_Fluxes_and_Errors")),Wavelength_for_error,Error,3);

    if(file.HasGlobalKey("ASCII_Errors") || file.HasGlobalKey("Fits_Errors")) {
      // separate file for errors and fluxes, we have to check stuff in here
      
      double flux_wavelength_step  = Wavelength(1)-Wavelength(0);
      double error_wavelength_step = Wavelength_for_error(1)-Wavelength_for_error(0);
      double eps = 0.001;
      if(fabs(flux_wavelength_step/error_wavelength_step-1)>eps ) {
	cerr << "flux and error files do not match " << endl;
	exit(-12);
      }
      if(Wavelength.size() != Wavelength_for_error.size()) { // we have to do something here
	cerr << "flux and error files do not match " << endl;
	
#ifdef STORAGE

	// find range of match
	int first_valid_bin=0;
	while(Wavelength(first_valid_bin)<Wavelength_for_error(0)*(1.-eps)) {first_valid_bin++;};
	int last_valid_bin=Wavelength.size()-1;
	while(Wavelength(last_valid_bin)>Wavelength_for_error(Wavelength_for_error.size()-1)*(1.+eps)) {last_valid_bin--;};
	
	if(last_valid_bin<=first_valid_bin) {
	  cerr << "flux and error files do not match " << endl;
	  exit(-12);
	}
	
	// find offset, so that Wavelength(i)=Wavelength_for_error(i+offset)*(1+-eps)
	int offset=-first_valid_bin;
	while(Wavelength(first_valid_bin)>Wavelength_for_error(first_valid_bin+offset)*(1.-eps) 
	      && (first_valid_bin+offset<int(Wavelength_for_error.size()))) 
	  {offset++;}; 
	if(offset>=int(Wavelength_for_error.size())) {
	  cerr << "flux and error files do not match " << endl;
	  exit(-12);
	}
	
	cout << "range of valid bins = " << first_valid_bin << "," << last_valid_bin << " in 0," <<  Wavelength.size()-1 << endl;
	cout << "offset = " << offset << endl;
#endif
      }


    } 
    
    // no errors were read so we cook them
    //====================================
    if(Error.Size()==0)  { 
      cout << "cook errors" << endl;
      double average_flux = 0;
      int iok = 0;
      for(size_t i=0;i<Flux.size();i++)
	if(Flux(i)>0) {
	  average_flux += Flux(i);
	  iok++;
	}
      average_flux/=iok;
      
      Error.allocate(Flux.size());
      for(size_t i=0;i<Flux.size();i++)
	Error(i)     = cooked_error(Flux(i),average_flux);
    }
    
  
    // if data are rest-frame, modify this
    //====================================
    if(file.HasGlobalKey("Frame") && (file.GlobalValue("Frame")=="Rest")) {
      Wavelength *= (1.+redshift);
      cout << "data were restframe!!" << endl;
    }

    LoadData(Wavelength,Flux,Error);
    
  } 
  
  if(mask) {
    ApplyMask(file);
  }
  if(Galaxy())
    galaxy->ApplyMask(file);
  
  
  if(file.HasGlobalKey("UNITS")) {
    double scale = atof(file.GlobalValue("UNITS").c_str());
    cout << "will rescale fluxes by " << scale << endl;

    for(list<CountedRef<MeasuredSpectrumPoint> >::iterator it=begin();it!=end();++it) {
      (*it)->SetFlux((*it)->Flux()*scale);
      (*it)->SetFluxErr((*it)->FluxErr()*scale);
    }
    if(galaxy) {
      for(list<CountedRef<MeasuredSpectrumPoint> >::iterator it=galaxy->begin();it!=galaxy->end();++it) {
	(*it)->SetFlux((*it)->Flux()*scale);
	(*it)->SetFluxErr((*it)->FluxErr()*scale);
       }
    }

  }

  cout << "done" << endl;
}

MeasuredSpectrum::~MeasuredSpectrum() {
  //cout << "destroying a measured spectrum" << endl;
  //cout << "clearing data ..." << endl;
  clear();
  //cout << "done" << endl;
  if(galaxy) {
    //cout << "deleting galaxy ... "  << endl;
    delete galaxy;
    //cout << "done" << endl;
  } 
}


bool MeasuredData::LoadSpectrum(const string& filename,double wavelenght_bin_size,bool warning) {
  if ( ! FileExists(filename)) {
    if(warning)
      cerr << "cannot find " << filename << endl;
    return false;
  }
  MeasuredSpectrum *spec = new MeasuredSpectrum(filename);
  if(spec->Name()=="") {
    char name[6];
    sprintf(name,"S%d",int(spectra.size()));
    spec->SetName(name);
  }
  spec->Rebin(wavelenght_bin_size);
  spec->MilkyWayAbsorptionCorrection(mwEbv); // divide by Galactic Extinction
  spectra.push_back(spec);
  
  return true;
}



void MeasuredSpectrum::Rebin(double wavelength_step) {
  
  if(wavelength_step<=0)
    return;

  if(galaxy) galaxy->Rebin(wavelength_step);

  // assume data are ordered ...
  double min_lambda = (*begin())->Wavelength();
  double max_lambda = (back())->Wavelength();
  
  MeasuredSpectrum rebinned_spectrum;
  size_t nbins = (size_t)((max_lambda-min_lambda)/wavelength_step);

  
  MeasuredSpectrum::const_iterator it=begin();
  double wavelength_step_in_spectrum_of_first_bin = -(*it)->Wavelength();
  it ++;
  wavelength_step_in_spectrum_of_first_bin += (*it)->Wavelength();
  
  double sumw,sumf;
  double bin_max_lambda,bin_min_lambda,weight;
  int n;
  bin_max_lambda = min_lambda;
  
  
  double wavelength_step_in_spectrum,previous_wavelength,wavelength;
  
  for(size_t ibin = 0; ibin < nbins ; ibin++) {
    sumw=0;
    sumf=0;
    n = 0;
    bin_min_lambda = bin_max_lambda;
    bin_max_lambda = bin_min_lambda + wavelength_step;
    
    //cout << "hello" << endl;

    previous_wavelength = 0;
    wavelength = 0;
    for(it=begin();it!=end();++it) {
      wavelength = (*it)->Wavelength();
      
      // step
      if(it!=begin()) {
	wavelength_step_in_spectrum = wavelength-previous_wavelength;
      }else{
	wavelength_step_in_spectrum = wavelength_step_in_spectrum_of_first_bin;
      }
      previous_wavelength = wavelength;
      
      double wavelength_spec_bin_min = wavelength-0.5*wavelength_step_in_spectrum;
      
      if(wavelength_spec_bin_min >  bin_max_lambda) break;

      double wavelength_spec_bin_max = wavelength+0.5*wavelength_step_in_spectrum;
      
      if(wavelength_spec_bin_max <  bin_min_lambda) continue;
      
      
      double bmin = max(wavelength_spec_bin_min,bin_min_lambda);
      double bmax = min(wavelength_spec_bin_max,bin_max_lambda);
      double frac = (bmax-bmin)/(wavelength_spec_bin_max-wavelength_spec_bin_min);
      
      
      //cout << bin_min_lambda << " " << bin_max_lambda << endl;
      //cout << wavelength_spec_bin_min << " " << wavelength_spec_bin_max << endl;
      //cout << bmin << " " << bmax << endl;
      //cout << "frac = " << frac << endl;
      
      weight = frac/sqr((*it)->FluxErr());
      sumw += weight;
      sumf += weight*((*it)->Flux());
      n++;
    }

#ifdef STORAGE
    while( ( it!=end() ) && ( (*it)->Wavelength() < bin_max_lambda ) ) {
      if( (*it)->Wavelength() >= bin_min_lambda ) {
	weight = 1./sqr((*it)->FluxErr());
	sumw += weight;
	//sumf += (*it)->Flux();
	sumf += weight*((*it)->Flux());
	n++;
      }else{
	cout << "mismatch" << endl;
      }
      ++it;
    }
    //--it;
#endif
    
    if(sumw>0) {
      //cout << n << endl;
      MeasuredSpectrumPoint *p = new MeasuredSpectrumPoint();
      //p->SetFlux(sumf/n);
      p->SetFlux(sumf/sumw);
      //p->SetFluxErr(sqrt(sumv)/n);
      p->SetFluxErr(sqrt(1./sumw));
      p->SetWavelength(0.5*(bin_min_lambda+bin_max_lambda));
      rebinned_spectrum.push_back(p);
    }
  }
  
  // load all of this in current spectrum
  clear();
  for(MeasuredSpectrum::const_iterator it=rebinned_spectrum.begin();it!=rebinned_spectrum.end();it++) {
    push_back(*it);
  }
  
}

bool MeasuredData::LoadSpectralSequence(const string& filename) {
  if( ! FileExists(filename)) {
    cout << "ERROR MeasuredData::LoadSpectralSequence file " << filename << " doesnt exist" << endl;
    return false;
  }
  DictFile file(filename);
  if(! file.HasGlobalKey("FORMAT")) {
    cout << "ERROR MeasuredData::LoadSpectralSequence no key FORMAT in file " << filename << endl;
    return false;
  }
  string format = file.GlobalValue("FORMAT");
  if(format != "SNFACTORY") {
    cout << "ERROR MeasuredData::LoadSpectralSequence read for FORMAT " << format  << " not implemented" << endl;
    return false;
  }
  z_heliocentric  = atof(file.GlobalValue("ZHELIO").c_str());
  z_cmb = atof(file.GlobalValue("ZCMB").c_str());

  if( file.HasGlobalKey("MWEBV")) {
    mwEbv = atof(file.GlobalValue("MWEBV").c_str());
  }else{
    cout << "WARNING MeasuredData::LoadSpectralSequence no key MWEBV, set it to 0" << endl;
    mwEbv = 0;
  }
  
  name = file.GlobalValue("OBJECT");
  
  int nspectra = atof(file.GlobalValue("NSPECTRA").c_str());
  if(nspectra<1) {
    cout << "ERROR MeasuredData::LoadSpectralSequence NSPECTRA =  " << nspectra << endl;
    return false;
  }
  
  
  
  double extinction_data[2];
  extinction_data[0] = mwEbv;
  extinction_data[1] = 3.1; // Rv
      
  

  double current_date = -12;
  
  MeasuredSpectrum *spectrum = 0;

  // now loop on data
  for (DictFileCIterator line = file.begin(); line != file.end(); 
       ++line)
    {
      double date = line->Value("MJD");
      
      if(current_date != date ) {
	// do something here
	if(spectrum) {
	  // save it
	  if(spectrum->size()>0)
	    spectra.push_back(spectrum);
	  else
	    cout << "WARNING MeasuredData::LoadSpectralSequence ignore empty spectrum " << spectrum->name << endl;
	}
	current_date = date;
	spectrum = new MeasuredSpectrum();
	spectrum->redshift = z_heliocentric;
	spectrum->date     = date;
	spectrum->phase    = 0; // unknown
	spectrum->number_of_calibration_parameters = 0; // SNFactory data are well calibrated
	spectrum->valid4pca = true; // don't know if still used
	spectrum->is_a_galaxy = false;
	spectrum->milky_way_absorption_is_corrected = false; // i don't know for the moment

	char specname[512];
	sprintf(specname,"%s-%d-%d",name.c_str(),int(date),int(100*date-100*int(date)));
	spectrum->name = specname;
      }
      
      // now add data point to current spectrum
      double flux_error = line->Value("SIG_FLUX");
      if(flux_error<0) {
	cout << "WARNING MeasuredData::LoadSpectralSequence ignore point with SIG_FLUX =  " << flux_error << endl;
	continue;
      }
      MeasuredSpectrumPoint * point = new MeasuredSpectrumPoint();
      point->wavelength = line->Value("WAVE");
      point->flux = line->Value("FLUX");
      point->fluxerr =  flux_error;
      if(mwEbv>0) {
	point->flux *= FluxExtinctionCardelli(point->wavelength,extinction_data);

      }
      spectrum->push_back(point);
    }


  

  return true;
}
