#ifndef MEASUREDDATA__H
#define MEASUREDDATA__H


#include <list>
#include <string>
#include <vector>
#include <math.h>

#include "countedref.h"
#include "filter.h"
#include "matvect.h"

class Filter;
//! \file
//! handles reading of photometry results

class Measurement : public RefCount
{
  protected :
    
    double flux, fluxerr;
  double zp;
  double date, errdate;
   
  public :
    Measurement() {}
  
  
  double Flux() const { return flux;}
  double FluxErr() const {return fluxerr;}
  double Zp() const { return zp;}
  double Day() const { return date;}
  double Date() const { return date;}
  double Mag() const { if(flux<=0) return 99; return -2.5*log10(flux)+zp;}
  double MagErr1() const {if((1.-fluxerr/flux)<=0) return 99; return -2.5*log10(1.-fluxerr/flux);}
  double MagErr2() const {if((1.+fluxerr/flux)<=0) return 99; return -2.5*log10(1.+fluxerr/flux);}
  
  void SetFlux(double v)  {  flux=v;}
  void SetFluxErr(double v)  { fluxerr=v;}
  void SetZp(double v)  {  zp=v;}
  void SetDay(double v)  {  date=v;}
  void SetDate(double v)  {  date=v;}
  
  
};

class MeasuredPoint : public Measurement
{ // use refcounts in case one wants to put the same point in several lists.


    private :
      
    const  Filter& filter; // effective filter including Milky Way extinction
    int index;
    
    public :
	MeasuredPoint(const Filter &F, int i) : Measurement(), filter(F), index(i) {}
    
    const Filter &GetFilter() const { return filter;}
    int Index() const {return index;} // index of measurement in the lighcturve (useful for fullfit)
    
    friend class MeasuredPointList;
};

class MeasuredSpectrumPoint : public Measurement
{ // use refcounts in case one wants to put the same point in several lists.


    private :
      double wavelength;
    
    public :
	MeasuredSpectrumPoint() : Measurement() {}
    
    double Wavelength() const {return wavelength;} 
    void SetWavelength(double w) {wavelength=w;} 
    
    friend class MeasuredSpectrum;
    friend class MeasuredData;
};


class DictFile;
class Vect;
class MeasuredSpectrum : public list<CountedRef<MeasuredSpectrumPoint> >, public RefCount
{      
  friend class MeasuredData;
  
  private :
  double redshift;
  double min_lambda;
  double max_lambda;
  double date;
  double phase;
  int number_of_calibration_parameters;
  string name; // can be used as an id.
  
  //double min_calibration_error; // like 0.7 for 30% error, lowe bound
  //double max_calibration_error; // like 1.3
  
  double sigma_calibration; // uncertainty on 'grey'=achromatic calibration df/f , 0 by default
  
  string fluxes_fits_filename;
  string errors_fits_filename;
  bool valid4pca;
  bool is_a_galaxy;
  MeasuredSpectrum * galaxy;
  //CountedRef<MeasuredSpectrum> galaxy;
  void ReadVLT0(const DictFile& file);
  void ReadVLT1(const DictFile& file);
  void ReadIUE(const DictFile& file);
  void ReadCFA(const DictFile& file);
  void ReadGEMINI(const DictFile& file); // CB 18/04/2007 for Gemini data
  void ReadKECK(const DictFile& file); // CB 21/05/2007
  void LoadData(const Vect& wavelength, const Vect& flux, const Vect& error);
  void LoadData(const Vect& wavelength, const Vect& flux);
  void ApplyMask(const DictFile& file);
  bool milky_way_absorption_is_corrected;

  public :
  MeasuredSpectrum() {redshift=0;date=0;galaxy=0;is_a_galaxy=false;number_of_calibration_parameters=0;from_space=false;milky_way_absorption_is_corrected=false;sigma_calibration=0;};
  MeasuredSpectrum(const string &FileName,bool mask=true);
  virtual ~MeasuredSpectrum();
  string Name() const {return name;};
  void SetName(const string &n);
  double Redshift() const {return redshift;};
  double Date() const {return date;};
  void SetRedshift(double z)  {redshift=z;};
  void SetDate(double d)  {date=d;};
  double GetMinLambda() const {return min_lambda;};
  double GetMaxLambda() const {return max_lambda;};
  //double GetMinCalibrationError() const {return min_calibration_error;};
  //double GetMaxCalibrationError() const {return max_calibration_error;};
  
  double GetPhase() const {return phase;};
  bool IsValid4PCA() const { return valid4pca;};
  int NumberOfCalibrationParameters() const {return number_of_calibration_parameters;}
  void SetNumberOfCalibrationParameters(unsigned n)  {number_of_calibration_parameters = n;}
  double SigmaCalibration() const {return sigma_calibration;};
  void SetSigmaCalibration(const double& s) {sigma_calibration=s;};
  void Rebin(double wavelength_step=10);
  const MeasuredSpectrum* Galaxy() const {return (const MeasuredSpectrum*)(galaxy);};
  MeasuredSpectrum* Galaxy()  {return ( MeasuredSpectrum*)(galaxy);};
  bool IsAGalaxy() const {return is_a_galaxy;};
  bool from_space;
  void MilkyWayAbsorptionCorrection(const double& MWEBV);
};


//! stuff of a given band
class MeasuredPointList : public list<CountedRef<MeasuredPoint> >, public RefCount
{
    private : 
      
      Filter EffectiveFilter; // effective filter of the instrument 
    Filter EffectiveFilterWithMilkyWay; // effective filter including MilkyWay extinction using Cardelli's law + ...
    
    string magSystem;
    string instrumentName;
    string WeightMatFileName;
    void ApplyScaling();    
    
    
    public :
      MeasuredPointList(){};
 MeasuredPointList(const Filter& effectivefilter) : EffectiveFilter(effectivefilter),EffectiveFilterWithMilkyWay(effectivefilter) {};
      MeasuredPointList(const string &FileName,bool applyscaling=true, double MWEBV=0,
			const string& filter2shift="none",double wavelengthshift=0);
      
    string BandName() const {return EffectiveFilter.Name();}
    const Filter &GetEffectiveFilter() const { return EffectiveFilter;}
    const Filter &GetEffectiveFilterWithMilkyWay() const { return EffectiveFilterWithMilkyWay;}
    
    Mat CovMat; // filled only once from the WeightMat
    
    double GetSimulationZP() const;
    const string& GetMagSystem() const {return magSystem;};
    const string& GetInstrumentName() const {return instrumentName;};
    const string& GetWeightMatFileName() const {return WeightMatFileName;};
    
    MeasuredPointList::iterator erase(MeasuredPointList::iterator toremove);
    //void KeepOnlyPointsBetween(double observer_datemin,double observer_datemax); 
    void ApplyErrorFloor(double errorfloordval = 0.);
    //void AddMeasuredPoint(CountedRef<MeasuredPoint> mp) {push_back(mp);};
    //void CreateWeightMat();
};


typedef MeasuredPointList::iterator MeasuredPointIterator;
typedef MeasuredPointList::const_iterator MeasuredPointCIterator;

ostream &operator << (ostream &, const MeasuredPointList&);
ostream &operator << (ostream &, const MeasuredPoint&);


class MeasuredData 
{
  private :
  
      vector<CountedRef<MeasuredPointList> > pointLists;
    vector<CountedRef<MeasuredSpectrum> > spectra;
    
    double z_heliocentric;
    double z_cmb;
    double mwEbv; // Milky Way E(B-V)
    string name; // of the object(SN)
    string ccd; 
    string field; 
    string snpath;
    string set;
    
    public :
      MeasuredData() {z_cmb=0;z_heliocentric=0;mwEbv=0;};
    MeasuredData(const vector<string> &LcNames,bool applyscaling=true,
		 const string& filter2shift="none",double wavelengthshift=0);

    bool LoadSpectralSequence(const string& filename);

    bool LoadSpectrum(const string& filename="spectrum.dat",double wavelenght_bin_size=-12,bool warning = true);

    //! number of available bands.
    size_t NBands() const { return pointLists.size();}
    
    //! number of spectra
    size_t NSpectra() const { return spectra.size();}
    
    const MeasuredPointList * GetMeasuredPointListByBand(const string& band) const;

    //! number of available bands.
    std::string BandName(const size_t N) const 
	{ return PointList(N).BandName();}

    //! access to bands  (labelled from 0 to NBands()-1)
    const MeasuredPointList &PointList(const size_t N) const
	{ return *pointLists[N];}

    //! access to bands  (labelled from 0 to NBands()-1
    MeasuredPointList &PointList(const size_t N) 
      { return *pointLists[N];}

    //! 
    const MeasuredSpectrum &Spectrum(const size_t N) const
	{ return *spectra[N];}

    //! 
    MeasuredSpectrum &Spectrum(const size_t N) 
      { return *spectra[N];}

    //! remove one pointlist
    void RemovePointList(const size_t N);
    void AddPointList(CountedRef<MeasuredPointList> mp) {pointLists.push_back(mp);};
    
    const string& SNPath() const {return snpath;};

    std::vector<std::string> BandNames() const;

    double Zhelio() const {return z_heliocentric;}
    double Zcmb() const {return z_cmb;}
    
    double MWEBV() const { return mwEbv;}

    std::string Name() const {return name;}
    
    std::string CCD() const {return ccd;}

    std::string Field() const {return field;}
    
    
    void Print (const int PrintLevel = 1) const;

    //void KeepOnlyPointsBetween(double observer_datemin,double observer_datemax) const; 

};

ostream &operator << (ostream &, const std::vector<std::string>);



#endif /* MEASUREDDATA__H */
