#include <iomanip>
#include <cmath>
#include <algorithm> // for swap

#include <cstring> // for strcmp

#include "param.h"
#include "fatalerror.h"
#include "snfitexception.h"

#ifndef SHOULD_HAVE_COVARIANCES
//#define SHOULD_HAVE_COVARIANCES
#endif

void LcParam::dump(ostream &stream) const
{
  stream << setprecision(12) << Name() << ' ' << val << ' ' << err << ' ';
  if (isfixed) stream << "F ";
}


//pour atteindre un param const (contenu ds) de lcParam
const LcParam* FitParams::LocateParam(const string &ParamName) const
{
  ParamCIterator i = begin();
  for ( ; i != end(); ++i)
    if ((*i)->Name() == ParamName) break;
  if (i == end()) return NULL;
  return (*i);
}

bool FitParams::DeleteParamIfExists(const string &ParamName)
{
  ParamIterator i = begin();
  for ( ; i != end(); ++i)
    if ((*i)->Name() == ParamName) {
      erase(i);
      return true;
    }
  return false;
}


//pour atteindre un param (contenu ds) de lcParam
LcParam* FitParams::LocateParam(const string &ParamName)
{
  ParamIterator i = begin();
  for ( ; i != end(); ++i)
    if ((*i)->Name() == ParamName) break;
  if (i == end()) return NULL;
  return (*i);
}

//fait un vecteur de class LcParam
const LcParam * FitParams::LocateOrAddParam(const string &ParamName)
{
  const LcParam* i = LocateParam(ParamName);
  if (i != NULL) return i;
  // no there : add one
  push_back(new LcParam(ParamName));
  return LocateParam(ParamName);
  //  ParamCIterator last = end();
  // return (--last); // i.e. the one just inserted (since end() is beyond last)
}

LcParam * FitParams::LocateOrAddNonConstParam(const string &ParamName)
{
  LcParam* i = LocateParam(ParamName);
  if (i != NULL) return i;
  // no there : add one
  push_back(new LcParam(ParamName));
  return LocateParam(ParamName);
  //  ParamCIterator last = end();
  // return (--last); // i.e. the one just inserted (since end() is beyond last)
}


size_t FitParams::VariableParamCount() const
{
  size_t npars = 0;
  // count and locate actually variable parameters
  for (ParamCIterator i=begin(); i != end(); ++i)
    if (!((*i)->isfixed)) npars++;
  return npars;
}


//descend ds le vecteur de class LcParam pr atteindre un param de LcParam

typedef map<string,double>::iterator CovIterator;
typedef map<string,double>::const_iterator CovCIterator;

void FitParams::dump(ostream &stream, const string &Message) const
{
  stream << "BEGIN_OF_FITPARAMS " << name << endl;
  if (Message != "") stream << Message << std::endl;
  for (ParamCIterator i=begin(); i != end(); ++i)
    stream << (*(*i)) << endl;
  for (CovCIterator i = covMat.begin(); i != covMat.end(); ++i)
    stream << "Cov"+i->first << " " << i->second << " -1 " << endl;
  stream << "END_OF_FITPARAMS " << name << endl;
}

void FitParams::dump_model(ostream &stream, const string &Message) const
{
  if (Message != "") stream << Message << std::endl;
  for(size_t i=0; i< this->size(); ++i)
    {
      stream  << setprecision(12) << (*this)[i]->name << " " <<(*this)[i]->val << endl  ;
    }
}
void FitParams::dump_correlation(ostream &stream, const string &Message) const
{ 
  stream << setfill(' ');
  stream  << setiosflags(ios::fixed) ;
  int oldprec = stream.precision();
  stream << setprecision(6);
  if (Message != "") stream << Message << std::endl;
  stream  << setw(12) << "-" ;
  for (size_t i = 0 ; i < this->size(); i++) 
    if ( ! ((*this)[i]->isfixed) )
      stream  <<  setw(12) << (*this)[i]->name ; 
  stream << endl ;
  for(size_t i=0; i< this->size(); ++i)
    {
      if ( ! ((*this)[i]->isfixed) )
	{
	  stream  << setw(12) << (*this)[i]->name ;
	  for (size_t j=0; j <i; ++j)
	    if ( ! ((*this)[j]->isfixed) )
	      stream  << setw(12) << "-" ;
	  for (size_t j=i; j < size() ; ++j)
	    if ( ! ((*this)[j]->isfixed) )
	      {
		double covij =  GetCovariance((*this)[i]->name, (*this)[j]->name );
		double covii =  GetCovariance((*this)[i]->name, (*this)[i]->name );
		double covjj =  GetCovariance((*this)[j]->name, (*this)[j]->name );
		stream << setw(12) <<covij /sqrt(covii*covjj) << " " ;
	      }
	  stream  << endl ;
	}

    }
  stream << setprecision(oldprec);
}



double FitParams::GetCovariance(const string &Name1, const string &Name2) const
{

  /* DEBUGGING
     cout << "===============" << endl;
     for(CovCIterator i = covMat.begin();i!=covMat.end();++i) {     
     cout <<i->first << " " << i->second << endl;
     }
     cout << "===============" << endl;
  */
  const string *n1 = &Name1;
  const string *n2 = &Name2;
  if (Name1 > Name2) std::swap(n1,n2);
  CovCIterator i = covMat.find((*n1)+(*n2));
  if (i != covMat.end()) return i->second;
  else
    {
#ifdef SHOULD_HAVE_COVARIANCES
      FatalError(" requested Covariance For params ('"+*n1+"','"+*n2+"') which was never set", true);
      return -1;
#else 
      if (Name1 == Name2) // return diagonal terms.
	{
	  const LcParam*p = LocateParam(Name1);
	  if (p) return (p->err*p->err);
	  throw(SnfitException(" FitParams::GetCovariance For param '"+Name1+"' and this param does not exist"));
	}
      else /* check for bad names before returning 0*/
	{
	  if (LocateParam(Name1) && LocateParam(Name2)) return 0; 	  
	  throw(SnfitException(" requested Covariance For params ('"+Name1+"','"+Name2+
			       "') and one does not exist"));
	  return 0; // never reached!
	}
#endif
    }
}


void FitParams::SetCovariance(const string &Name1, const string &Name2, const double &Val)
{
  const string *n1 = &Name1;
  const string *n2 = &Name2;
  if (Name1 > Name2) std::swap(n1,n2);

  // some sanity checks
  const LcParam *p1 = LocateParam(Name1);
  if (!p1) throw(SnfitException("SetCovariance : don't have any param named "+Name1));
  const LcParam *p2 = LocateParam(Name2);
  if (!p2) throw(SnfitException("SetCovariance : don't have any param named "+Name2));

  // do not allow unphysyical covariance
  double maxval = p1->err*p2->err;
  if(maxval>0 && fabs(Val)>maxval) {
    covMat[(*n1)+(*n2)]= Val*(maxval/fabs(Val));
  }else{
    covMat[(*n1)+(*n2)]= Val;
  }
}


#include <fstream>

//permet de recuperer les diff LcParam ds un vector de LcParam
void FitParams::read(const std::string &FileName)
{
  ifstream f(FileName.c_str());
  if (!f) throw(SnfitException("FitParams::read : cannot find \""+FileName+"\""));
  try {
    read(f);
  }
  catch (SnfitException e)
    {
      throw(SnfitException(e.message()+" in file \""+FileName+"\""));
    }
  f.close();
}
void FitParams::read(istream &S)
{  
  char line [4096];
  while (S.getline(line,4096))
    {
      
      char read_name[100];
      int status = sscanf(line,"BEGIN_OF_FITPARAMS %s",read_name);
      //cout << line << " " << status << endl;
      if (status != 1) continue;
      //cout << "'" << read_name << "'  '" << name << "'" << endl;
      if (name =="NoName") name = string(read_name);
      if(strcmp(name.c_str(),read_name)==0) break;
    }
  if(S.eof()) {
    throw(SnfitException("FitParams::read : cannot find fitparams called \""+name+"\""));
  }
  while (S.getline(line,4096))
    {
      if (strstr(line,"END_OF_FITPARAMS")!=0) break;
      char name_read[100];
      double val, err;
      int nchar;
      if (sscanf(line,"%s %lf %lf%n",name_read,&val,&err,&nchar)!=3)
	throw(SnfitException("FitParams::read : cannot decode\n"+string(line)+"\n"));
      string name(name_read);
      if (name.find("Cov") == 0) // this is a covariance
	{
	  name.erase(0,3);// drop "Cov"
	  covMat[name] = val;
	  continue;
	}
      LcParam *p=LocateOrAddNonConstParam(name);
      p->val = val;
      p->err = err;
      char fixed[16]; fixed[0] = '\0';
      if (sscanf(line+nchar,"%s",fixed) == 1 || fixed[0] == 'F') p->isfixed = true;
    }
}


void FitParams::operator = (const FitParams &other) {

  if (name == "NoName" ) name = other.name;

  // first clear the parameters in this instance that are not in the other
  FitParams::iterator i=begin();
  while(i != end()) {
    if ( ! other.LocateParam((*i)->name)) {
      delete *i; // free memory (delete parameters)
      i = erase(i);
    }else{
      ++i;
    }
  }
  
  
  for (FitParams::const_iterator i=other.begin(); i != other.end(); ++i) {
    LcParam *p = LocateOrAddNonConstParam((*i)->name); // get a pointer to the parameter of this instance
    *p = *(*i); // copy the content of parameters
  }
  
  // copy covariance matrix
  covMat = other.covMat;
}

FitParams::FitParams(const FitParams &other) :
  vector<LcParam*>()
{
  (*this) = other;
}

#include <cstdio>

vector<string> FitParamNames(const string &FileName)
{
  vector<string> names;
  FILE* f = fopen(FileName.c_str(), "r");
  if (!f) throw(SnfitException(" FitParamNames : cannot open \""+FileName+"\""));
  char line[4096];
  while (fgets(line,4096,f))
    {
      char name[100];
      if (sscanf(line,"BEGIN_OF_FITPARAMS %s",name) == 1)
	names.push_back(name);
    }
  return names;
}
  
