#include <iostream>
#include <sstream>

#include "salt2extinction.h"
#include "snfitexception.h"

#define BSPLINE_ORDER 3
#define PHASE_STEP 0.5
#define WAVELENGTH_STEP 10

using namespace std;


Salt2ExtinctionLaw::Salt2ExtinctionLaw() {
  constant =  log(10.0)/2.5;
  version  = -1; // must be set before usage
  min_reduced_lambda = 1000;
  max_reduced_lambda = 1.e6;
}

void Salt2ExtinctionLaw::update() {
  if (version==1) {
    Pval_min_reduced_lambda = polynomial_value_reduced_lambda(min_reduced_lambda);
    dPval_min_reduced_lambda = alpha;
    double lp = min_reduced_lambda;
    for(size_t i =0; i< Params.size(); i++, lp *= min_reduced_lambda) {
      dPval_min_reduced_lambda += (i+2)*Params(i)*lp;
    }
    
    Pval_max_reduced_lambda = polynomial_value_reduced_lambda(max_reduced_lambda);
    dPval_max_reduced_lambda = alpha;
    lp = max_reduced_lambda;
    for(size_t i =0; i< Params.size(); i++, lp *= max_reduced_lambda) {
      dPval_max_reduced_lambda += (i+2)*Params(i)*lp;
    }
  }
}

void Salt2ExtinctionLaw::SetParams(const Vect& i_params) {
  Params = i_params;
  if(version==1) {
    // do something here
    //cout << "computing alpha" << endl;
    alpha = 1;
    for(size_t i =0; i< Params.size(); i++) {
      // alpha -= (i+2)*Params(i);
      alpha -= Params(i);
    }
    update();
  }
  //never use the basis parameters actually
  //used only to compute the H vectors
  //if(version==2) {
  //basis.SetCoefficients(0.5*Vect(Pass.transposed()*Params));
  //}
}

void Salt2ExtinctionLaw::SetMinLambda(const double& i_min_lambda) {
  min_lambda = i_min_lambda;
  min_reduced_lambda = reduced_lambda(min_lambda);
  if (version==1) update();

}

void Salt2ExtinctionLaw::SetMaxLambda(const double& i_max_lambda) {
  max_lambda = i_max_lambda;
  max_reduced_lambda = reduced_lambda(max_lambda);
  if (version==1) update();

}

void Salt2ExtinctionLaw::Read(istream& stream) {

  bool debug = false;

  Vect p;
  p.readASCII(stream);
  
  string a_string;
  int    an_int;
  double a_double;
  version = 0;
  while(stream >> a_string >> an_int) {
    if(a_string=="Salt2ExtinctionLaw.version") {
      version = an_int;
      break;
    }
  }
  
  
  if(debug) cout << "INFO Salt2ExtinctionLaw::Read version = " << version << endl;

  switch(version) {

  case 0 : SetParams(p); break;
  case 1 : {
    
    SetParams(p);
    
    min_lambda = -1;
    while(stream >> a_string >> a_double) {
      if(a_string=="Salt2ExtinctionLaw.min_lambda") {
	SetMinLambda(a_double);
	break;
      }
    }
    if(min_lambda<0) {
      cerr << "fatal, could not read min_lambda in stream" << endl;
      abort();
    }
    max_lambda = -1;
    while(stream >> a_string >> a_double) {
      if(a_string=="Salt2ExtinctionLaw.max_lambda") {
	SetMaxLambda(a_double);
	break;
      }
    }
    if(max_lambda<0) {
      cerr << "fatal, could not read max_lambda in stream" << endl;
      abort();
    }
    
    if(debug) {
      cout << "INFO Salt2ExtinctionLaw::Read min_lambda = " << min_lambda << endl;
      cout << "INFO Salt2ExtinctionLaw::Read max_lambda = " << max_lambda << endl;
    }
  } break;
  
  default :
    stringstream message;
    message << "ERROR Salt2ExtinctionLaw::Read unknown version " << version;
    cerr << message << endl;
    throw(SnfitException(message.str()));
    
  }
  
}


void Salt2ExtinctionLaw::Write(ostream& stream) const {
  Params.writeASCII(stream);
  // do something else here
  stream << "Salt2ExtinctionLaw.version " << version << endl;
  if(version==1) {
    stream << "Salt2ExtinctionLaw.min_lambda " << GetMinLambda() << endl;
    stream << "Salt2ExtinctionLaw.max_lambda " << GetMaxLambda() << endl;
  }
  
}


/*
  ========================================================
  VERSION 0
  ========================================================
  ext = exp( color * constant * ( l + params(0)*l^2 + params(1)*l^3 + ... ) / ( 1 + params(0) + params(1) + ... ) )
      = exp( color * constant *  numerator / denominator )
      = exp( color * expo_term )
  ======================================================== 
*/


double Salt2ExtinctionLaw::get_expo_term(const double& lambda, double& numerator, double& denominator) const {
  numerator = lambda;
  denominator = 1;
  
  const size_t n = Params.Size();
  if(n>0) {
    double lambda_i = lambda*lambda;
    const double *p = Params.Data();
    for(size_t i=0;i<n;i++,p++,lambda_i*=lambda) {
      numerator += lambda_i*(*p);
      denominator += (*p);
    }
  }
  // cout << "expo_term(" << lambda << ")=" << constant*numerator/denominator << endl;
  return constant*numerator/denominator;
}

double Salt2ExtinctionLaw::extinction_reduced_lambda_version_0(const double& color, const double& lambda)  const{
  double numerator,denominator;
  return exp(color*get_expo_term(lambda,numerator,denominator));
} 

double Salt2ExtinctionLaw::derivative_extinction_reduced_lambda_d_color_version_0(const double& color, const double& lambda) const {
  double numerator,denominator;
  double expo_term = get_expo_term(lambda,numerator,denominator);
  return expo_term*exp(expo_term*color);
}

double Salt2ExtinctionLaw::derivative_extinction_reduced_lambda_d_param_version_0(const double& color, const double& lambda, int parameter_index) const {
  double numerator,denominator;
  double expo_term = get_expo_term(lambda,numerator,denominator);
  return constant*color
    *(
      pow(lambda,parameter_index+2)*denominator
      -numerator
      )/(denominator*denominator)
    *exp(expo_term*color);
}

double Salt2ExtinctionLaw::relative_derivative_extinction_reduced_lambda_d_color_version_0(const double& lambda) const {
  double numerator,denominator;
  return get_expo_term(lambda,numerator,denominator);
}

double Salt2ExtinctionLaw::relative_derivative_extinction_reduced_lambda_d_param_version_0(const double& color, const double& lambda, int parameter_index) const {
  double numerator,denominator;
  get_expo_term(lambda,numerator,denominator);
  return constant*color
    *(
      pow(lambda,parameter_index+2)*denominator
      -numerator
      )/(denominator*denominator); 
}
Vect Salt2ExtinctionLaw::relative_derivative_extinction_reduced_lambda_d_params_version_0(const double& color, const double& lambda) const {
  double numerator,denominator;
  get_expo_term(lambda,numerator,denominator);
  Vect res(Params.size());
  double *resv=res.NonConstData();
  double a = constant*color/denominator;
  double b = constant*color*numerator/(denominator*denominator); 
  double li = lambda*lambda;
  for(size_t i=0;i<res.size();i++,resv++,li*=lambda) {
    *resv = a*li-b;
  }
  return res;
  /*
    constant*color*(
    pow(lambda,i+2)*denominator
    -numerator
    )/(denominator*denominator); 
  */
} 

double Salt2ExtinctionLaw::extinction_lambda_error_version_0(const double& color, const double& lambda, const Mat& covmat) const {
  
  size_t size= Params.Size();
  // assume covmat is covariance matrix of Params
  if( (covmat.SizeX() != size) || covmat.SizeY() != size) {
    cerr << "Fatal error in extinction_lambda_error" << endl;
    cerr << Params << endl;
    cerr << covmat << endl;
    abort();
  }
  Vect relative_derivative(size);
  double rl = reduced_lambda(lambda);
  for(size_t i=0;i<size;i++)
    relative_derivative(i) = relative_derivative_extinction_reduced_lambda_d_param(color,rl,i);
  double var = scalar_product(relative_derivative,covmat,relative_derivative);
  if(var<0) {
    cerr << "WARNING error in extinction_lambda_error, negative variance = " << var << endl;
    return 0;
  }
  return sqrt(var)*extinction_reduced_lambda(color,rl);
}


/*
  ========================================================
  VERSION 1
  ========================================================
  if(l_B<=l<=l_R)
  ext = exp( color * constant * ( alpha*l + params(0)*l^2 + params(1)*l^3 + ... ))
      = exp( color * constant * P(l) )
  alpha = 1-params(0)-params(1)-...
  if(l>l_R)
  ext = exp( color * constant * ( P(l_R) + P'(l_R)*(l-l_R) ) )
  if(l<l_B)
  ext = exp( color * constant * ( P(l_B) + P'(l_B)*(l-l_B) ) )
  
  ======================================================== 
*/

// P(l)
double Salt2ExtinctionLaw::polynomial_value_reduced_lambda(const double& lambda) const {
  
  if(lambda>=min_reduced_lambda && lambda<=max_reduced_lambda) {
    double v        = alpha*lambda;
    const double *p = Params.Data();
    double lp       = lambda*lambda;
    
    for(size_t i =0; i< Params.size(); i++, lp *= lambda, p++) {
      v += (*p)*lp;
    }
    return v;
  }
  if(lambda<min_reduced_lambda)
    return Pval_min_reduced_lambda + dPval_min_reduced_lambda * (lambda-min_reduced_lambda);
      
  return Pval_max_reduced_lambda + dPval_max_reduced_lambda * (lambda-max_reduced_lambda);

}


double Salt2ExtinctionLaw::extinction_reduced_lambda_version_1(const double& color, const double& lambda) const {
  return exp(color*constant*polynomial_value_reduced_lambda(lambda));
} 

double Salt2ExtinctionLaw::derivative_extinction_reduced_lambda_d_color_version_1(const double& color, const double& lambda) const {
  double P = polynomial_value_reduced_lambda(lambda);
  return constant*P*exp(color*constant*P);
}

double Salt2ExtinctionLaw::derivative_extinction_reduced_lambda_d_param_version_1(const double& color, const double& lambda, int parameter_index) const {
  return relative_derivative_extinction_reduced_lambda_d_param_version_1(color,lambda,parameter_index)*extinction_reduced_lambda_version_1(color,lambda);
}

double Salt2ExtinctionLaw::relative_derivative_extinction_reduced_lambda_d_color_version_1(const double& lambda) const {
  return constant*polynomial_value_reduced_lambda(lambda);
}

double Salt2ExtinctionLaw::relative_derivative_extinction_reduced_lambda_d_param_version_1(const double& color, const double& lambda, int parameter_index) const {
  double dP=0;
  if(lambda>=min_reduced_lambda && lambda<=max_reduced_lambda) {
    dP = -lambda+pow(lambda,2.+parameter_index);
  }else{

    if(lambda<min_reduced_lambda) {
      
      dP =  -min_reduced_lambda+pow(min_reduced_lambda,2.+parameter_index)
	+ ( 
	   (-1+(2.+parameter_index)*pow(min_reduced_lambda,1.+parameter_index))
	   *(lambda-min_reduced_lambda)
	   );
      

    }else{

      dP =  -max_reduced_lambda+pow(max_reduced_lambda,2.+parameter_index)
	+ ( 
	   (-1+(2.+parameter_index)*pow(max_reduced_lambda,1.+parameter_index))
	   *(lambda-max_reduced_lambda)
	   );
    }
  }
  return color*constant*dP;
}

Vect Salt2ExtinctionLaw::relative_derivative_extinction_reduced_lambda_d_params_version_1(const double& color, const double& lambda) const {
  Vect der(Params.size());
  if(lambda>=min_reduced_lambda && lambda<=max_reduced_lambda) {
    double lp2 = lambda*lambda;
    for(size_t i=0;i<Params.size();i++,lp2*=lambda) {
      der(i)=-lambda+lp2;
    }
  }else{
    if(lambda<min_reduced_lambda) {
      double lp1 = min_reduced_lambda;
      double lp2 = min_reduced_lambda*min_reduced_lambda;
      for(size_t i=0;i<Params.size();i++,lp1*=min_reduced_lambda,lp2*=min_reduced_lambda) {     
	der(i)= -min_reduced_lambda+lp2
	  + (-1+(2.+i)*lp1)*(lambda-min_reduced_lambda);
      }
    }else{
      double lp1 = max_reduced_lambda;
      double lp2 = max_reduced_lambda*max_reduced_lambda;
      for(size_t i=0;i<Params.size();i++,lp1*=max_reduced_lambda,lp2*=max_reduced_lambda) {     
	der(i)= -max_reduced_lambda+lp2
	  + (-1+(2.+i)*lp1)*(lambda-max_reduced_lambda);
      }
    }
  }
  der *= (color*constant);
  
  return der;
}

double Salt2ExtinctionLaw::extinction_lambda_error_version_1(const double& color, const double& lambda, const Mat& covmat) const {
  
  size_t size= Params.Size();
  // assume covmat is covariance matrix of Params
  if( (covmat.SizeX() != size) || covmat.SizeY() != size) {
    cerr << "Fatal error in extinction_lambda_error" << endl;
    cerr << Params << endl;
    cerr << covmat << endl;
    abort();
  }
  Vect relative_derivative = relative_derivative_extinction_lambda_d_params(color,lambda);
  double var = scalar_product(relative_derivative,covmat,relative_derivative);
  if(var<0) {
    cerr << "WARNING error in extinction_lambda_error, negative variance = " << var << endl;
    return 0;
  }
  return sqrt(var)*extinction_lambda(color,lambda);
  
}

