#ifndef SALT2EXTINCTION
#define SALT2EXTINCTION

#include <math.h>
#include <iostream>
#include <fstream>
#include <string>


#include "wavelength.h"
#include "matvect.h"

#define BUFFERED_H_VECTORS


class Salt2ExtinctionLaw {
  protected :
    double constant; // log(10)/2.5
  Vect Params; 
  
  
  // for version 1
  double alpha;
  double Pval_min_reduced_lambda;
  double dPval_min_reduced_lambda;
  double min_lambda;
  double min_reduced_lambda;

  double Pval_max_reduced_lambda;
  double dPval_max_reduced_lambda;
  double max_lambda;
  double max_reduced_lambda;
  
  // for version 0
  double get_expo_term(const double& lambda, double& numerator, double& denominator) const;
  double extinction_reduced_lambda_version_0(const double& color, const double& lambda) const;
  double derivative_extinction_reduced_lambda_d_color_version_0(const double& color, const double& lambda) const;
  double derivative_extinction_reduced_lambda_d_param_version_0
    (const double& color, const double& lambda, int parameter_index) const;
  double relative_derivative_extinction_reduced_lambda_d_color_version_0(const double& lambda) const;
  double relative_derivative_extinction_reduced_lambda_d_param_version_0
    (const double& color, const double& lambda, int parameter_index) const;
  Vect relative_derivative_extinction_reduced_lambda_d_params_version_0
    (const double& color, const double& lambda) const;
  double extinction_lambda_error_version_0(const double& color, const double& lambda, const Mat& covmat) const;
  // for version 1
  void update(); 
  double polynomial_value_reduced_lambda(const double & lambda) const;
  double extinction_reduced_lambda_version_1(const double& color, const double& lambda) const;
  double derivative_extinction_reduced_lambda_d_color_version_1(const double& color, const double& lambda) const;
  double derivative_extinction_reduced_lambda_d_param_version_1
    (const double& color, const double& lambda, int parameter_index) const;
  double relative_derivative_extinction_reduced_lambda_d_color_version_1(const double& lambda) const;
  double relative_derivative_extinction_reduced_lambda_d_param_version_1
    (const double& color, const double& lambda, int parameter_index) const;
  Vect relative_derivative_extinction_reduced_lambda_d_params_version_1
    (const double& color, const double& lambda) const;
  double extinction_lambda_error_version_1(const double& color, const double& lambda, const Mat& covmat) const;
  
  public :
  /* 
     Parameters for extinction law 

     extinction = exp [ color*log(10)*0.4/(1.+ext(0)+ext(1)+...)*(lr+lr^2*ext(0)+lr^3*ext(1)+...) ]
     
     where 
     lr = (lambda - lambda_B)/(lambda_V - lambda_V)
     lr(lambda_B)=0
     lr(lambda_V)=1
     E(B-V)  = color (as in SALT1)
     
  */
    const Vect& GetParams() const { return Params;}
  size_t NumberOfParams() const { return Params.size();} 
  void SetParams(const Vect& i_params);
  const double& GetMaxLambda() const { return max_lambda;}
  const double& GetMinLambda() const { return min_lambda;}
  void SetMaxLambda(const double& i_max_lambda);
  void SetMinLambda(const double& i_min_lambda);
  
  void Read(std::istream& stream);
  void Read(const std::string &filename) {
    std::ifstream is(filename.c_str());
    Read(is);
    is.close();
  };
  void Write(std::ostream &stream) const;
  void Write(const std::string &filename) const {
    std::ofstream os(filename.c_str());
    Write(os);
    os.close();
  }
  friend std::ostream& operator << (std::ostream &stream, const Salt2ExtinctionLaw &f) {
    f.Write(stream); 
    return stream;
  }
  friend std::istream& operator >> (std::istream &stream, Salt2ExtinctionLaw &f) {
    f.Read(stream); 
    return stream;
  }
  
  
  int version; // 0= as in paper, 1=new
  
  
  

  Salt2ExtinctionLaw();
  
  double extinction_reduced_lambda(const double& color, const double& lambda) const {
    switch(version) {
    case 0 :
      return extinction_reduced_lambda_version_0(color,lambda);
    case 1 :
      return extinction_reduced_lambda_version_1(color,lambda);
    default :
      std::cerr << "fatal error, cannot deal with version = " << version << std::endl;
      abort();
    }
  }
  double derivative_extinction_reduced_lambda_d_color(const double& color, const double& lambda) const {
    switch(version) {
    case 0 :
      return derivative_extinction_reduced_lambda_d_color_version_0(color,lambda);
    case 1 :
      return derivative_extinction_reduced_lambda_d_color_version_1(color,lambda);
      
    default :
      std::cerr << "fatal error, cannot deal with version = " << version << std::endl;
      abort();
    }
  }
  double derivative_extinction_reduced_lambda_d_param
    (const double& color, const double& lambda, int parameter_index) const {
    switch(version) {
    case 0 :
      return derivative_extinction_reduced_lambda_d_param_version_0(color,lambda,parameter_index);
    case 1 :
      return derivative_extinction_reduced_lambda_d_param_version_1(color,lambda,parameter_index);
 
      
    default :
      std::cerr << "fatal error, cannot deal with version = " << version << std::endl;
      abort();
    }

  }
  
  double relative_derivative_extinction_reduced_lambda_d_color(const double& lambda) const {
    switch(version) {
    case 0 :
      return relative_derivative_extinction_reduced_lambda_d_color_version_0(lambda);
    case 1 :
      return relative_derivative_extinction_reduced_lambda_d_color_version_1(lambda);

    default :
      std::cerr << "fatal error, cannot deal with version = " << version << std::endl;
      abort();
    }
  }
  double relative_derivative_extinction_reduced_lambda_d_param
    (const double& color, const double& lambda, int parameter_index) const {
    switch(version) {
    case 0 :
      return relative_derivative_extinction_reduced_lambda_d_param_version_0(color,lambda,parameter_index);
    case 1 :
      return relative_derivative_extinction_reduced_lambda_d_param_version_1(color,lambda,parameter_index);

    default :
      std::cerr << "fatal error, cannot deal with version = " << version << std::endl;
      abort();
    }
  }
  Vect relative_derivative_extinction_reduced_lambda_d_params
    (const double& color, const double& lambda) const {
    switch(version) {
    case 0 :
      return relative_derivative_extinction_reduced_lambda_d_params_version_0(color,lambda);
    case 1 :
      return relative_derivative_extinction_reduced_lambda_d_params_version_1(color,lambda);

    default :
      std::cerr << "fatal error, cannot deal with version = " << version << std::endl;
      abort();
    }
  }
  
  double extinction_lambda_error(const double& color, const double& lambda, const Mat& covmat) const {
     switch(version) {
     case 0 :
       return extinction_lambda_error_version_0(color,lambda,covmat);
     case 1 :   
       return extinction_lambda_error_version_1(color,lambda,covmat);
       
    default :
      std::cerr << "fatal error, cannot deal with version = " << version << std::endl;
      abort();
    }
  }

  
  double reduced_lambda(const double& lambda) const {
    if(version==2) return lambda;
    return (lambda-B_WAVELENGTH)/(V_WAVELENGTH-B_WAVELENGTH);
  }
  double extinction_lambda(const double& color, const double& lambda) const {
    return extinction_reduced_lambda(color,reduced_lambda(lambda));
  }
  double derivative_extinction_lambda_d_color(const double& color, const double& lambda) const {
    return derivative_extinction_reduced_lambda_d_color(color,reduced_lambda(lambda));
  }
  double derivative_extinction_lambda_d_param(const double& color, const double& lambda, int parameter_index) const {
    return derivative_extinction_reduced_lambda_d_param(color,reduced_lambda(lambda),parameter_index);
  }
  double relative_derivative_extinction_lambda_d_color(const double& lambda) const {
    return relative_derivative_extinction_reduced_lambda_d_color(reduced_lambda(lambda));
  }
  double relative_derivative_extinction_lambda_d_param(const double& color, const double& lambda, int parameter_index) const {
    return relative_derivative_extinction_reduced_lambda_d_param(color,reduced_lambda(lambda),parameter_index);
  }
  Vect relative_derivative_extinction_lambda_d_params(const double& color, const double& lambda) const {
    return relative_derivative_extinction_reduced_lambda_d_params(color,reduced_lambda(lambda));
  }
  
  

};
#endif

