// -*- C++ -*-
// 
// 
// 
// 
#include <iostream>
#include <string>
#include <fstream>
#include <algorithm>
#include <cstdio>

#include "pathutils.h"
#include "fileutils.h"
#include "vutils.h"
#include "dictfile.h"


#include "integflux.h"
#include "magsys.h"

#include "snfitexception.h"
#include "hostabsorption.h"
#include "measureddata.h"
#include "salt2model.h"
#include "hostabsorption.h"
#include "wavelength.h"
#include "galaxytemplate.h"

// uncomment this to include corrections in the integration over lambda
//#define WAVELENGTH_DEPENDENT_CORRECTIONS

// step for integration over wavelength in the product of
// template spectrum and corrections
#define LAMBDA_STEP 5


#ifdef ALLOW_WARPING
#define WARPING_WAVELENGTH_0 4400.
#define WARPING_WAVELENGTH_1 5900.
#endif

//==============================================================================

static double sqr(const double& x) {return x*x;}

Salt2Model::Salt2Model() 
{
#ifdef DEBUG
  cout << " > Salt2Model::Salt2Model() " << endl;
#endif
  params.SetName("Salt2Model");
  forcepositiveEBV = false;
  ReferenceWavelength = B_WAVELENGTH; // central wavelength of the Johnson B filter
  with_lc_error_scaling = true;
  with_spec_error_scaling = true;
  
  BlueCutoff = 3000; // 
  RedCutoff  = 8000; // 

  
  
  // create basic parameters
  pDayMax  = params.LocateOrAddNonConstParam("DayMax");  
  pDayMax->val = 0.; // range for fit set in Salt2Model::InitParams
  pDayMax->step = 1;

  pRedshift = params.LocateOrAddNonConstParam("Redshift");
  pRedshift->val = 0.05;
  pRedshift->minv = -0.1;
  pRedshift->maxv = 2.;
  pRedshift->step = 0.2;

#ifdef WITH_HOST_EXTINCTION
  pRv  = params.LocateOrAddNonConstParam("Rv");
  pRv->val  = 3.1;
  pRv->minv = 0.5;
  pRv->maxv = 6.;
  pRv->step = 0.1;
  
  pTau = params.LocateOrAddNonConstParam("Tau");
  pTau->val = 0.;
  pTau->minv = -2;
  pTau->maxv = 2;
  pTau->step = 0.1;
#endif

#ifdef WITH_UV_FIT
  pUV  = params.LocateOrAddNonConstParam("UV");
  pUV->val  = 0;
  pUV->minv = -1;
  pUV->maxv = 1;
  pUV->step = 0.1;
#endif

  pColor = params.LocateOrAddNonConstParam("Color");
  pColor->val = 0.;
  pColor->minv = -4;
  pColor->maxv = 4;
  pColor->step = 0.1;
  
  
  string salt2_data_dir = LocateFileFromCard("SALT2");

  
  // read the components  
  ReadComponents(salt2_data_dir);
  
  // read color law
  {
    string filename = salt2_data_dir+"/salt2_color_correction.dat";
    if(! FileExists(filename)) {
      cerr << "cannot find " << filename << endl;
      exit(-12);
    }
    extinction_law.Read(filename);
  }
  
  if(F.size()==0) {
    cerr << "Fatal error in Salt2Model::Salt2Model" << endl;
    cerr << "No component was read" << endl;
    exit(-12);
  }

  char pname[10];
  for(size_t c=0;c<F.size();c++) {
    sprintf(pname,"X%d",int(c));
    LcParam *p = params.LocateOrAddNonConstParam(pname);
    if(c==0)
      p->val=1;
    else
      p->val=0;
    pX.push_back(p);
  }

  galaxy_template = 0;
  pGalaxyContam = 0;
  pOnlyGalaxyNorm = 0;
  pGalaxyType = 0;

#ifdef ALLOW_WARPING 
  warping_constant_0 = 1./sqr(U_WAVELENGTH-WARPING_WAVELENGTH_0);
  warping_constant_1 = 1./sqr(R_WAVELENGTH-WARPING_WAVELENGTH_1);
#endif
}



void Salt2Model::FitGalaxySpectrum(const MeasuredSpectrum* spec) {

  bool debug = true;
  
  if(debug)
    cout << "Salt2Model::FitGalaxySpectrum" << endl;
  
  // create template
  if(!galaxy_template)
    galaxy_template = new GalaxyTemplate();
  
  // add parameters
  bool already_defined = (params.LocateParam("GalaxyType")!=0);
  pGalaxyType  = params.LocateOrAddNonConstParam("GalaxyType");  
  if(!already_defined) {
    
    pGalaxyType->step = 1;
    pGalaxyType->minv = galaxy_template->MinType();
    pGalaxyType->maxv = galaxy_template->MaxType();
    pGalaxyType->val = (pGalaxyType->minv+pGalaxyType->maxv)/2.;
  }
  
  already_defined = (params.LocateParam("GalaxyCont")!=0);
  pGalaxyContam  = params.LocateOrAddNonConstParam("GalaxyCont");  
  if(!already_defined) {
    pGalaxyContam->val = 1.e-6;
    pGalaxyContam->step = 1;
    pGalaxyContam->minv = -1.e4;
    pGalaxyContam->maxv = 1.e4;

    if(spec) { // use spec to set a default value 
      double sumModelFluxes = 0;
      double sumDataFluxes = 0;
      pGalaxyContam->val = 1;
      for (MeasuredSpectrum::const_iterator iti = spec->begin(); iti!= spec->end(); ++iti)
	{
	  const MeasuredSpectrumPoint *data= *iti;
	  sumDataFluxes += data->Flux();
	  sumModelFluxes += GalaxySpectrumFlux(data->Wavelength());
	}
      double ratio = 1;
      if(sumModelFluxes>0 ) ratio = sumDataFluxes/sumModelFluxes;
      pGalaxyContam->val = ratio*0.1;
      pGalaxyContam->maxv = ratio*20;
      pGalaxyContam->minv = -ratio*20;
      pGalaxyContam->step = pGalaxyContam->val;
    
      if(spec->Galaxy()) {
	already_defined = (params.LocateParam("OnlyGalaxy")!=0);
	pOnlyGalaxyNorm  = params.LocateOrAddNonConstParam("OnlyGalaxy");  
	if(!already_defined) {
	  double sumDataFluxes = 0;
	  for (MeasuredSpectrum::const_iterator iti = spec->Galaxy()->begin(); iti!= spec->Galaxy()->end(); ++iti)
	    {
	      const MeasuredSpectrumPoint *data= *iti;
	      sumDataFluxes += data->Flux();
	    }
	  if(sumModelFluxes>0 ) 
	    pOnlyGalaxyNorm->val =  sumDataFluxes/sumModelFluxes;
	  else
	    pOnlyGalaxyNorm->val = 1;
	  pOnlyGalaxyNorm->minv = 0;
	  pOnlyGalaxyNorm->maxv = pOnlyGalaxyNorm->val*10;
	  pOnlyGalaxyNorm->step = pOnlyGalaxyNorm->val/2.;
	}
      }
    }
  }
}

bool  readtemplate(Grid2DFunction& func, const string& filename, const string& data_dir,bool is_fatal=true) {
  char fullfilename[500];
  sprintf(fullfilename,"%s/%s",data_dir.c_str(),filename.c_str());
  if(FileExists(fullfilename)) {
    func.Read(fullfilename);
  } else {
    if(is_fatal) {
      cout << "ERROR Salt2Model readtemplate :  cannot find " << fullfilename << endl;
      throw(SnfitException(" Salt2Model readtemplate :  cannot find "+string(fullfilename)));
    }else{
      cout << "WARNING Salt2Model readtemplate : cannot find " << fullfilename << endl;
      return false;
    }
  }
  return true;
}

int Salt2Model::ReadComponents(const string& data_dir) {
  
  
  char filename[500];
  
  // read components (and count them)
  size_t c=0;
  while(true) {
    sprintf(filename,"%s/salt2_template_%d.dat",data_dir.c_str(),int(c));
    if(!FileExists(filename))
      break;
    cout  << "INFO Salt2Model::ReadComponents Reading component " << c << endl;
    Grid2DFunction Fc;
    Fc.Read(filename);
    F.push_back(Fc);
    c++;
  }
  

  readtemplate(photometric_relative_variance_0,
	       "salt2_lc_relative_variance_0.dat",data_dir);
  readtemplate(spectroscopic_variance_0,
	       "salt2_spec_variance_0.dat",data_dir);


  if(F.size()>=1) {
    readtemplate(photometric_relative_variance_1,
		 "salt2_lc_relative_variance_1.dat",data_dir);
    readtemplate(photometric_relative_covariance_01,
		 "salt2_lc_relative_covariance_01.dat",data_dir);

    readtemplate(spectroscopic_variance_1,
		 "salt2_spec_variance_1.dat",data_dir);
    readtemplate(spectroscopic_covariance_01,
		 "salt2_spec_covariance_01.dat",data_dir);
  }

  


  with_lc_error_scaling = readtemplate(photometric_errors_scaling,"salt2_lc_dispersion_scaling.dat",data_dir,false);
  
  with_spec_error_scaling = readtemplate(spectroscopic_errors_scaling,"salt2_spec_dispersion_scaling.dat",data_dir,false);
 
#ifndef NEW_ERRORS 
  // Color dispersion
  sprintf(filename,"%s/salt2_color_dispersion.dat",data_dir.c_str());
  if(FileExists(filename)) {
    Ksigma = General1DFunction(filename);

    /* old stuff
    DictFile dfile(filename);
    if(! dfile.HasGlobalKey("kU")) {
      cerr << "FATAL error in SALT2: no key kU in " << filename << endl;
      exit(12);
    }
    if(! dfile.HasGlobalKey("kR")) {
      cerr << "FATAL error in SALT2: no key kR in " << filename << endl;
      exit(12);
    }
    KVect.allocate(2);
    KVect(0)=atof(dfile.GlobalValue("kU").c_str());
    KVect(1)=atof(dfile.GlobalValue("kR").c_str());
    // for debugging today
    cout << "kU,kR = " << KVect(0) << "," << KVect(1) << endl;
    */
    

  } else {
    cerr << "FATAL error in SALT2: cannot find " << filename << endl;
    exit(12);
  }
  
#else
  sprintf(filename,"%s/salt2_covariance_model.dat",data_dir.c_str());
  if(FileExists(filename)) {
    covariance_model.Read(filename);
  }else{
    cerr << "FATAL error in SALT2: cannot find " << filename << endl;
    exit(12);
  }

#endif
  return 0;
}


/* 
   reads parameters in a file and set default values if parameter names 
   match those already in memory.
   first check is this parameter set correspond to the program used for fit  
*/
void Salt2Model::ReadParams(const string& filename) {
  params.read(filename);


#ifdef ALLOW_WARPING
  pW.clear();
#endif  
    
  // 
  for(size_t i=0; i<params.size();++i) {
    const string &pname = params[i]->Name();
    
    LcParam* p = params[i];
    
    // check if it is a calibration parameter of the form
    //  C%d_hashcode(spectrumname)
    
    int calib_index;
    char cst_specname[1000];
    if(sscanf(pname.c_str(),"c%d_%s",&calib_index,cst_specname)==2) {
      string specname=cst_specname;
      if ( calibparams.find(specname) == calibparams.end()) {
	vector<LcParam*> spec_calibparams;
	calibparams[specname]=spec_calibparams;
      }
      
      vector<LcParam*>& spec_calibparams = calibparams.find(specname)->second;
      if(int(spec_calibparams.size())!=calib_index) {
	cerr << "io pb when reading parameter " << pname << endl;
	continue;
      }
      cout << "Adding calib parameter for spectrum " << specname << endl;
      spec_calibparams.push_back(p);

      continue;
      
    }
    if(pname=="GalaxyCont" || pname=="GalaxyType" || pname=="OnlyGalaxy"  ) {
      FitGalaxySpectrum(); // i can fit the galaxy
      if(pname=="OnlyGalaxy")
	pOnlyGalaxyNorm  = p;
      
      continue;
    }
#ifdef ALLOW_WARPING
    int w_index;
    if(sscanf(pname.c_str(),"W%d",&w_index)==1) {
      if( w_index != int(pW.size())) {
	cerr << "fatal error in Salt2Model::ReadParams, expect to red W0,W1... in inc. order" << endl;
	exit(12);
      }
      pW.push_back(p);
    }

#endif

#ifdef DEBUG    
      cout << "read: " << *p << endl;
#endif
  }

  Update();
}   
  
vector<LcParam*>* Salt2Model::GetCalibParams(const string& spectrum_name) {
  if ( calibparams.find(spectrum_name) == calibparams.end()) { // need a new set of calib params.
    return 0;
  }
  return & (calibparams.find(spectrum_name)->second);
}

const vector<LcParam*>* Salt2Model::GetCalibParams(const string& spectrum_name) const {
  
  if ( calibparams.find(spectrum_name) == calibparams.end()) { // need a new set of calib params.
    return 0;
  }
  return  & (calibparams.find(spectrum_name)->second);
}


bool Salt2Model::CanComputeFluxFor(const Filter &observerfilter,bool quiet) const 
{
  
  double min_restframewavelength = observerfilter.Xmin()/(1.+pRedshift->val);
  //  double max_restframewavelength = observerfilter.Xmax()/(1.+pRedshift->val);
  double mean_restframewavelength = observerfilter.Mean()/(1.+pRedshift->val);
  
  /* Desactivate the following.
     this basically means we assume flux=0 outside the range of the template
  if(min_restframewavelength<F[0].Ymin()) {
    if(!quiet)
      cout << "I won't fit " << observerfilter.Band() << " cause min restframewavelength = " << min_restframewavelength << " < " << F[0].Ymin() << " (limit of the model)" << endl;
    return false;
  }
  if(max_restframewavelength>F[0].Ymax()) {
    if(!quiet)
      cout << "I won't fit " << observerfilter.Band() << " cause max restframewavelength = " << max_restframewavelength << " > " << F[0].Ymax() << " (limit of the model)" << endl;
    return false;
  }
  */
    
  if(mean_restframewavelength>BlueCutoff && mean_restframewavelength<RedCutoff) // limitation due to the nearby data 
    return true;
  else {
    if(!quiet) {
      if(min_restframewavelength<BlueCutoff)
	cout << "INFO Salt2Model::CanComputeFluxFor : I won't fit " << observerfilter.Band() << " cause restframewavelength = " << mean_restframewavelength << " < " << BlueCutoff << " (limit in U)" << endl;
      else
	cout << "INFO Salt2Model::CanComputeFluxFor : I won't fit " << observerfilter.Band() << " cause restframewavelength = " << mean_restframewavelength << " > " << RedCutoff << " (limit in R)" << endl;
    }
    return false;
  }
}


void Salt2Model::Update() {
  
}

double Salt2Model::RestFrameFlux(const double lambda, 
				 const double observerday) const
{
  double phase = GetPhase(observerday);
  double flux = 0;
  for(size_t c = 0; c<F.size(); c++) {
    if(c==0) {
      flux += F[c].Value(phase,lambda);
    }else{
      flux += (pX[c]->val)*F[c].Value(phase,lambda);
    }
  }
  flux *= extinction_law.extinction_lambda(pColor->val,lambda);
  flux *= ( (1+pRedshift->val)*(pX[0]->val)*1.e-12 );
  
  
  // optional modifications

#ifdef WITH_HOST_EXTINCTION
  // compute extinction law
  double data[]={pTau->val,pRv->val}; // B-V excess and Rv 
  flux *= FluxExtinctionCardelli(lambda,data); // check this
#endif
#ifdef WITH_UV_FIT
  if(lambda< B_WAVELENGTH) {
    flux *= exp(pUV->val*(lambda-B_WAVELENGTH)/(U_WAVELENGTH-B_WAVELENGTH));
  }
#endif 

#ifdef ALLOW_WARPING
  //double lambda = MeasuredPassBand->Mean()/(1+pRedshift->val);
  
  if(pW.size()>0 && (lambda<WARPING_WAVELENGTH_0)) {
    flux *=(1.+pW[0]->val*warping_constant_0*sqr(lambda-WARPING_WAVELENGTH_0));
  }
  if(pW.size()>1 && (lambda>WARPING_WAVELENGTH_1)) {
    flux *=(1.+pW[1]->val*warping_constant_1*sqr(lambda-WARPING_WAVELENGTH_1));
  }
#endif
  return flux;
}

class ModelAndDate {
public :
  const Salt2Model& model;
  const double& date;
  ModelAndDate(const Salt2Model& i_model,const double& i_date) :
    model(i_model),
    date(i_date)
  {
  }
};

static double FluxAtDate(const double &lambda, const void *Data) {

  const Salt2Model& model = ((const ModelAndDate*) Data)->model;
  const double&      date = ((const ModelAndDate*) Data)->date;
  return model.RestFrameFlux(lambda,date);
}


double Salt2Model::Flux(const Filter* MeasuredPassBand,
			const double observerday) const
{
  const ModelAndDate md(*this,observerday);
  Computed1DFunction FluxFunc(&FluxAtDate, (const void*)(&md));
  FluxFunc.SetXmin(F[0].Xmin()); 
  FluxFunc.SetXmax(F[0].Xmax()); 
  
  bool check_range = false; // this basically means we assume flux=0 outside the range of the template
  double flux = IntegFlux(FluxFunc,*MeasuredPassBand, pRedshift->val, check_range);




  return flux;
}


double Salt2Model::SpectrumFlux(const double observerlambda, 
				const double observerday) const
{  
  return RestFrameFlux(observerlambda/(1.+pRedshift->val),observerday);
}


double Salt2Model::SpectrumCalibration(const double observerlambda, 
				       const MeasuredSpectrum& spectrum) {
  
  const vector<LcParam*>* spec_calibparams = GetCalibParams(spectrum.Name());
  
  if( !spec_calibparams) {
    cerr << "WARNING Salt2Model::SpectrumCalibration : cannot find recalibration params for spec '" << spectrum.Name() << "'" << endl;
    return 1;
  }

  
  double lrc = 0;
  if(spec_calibparams->size()==1) {
    lrc = (*spec_calibparams->begin())->val;
  }else{
    double reduced_lambda = (observerlambda/(1.+pRedshift->val)/4400.)-1;
    double rli = 1;
    for(size_t i=0;i<spec_calibparams->size();i++,rli*=reduced_lambda) {
      lrc += (*spec_calibparams)[i]->val*rli;
    }
  }
  
  //double min = spectrum.GetMinCalibrationError();
  //double max = spectrum.GetMaxCalibrationError();
  //double val = min+(0.5+atan(lrc)/3.14159265358979312)*(max-min);

  if(lrc>100) lrc=100;
  if(lrc<-100) lrc=-100;
  return exp(lrc);
}

double Salt2Model::GalaxySpectrumFlux(const double observerlambda, bool only_galaxy) {
  if(!galaxy_template) {
    cerr << "fatal error in GalaxySpectrumFlux, must call FitGalaxySpectrum() before" << endl;
    abort();
  }
  double flux = galaxy_template->Value(observerlambda/(1.+pRedshift->val) , pGalaxyType->val);
  if(only_galaxy) {
    return pOnlyGalaxyNorm->val*flux; // flux de la galaxie seule
  }else{
    return pGalaxyContam->val*flux; // flux de la contamination galactique
  }
}

static double DayWithMaxFlux(const MeasuredPointList &List, double &maxflux) {
  
  // evaluate median error
  double *errors = new double[List.size()];
  int i=0;
  for (MeasuredPointCIterator iti=List.begin(); iti!= List.end(); ++iti, ++i) {
    errors[i] = (*iti)->FluxErr();
  }
  double median_error = DArrayMedian(errors, List.size());
  delete [] errors;
  
  
  double maxDay = 0;
  maxflux = 0;
  for (MeasuredPointCIterator iti=List.begin(); iti!= List.end(); ++iti) {
    const MeasuredPoint *data= *iti;
    if( (data->FluxErr()<3.*median_error) && (data->Flux()>maxflux) )  {
      maxDay = data->Day();
      maxflux = data->Flux();
    }
  }
  return maxDay;
}

void Salt2Model::InitParams(const MeasuredData* data) {
  
#ifdef DEBUG
  cout << " > Salt2Model::InitParams" << endl;
#endif

  
 
  // compute first guess of maxDay;
  double maxDay = 0;
  //  size_t bandwithbestsampling=0;
  if( data->NBands()) {
    //select band with most data
    //size_t nmax=0;
    //for (size_t i=0; i < data->NBands(); ++i) {
    //if(data->PointList(i).size()>nmax) {
    //nmax=data->PointList(i).size();
    //bandwithbestsampling=i;
    //}
    //}
    double maxflux=0;
    for (size_t i=0; i < data->NBands(); ++i) {   
      double maxflux_for_band=0;
      double maxDay_for_band = DayWithMaxFlux(data->PointList(i),maxflux_for_band);
      if(maxflux_for_band>maxflux) {
	maxflux=maxflux_for_band;
	maxDay=maxDay_for_band;
      }
    }
  
  // 
  if(!pDayMax->isfixed) {
     pDayMax->val = maxDay;
    pDayMax->step = 1;
    pDayMax->minv = maxDay-30;
    pDayMax->maxv = maxDay+30;
  }
  }
  
  //
  if(!pX[0]->isfixed) {
    pX[0]->val = 1;
    
    if( data->NBands()) {
      // compute first guess for dflux 
      
      
      double sumModelFluxes = 0;
      double sumDataFluxes = 0;
      for(size_t i=0; i < data->NBands(); ++i) {
	const MeasuredPointList &datapoints = data->PointList(i);
	const Filter* observerfilter = &(datapoints.GetEffectiveFilterWithMilkyWay());
	for (MeasuredPointCIterator iti = datapoints.begin(); iti!= datapoints.end(); ++iti)
	  {
	    const MeasuredPoint *data= *iti;
	    if(ModelRelativeError(observerfilter,data->Day(),true)<0.1) {
	      
	      sumDataFluxes  += data->Flux();
	      sumModelFluxes += Flux(observerfilter,data->Day());
	    }
	  }
      }
      if(sumModelFluxes<=0) {
	cerr << "in Salt2Model::InitParams model errors to large for first guess of X0" << endl;
	sumDataFluxes = 0;
	sumModelFluxes = 1;
      }
      pX[0]->val = sumDataFluxes/sumModelFluxes; 
      pX[0]->step = 0.2 * pX[0]->val;
      //pX[0]->minv = 0.;
      //pX[0]->maxv = 100. * pX[0]->val;
      pX[0]->minv = 0;
      pX[0]->maxv = 10000. * pX[0]->val;
    }
  }
  

  for(size_t c=1;c<pX.size();c++) { 
    pX[c]->step = 0.01;
    pX[c]->minv = -5.;
    pX[c]->maxv = 5.;
    if(!pX[c]->isfixed) {
      pX[c]->val = 0;
    }else{     
      pX[c]->minv += pX[c]->val;
      pX[c]->maxv += pX[c]->val;
    }
    
  }
#ifdef ALLOW_WARPING
  for(size_t c=0;c<pW.size();c++) { 
    pW[c]->val = 0;
    pW[c]->step = 0.1;
    pW[c]->minv = -1000.;
    pW[c]->maxv = 1000.;
  }
#endif

  pColor->step = 0.1;
  pColor->minv = -4;
  pColor->maxv = 4;
    
  if( ! pColor->isfixed) {
    pColor->val = 0.;
  }else{
    pColor->minv += pColor->val;
    pColor->maxv += pColor->val;
  }


  Update();
}

vector<LcParam*> Salt2Model::InitCalibrationParams(const MeasuredSpectrum *spec, int n_recalibration_params) {
  
  
  if ( calibparams.find(spec->Name()) == calibparams.end()) { // need a new set of calib params.
    vector<LcParam*> spec_calibparams;
    for(int i=0;i<n_recalibration_params;i++) {
      char name[100];
      sprintf(name,"c%d_%s",i,spec->Name().c_str());
      spec_calibparams.push_back(params.LocateOrAddNonConstParam(name));
    }
    calibparams[spec->Name()] = spec_calibparams;
  }

  vector<LcParam*> & spec_calibparams = calibparams.find(spec->Name())->second;
  
  for(size_t i=0;i<spec_calibparams.size();i++) {
    LcParam *pi = spec_calibparams[i];
    if(!pi->isfixed) {
      pi->val = 0;
      pi->step = 1;
      pi->minv = -1.e6;
      pi->maxv = 1.e6;
      if(i>=1) {
	pi->minv = -10;
	pi->maxv = 10;
	
      }
    }
  }

  // check if daymax and X0 are initiated
  if(pDayMax->val == 0) {
    pDayMax->val = spec->Date();
  }
  pDayMax->minv = pDayMax->val-100;
  pDayMax->maxv = pDayMax->val+100;
  
  
  double sumModelFluxes = 0;
  double sumDataFluxes = 0;
  for (MeasuredSpectrum::const_iterator iti = spec->begin(); iti!= spec->end(); ++iti)
    {
      const MeasuredSpectrumPoint *data= *iti;
      sumDataFluxes += data->Flux();
      sumModelFluxes += SpectrumFlux(data->Wavelength(),spec->Date());
    }
  if( sumModelFluxes>0)  {
 
    if( ! pX[0]->isfixed)  {
      pX[0]->val = sumDataFluxes/sumModelFluxes;
      pX[0]->step = pX[0]->val *0.5;
      pX[0]->minv = pX[0]->val *0.001;
      pX[0]->maxv = pX[0]->val *1000;
      
    }
    //else{
      //if(spec_calibparams.size()) {
	/// set default val for calib
	//spec_calibparams[0]->val = 0; //log(sumDataFluxes/sumModelFluxes);
	//spec_calibparams[0]->minv = -100;
	//spec_calibparams[0]->maxv = 100;
    // }
    //}
  }
  
  Update();

  return spec_calibparams;
}


#ifdef STORAGE  
void Salt2Model::KeepOnlyNBands(int maxnumber, MeasuredData* data) {
  
  list<int> restwavelengthdiff;
  map<int,int> bandmap;
  map<int,double> restwavelengthmap;
  
  // fill list
  for (size_t i=0; i < data->NBands(); ++i) {
    if (data->PointList(i).isDiscarded()) continue;
    const Filter& observerfilter = data->PointList(i).GetEffectiveFilterWithMilkyWay();
    double wl = observerfilter.CentralWaveLength()/(1+pRedshift->val);
    int iwl   = abs(int(wl-ReferenceWavelength)); 
    restwavelengthdiff.push_back(iwl);
    bandmap[iwl]=i;
    restwavelengthmap[iwl]=wl;
  }
  restwavelengthdiff.sort();
  
  int count = 0;
  cout << "Bands kept close to reference wavelength  (" <<  ReferenceWavelength << "):"<< endl;
  for(list<int>::iterator wl = restwavelengthdiff.begin();  wl != restwavelengthdiff.end(); ++wl) {
    count++;
    int keptband = bandmap.find(*wl)->second;
    cout << keptband << endl;
    if(count>maxnumber) {
      cout << "Discarding band " << data->PointList(keptband).GetEffectiveFilterWithMilkyWay().Band() << " (" << restwavelengthmap.find(*wl)->second << ")" << endl;
      data->PointList(keptband).Discard();
    }else{
      cout << "Keeping band " << data->PointList(keptband).GetEffectiveFilterWithMilkyWay().Band() << " (" << restwavelengthmap.find(*wl)->second << ")" << endl;
    }
  }
}
#endif


double Salt2Model::GetPhase(double observerday) const {
  return (observerday - pDayMax->val)/(1+pRedshift->val);
}



#ifdef STORAGE
double  Salt2Model::GetRestFrameMagAtPhase(const Filter& filter, double phase, const string &magsys) {
#ifdef DEBUG 
  cout << " Salt2Model::GetRestFrameMagAtPhase " << endl;
#endif
  
  double saved_z =  pRedshift->val;
  pRedshift->val=0;
  bool can_compute = CanComputeFluxFor(filter);
  pRedshift->val=saved_z;
  if(!can_compute) return 99;

  
  double day = phase*(1.+pRedshift->val)+pDayMax->val;
  Filter filter_red = filter;
  filter_red.RescaleX(1+pRedshift->val);

  double flux_sn = Flux(&filter_red,day)/(1+pRedshift->val);
  if(flux_sn<=0)
    return 99;
  
  return (-2.5*log10(flux_sn/flux_ref));
}
  
LcParam  Salt2Model::GetMagAtPhase(const Filter& filter, double phase, const string &magsys) {
#ifdef DEBUG 
  cout << " Salt2Model::GetMagAtPhase " << endl;
#endif
  
  LcParam mag;
  char name[100];
  sprintf(name,"Mag_%d_%s",(int)phase,filter.Band().c_str());
  mag.name = name;
  mag.val = 99;
  mag.err = 99;

  if(! CanComputeFluxFor(filter))
    return mag;

  double day = phase*(1.+pRedshift->val)+pDayMax->val;
  double flux_ref = IntegPhotons(RefSpectrumForMags(magsys), filter,0.);
  double flux_sn = Flux(&filter,day);
  if(flux_sn<=0)
    return mag;

  mag.val = -2.5*log10(flux_sn/flux_ref);
  
  // error is just that of dflux
  mag.err = 2.5/log(10.)*pX[0]->err/pX[0]->val;
  
  return mag;
}
#endif /* STORAGE */


#ifdef NEW_ERRORS
double Salt2Model::ModelRelativeCovariance(const Filter* filter,
				       const double observerday1,
				       const double observerday2) {
  double phase1 = GetPhase(observerday1);
  double phase2 = GetPhase(observerday2);
  double lambda = filter->Mean()/(1.+pRedshift->val);
  cout << phase1 << " " << phase2 << " " << lambda << " " << sqrt(covariance_model.Cov(phase1,phase2,lambda)) << endl;
  return covariance_model.Cov(phase1,phase2,lambda);
}
#endif




double Salt2Model::ModelRelativeError(const Filter* filter,
			      const double observerday, bool with_scaling) {
  
  if(F.size()>2) {
    cout << "sorry, Salt2Model::ModelRelativeError is not implemented for more than 1 var. comp" << endl;
    exit(12);
  }

   
  double p = GetPhase(observerday);
  double rl = filter->Mean()/(1.+pRedshift->val);
  double error_scaling = 1;
  if(with_lc_error_scaling && with_scaling) error_scaling = photometric_errors_scaling.Value(p,rl);
  
  double f0 = 0;
  double ftot = 0;
  double model_relative_var = 0; // relative to flux^2 of av. component
  bool check_range = 0;
  {
    Grid2DXSlice S(F[0], p);
    f0 = IntegFlux(S,*filter, pRedshift->val, check_range); // the flux in comp. 0
    ftot = f0;
    model_relative_var = photometric_relative_variance_0.Value(p,rl);
  }
  if(F.size()>=1) {
    Grid2DXSlice S(F[1], p);
    const double& X = pX[1]->val;
    ftot += X*IntegFlux(S,*filter, pRedshift->val, check_range); // the flux in comp. 1
    model_relative_var += sqr(X)*photometric_relative_variance_1.Value(p,rl);
    model_relative_var += 2.*X*photometric_relative_covariance_01.Value(p,rl);
  }
  if(ftot<=0) 
    return 100;
  
  if(model_relative_var<=0) {
    cerr << "Warning : model_relative_var = " << model_relative_var << endl;
    model_relative_var = 0.0001;
  }
  return (f0*sqrt(model_relative_var)/ftot)*error_scaling;
}

#ifndef NEW_ERRORS
double Salt2Model::ModelKError(const Filter* filter) {
  
  double rl = filter->Mean()/(1.+pRedshift->val);
  double val = Ksigma.Value(rl);
  
  // !! for error propto color here !!
  // val *= fabs(pColor->val);
  
  return val;
  /* old stuff
  if(rl<B_WAVELENGTH)
    return KVect(0)*pow((rl-B_WAVELENGTH)/(U_WAVELENGTH-B_WAVELENGTH),3);
  if(rl>V_WAVELENGTH)
    return KVect(1)*pow((rl-V_WAVELENGTH)/(R_WAVELENGTH-V_WAVELENGTH),2);
  return 0;
  */
}
#endif

double Salt2Model::SpectrumModelError(const double observerlambda,  
				      const double observerday,
				      bool with_scaling) {
  
  if(F.size()>2) {
    cout << "sorry, Salt2Model::SpectrumModelError is not implemented for more than 1 var. comp" << endl;
    exit(12);
  }
  
  double p = GetPhase(observerday);
  double rl =observerlambda /(1.+pRedshift->val);
  
  double error_scaling = 1;

  ///if(with_scaling) error_scaling = photometric_errors_scaling.Value(p,rl);
  if(with_spec_error_scaling && with_scaling) error_scaling = spectroscopic_errors_scaling.Value(p,rl);

  double var = spectroscopic_variance_0.Value(p,rl);
  if(F.size()>=1) {
    const double& X = pX[1]->val;
    var += sqr(X)*spectroscopic_variance_1.Value(p,rl);
    var += 2.*X*spectroscopic_covariance_01.Value(p,rl);
  }
  //cout << p << " " << rl << " " << sqrt(var) << " " << error_scaling << endl;

  double transmission = 1;
#ifdef WITH_HOST_EXTINCTION
  // compute extinction law
  double data[]={pTau->val,pRv->val}; // B-V excess and Rv
  Computed1DFunction HostGalacticExtinction(&FluxExtinctionCardelli, data);
  transmission = FluxExtinctionCardelli(rl,data); // check this
#endif 
  
  
  //if(rl<3400) {
  //if(p<-7 || p>+12) error_scaling*=1000; // hack
  //}
  //if(fabs(p)>=10 && rl<3400) error_scaling*=1000; // hack
  

  return (1+pRedshift->val)*(pX[0]->val)*sqrt(var)*transmission*1.e-12*extinction_law.extinction_lambda(pColor->val,rl)*error_scaling;
  
} 



#ifdef ALLOW_WARPING
void Salt2Model::SetNumberOfWarpingParams(size_t n) {
  if(n>2) {
    cout << "fatal : cannot have " << n << " warping coefficients" << endl;
    abort();
  }
  
  cout << "using " << n << " warping coefficients" << endl;
  size_t Wsize = pW.size();
  if(n==Wsize)
    return;
  pW.clear();
  
  char pname[100];
  if(Wsize>n) {
    for(size_t i=n;i<Wsize;i++) {
      sprintf(pname,"W%d",int(i));
      params.DeleteParamIfExists(pname);
    }
  }
  for(size_t i=0;i<n;i++) {
    sprintf(pname,"W%d",int(i));
    pW.push_back(params.LocateOrAddNonConstParam(pname));
  }
}
#endif
