// -*- C++ -*-
// 
// 
// 
// 
#include <iostream>
#include <string>
#include <fstream>
#include <algorithm>
#include <cstdio>

#include "pathutils.h"
#include "fileutils.h"
#include "vutils.h"

#include "snfitexception.h"
#include "hostabsorption.h"
#include "measureddata.h"
#include "saltmodel.h"

#include "wavelength.h"
#include "integflux.h"
#include "magsys.h"

// uncomment this to include corrections in the integration over lambda
//#define WAVELENGTH_DEPENDENT_CORRECTIONS

// step for integration over wavelength in the product of
// template spectrum and corrections
#define LAMBDA_STEP 5

//==============================================================================


SaltModel::SaltModel() 
{
#ifdef DEBUG
  cout << " > SaltModel::SaltModel() " << endl;
#endif
  params.SetName("SaltModel");
  forcepositiveEBV = false;
  ReferenceWavelength = B_WAVELENGTH_SALT1; // central wavelength of the Johnson B filter
  
  BlueCutoff = 3460; // as in Astier05
  RedCutoff  = 6600; // as in Astier05
  

  // create parameters
  pDayMax  = params.LocateOrAddNonConstParam("DayMax");
  pStretch = params.LocateOrAddNonConstParam("Stretch");
  pDflux   = params.LocateOrAddNonConstParam("Dflux");
  pColor   = params.LocateOrAddNonConstParam("Color");
 
#ifdef WITH_HOST_EXTINCTION
  pRv      = params.LocateOrAddNonConstParam("Rv");
  pTau     = params.LocateOrAddNonConstParam("Tau");
#endif
  
  pRedshift = params.LocateOrAddNonConstParam("Redshift");
  
  pDayMax->val = 0.;
  pStretch->val = 1.;
  pDflux->val = 1.;
#ifdef WITH_HOST_EXTINCTION 
  pTau->val = 0.;
  pRv->val = 0.;
#endif 
  pColor->val = 0.;
  pRedshift->val = -12;
  
  string data_dir = LocateFileFromCard("SALT1");
  
  // create and read the spectral template
  string templatefilename = data_dir+"/salt_template.dat";
  if(!FileExists(templatefilename)) {
    cerr << "Cannot find " << templatefilename << " for salt template " << endl;
    cerr << "Quit" << endl;
    exit(-1);
  }

  SpectralTemplate = new SnSpectra(templatefilename); 
  
  // create and read polynomes if exist 
  string corrections_stretch_filename = data_dir+"/corrections_stretch"; 
  string corrections_color_filename = data_dir+"/corrections_color"; 
  if(!FileExists(corrections_stretch_filename)) {
    polynomes = 0;
    cerr << "Cannot find " << corrections_stretch_filename << " for corrections to template " << endl;
    cerr << "Quit" << endl;
    exit(-1);
  }else{
    if(!FileExists(corrections_color_filename)) {
      polynomes = 0;
      cerr << "Cannot find " << corrections_stretch_filename << " for corrections to template " << endl;
      cerr << "Quit" << endl;
      exit(-1);
    }else{
      polynomes = new Corrector(corrections_stretch_filename,corrections_color_filename);
    }
  }
}

#ifdef STORAGE
/* 
   reads parameters in a file and set default values if parameter names 
   match those already in memory.
   first check is this parameter set correspond to the program used for fit  
*/
void SaltModel::ReadParams(const string& filename) {
  ifstream filestream(filename.c_str());
  if(!filestream) {
    cerr << "SaltModelFit::ReadParams error, cannot open file " << filename << endl;
    return;
  }
  // read model
  string classname;
  if(!(filestream >> classname)) {
    cerr << "SaltModelFit::ReadParams error when reading model name " << filename << endl;
    return;
  }
  if(classname!=ModelName()) {
    cerr << "SaltModelFit::ReadParams error model \"" << classname << "\"!=\"" << ModelName() << "\"" << endl;
    return;
  }
  
  // read list of parameters and put them in params_in_file
  FitParams params_in_file;
  params_in_file.read(filestream,false);
  filestream.close();
  
    
  // for those parameters of params_in_file which are in params, set values
  for(size_t i=0; i<params_in_file.size();++i) {
    const string &pname = params_in_file[i]->Name();
    LcParam* p = params.LocateParam(pname);
    if(!p) {
      cerr << "no parameter named " << pname << endl;
      abort();
    }
    p->val = params_in_file[i]->val;
    p->err = params_in_file[i]->err;
#ifdef DEBUG    
      cout << "read: " << *p << endl;
#endif
  } 
}   
#endif  

bool SaltModel::CanComputeFluxFor(const Filter &observerfilter,bool quiet) const 
{
  bool templatesays = SpectralTemplate->CheckRange4Filter(observerfilter,pRedshift->val);
  if(!templatesays)
    return false;

  
  double restframewavelenght = observerfilter.Mean()/(1.+pRedshift->val);
  if(restframewavelenght>BlueCutoff && restframewavelenght<RedCutoff) // limitation due to the nearby data 
    return true;
  else {
    if(!quiet) {
      if(restframewavelenght<BlueCutoff)
	cout << "I won't fit " << observerfilter.Band() << " cause restframewavelenght = " << restframewavelenght << " < " << BlueCutoff << " (limit in U)" << endl;
      else
	cout << "I won't fit " << observerfilter.Band() << " cause restframewavelenght = " << restframewavelenght << " > " << RedCutoff << " (limit in R)" << endl;
    }
    return false;
  }
}


#ifdef WAVELENGTH_DEPENDENT_CORRECTIONS
class AdHocComputed1DFunction : public Sampled1DFunction 
{
 private:
  const Grid2DXSlice* spectrum; 
  const double& phase; 
  const double& stretch; 
  const double& color;
  Corrector* polynomes;
public:
  AdHocComputed1DFunction(const Grid2DXSlice* ispectrum, 
			  Corrector* ipolynomes, 
			  const double& iphase, 
			  const double& istretch, 
			  const double& icolor) :
    spectrum(ispectrum),
    polynomes(ipolynomes),
    phase(iphase),
    stretch(istretch),
    color(icolor) {}

  double Value(const double &lambda) const {
    double parameters[] = {phase,stretch,lambda,color};
    if(polynomes)
      return spectrum->Value(lambda)*pow(10,-0.4* polynomes->Value(parameters) + 0.4*(color*(lambda-B_WAVELENGTH_SALT1)/(V_WAVELENGTH_SALT1-B_WAVELENGTH_SALT1)));
    else
      return spectrum->Value(lambda)*pow(10,0.4*(color*(lambda-B_WAVELENGTH_SALT1)/(V_WAVELENGTH_SALT1-B_WAVELENGTH_SALT1)));
  }
  double Xmin() const { return -1e30;}
  double Xmax() const { return  1e30;}
  void dump(ostream & stream = cout) const { 
    stream << " AdHocComputed1DFunction" <<  endl;
  }
};
#endif

double SaltModel::FluxWithinPhaseRange(const Filter* MeasuredPassBand,
				       const double &Phase,
				       const double &RF_lambda) const
{
  // compute extinction law
#ifdef WITH_HOST_EXTINCTION
  double data[]={pTau->val,pRv->val}; // B-V excess and Rv
  Computed1DFunction HostGalacticExtinction(&FluxExtinctionCardelli, data);
#endif
  // load spectrum for the given phase
  Grid2DXSlice snSpectrum(*SpectralTemplate, Phase);
  
#ifdef WAVELENGTH_DEPENDENT_CORRECTIONS // compute corrections for each step of integration over lambda
  // ===================================================================================================

  AdHocComputed1DFunction correctedspectrum(&snSpectrum,polynomes,Phase,pStretch->val,pColor->val);
  
  //return pDflux->val*(1+pRedshift->val)*IntegPhotons(correctedspectrum, *MeasuredPassBand, HostGalacticExtinction, pRedshift->val);
  return pDflux->val*(1+pRedshift->val)*IntegPhotons(correctedspectrum, *MeasuredPassBand, pRedshift->val);

#else // compute corrections on mean wavelength of the filter (as in 1st paper submission)
  // ===================================================================================================
  double correction = 1;
  // compute correction if polynomes exist
  if(polynomes) {
    double parameters[]={Phase,pStretch->val,RF_lambda,pColor->val};
    // hack to remove divergence of corrections in an untrained parameter space area 
    //if(Phase<-10) {
    //  cout << Phase << " : hack to remove divergence of corrections in an untrained parameter space area " << endl;
    //  parameters[0]=-10;
    //}
    correction = pow(10,-0.4* polynomes->Value(parameters));
    
  }
  double colorcorrection = pow(10,0.4*(pColor->val*(RF_lambda-B_WAVELENGTH_SALT1)/(V_WAVELENGTH_SALT1-B_WAVELENGTH_SALT1))); 

#ifdef WITH_HOST_EXTINCTION
  double expectedflux = pDflux->val*(1+pRedshift->val)*IntegFlux(snSpectrum, *MeasuredPassBand, HostGalacticExtinction, pRedshift->val)*correction*colorcorrection;
#else
  double expectedflux = pDflux->val*(1+pRedshift->val)*IntegFlux(snSpectrum, *MeasuredPassBand, pRedshift->val)*correction*colorcorrection;
#endif
  return expectedflux;
#endif
  // ===================================================================================================
}

double SaltModel::SpectrumFluxWithinPhaseRange(const double &Phase,
					       const double &RF_lambda) const
{
  
  double correction = 1;
  if(polynomes) {
    double parameters[]={Phase,pStretch->val,RF_lambda,pColor->val};
    correction = pow(10,-0.4* polynomes->Value(parameters));  
  }
  
  double colorcorrection = pow(10,0.4*(pColor->val*(RF_lambda-B_WAVELENGTH_SALT1)/(V_WAVELENGTH_SALT1-B_WAVELENGTH_SALT1)));
  
  double template_value = SpectralTemplate->Value(Phase,RF_lambda);
  
  //double data[]={pTau->val,pRv->val}; // B-V excess and Rv
  //double host_extinction_value = FluxExtinctionCardelli(RF_lambda,data);
  
  //cout << template_value << " " << host_extinction_value << " " << correction << " " << colorcorrection << endl;
  
  double expectedflux = pDflux->val
    *(1+pRedshift->val)
    *template_value
    // *host_extinction_value
    *correction
    *colorcorrection;
  return expectedflux;
}




double SaltModel::SpectrumFlux(const double observerlambda, 
			      const double observerday) const
{
  double RF_lambda = observerlambda/(1+pRedshift->val);  
  double phase = GetPhase(observerday);
  
  if (phase < -20) return 0;
  double expectedflux;
  if (phase < -19) // i.e. -20<= phase < -19
    {
      // make the function continuous
      double fluxm19 =  SpectrumFluxWithinPhaseRange(-19,RF_lambda);
      expectedflux = (phase + 20)*fluxm19;
    }

  else if (phase > 69) 
    {
      /* extrapolate from available data : returning 0 causes a
	 discontinuity of chi2 w.r.t. stretch */
      double ph1 = 60;
      double ph2 = 68;
      double fl1 =  SpectrumFluxWithinPhaseRange(ph1,RF_lambda);
      double fl2 = SpectrumFluxWithinPhaseRange(ph2,RF_lambda);

      if (fl2 < 0 || fl1 <0)
	throw(SnfitException(" SaltModel::Flux  : negative flux"));
      if (fl2 == 0 || fl1 == 0)
	return 0;
      
      expectedflux = (phase-ph1)*(log(fl2)-log(fl1))/(ph2-ph1)+log(fl1);
      expectedflux = exp(expectedflux);
    }
  else
    {
      expectedflux = SpectrumFluxWithinPhaseRange(phase,RF_lambda);
    }
      
  return expectedflux; 
}




double SaltModel::Flux(const Filter* MeasuredPassBand,
			const double observerday) const
{
  double restframe_lambda_measuredpassband = MeasuredPassBand->Mean()/(1+pRedshift->val);  
  double phase = GetPhase(observerday);
  
  if (phase < -20) return 0;
  double expectedflux;

  if (phase < -19) // i.e. -20<= phase < -19
    {
      // make the function continuous
      double fluxm19 =  FluxWithinPhaseRange(MeasuredPassBand, -19,
					     restframe_lambda_measuredpassband);
      expectedflux = (phase + 20)*fluxm19;
    }

  else if (phase > 69) 
    {
      /* extrapolate from available data : returning 0 causes a
	 discontinuity of chi2 w.r.t. stretch */
      double ph1 = 60;
      double ph2 = 68;
      double fl1 =  FluxWithinPhaseRange(MeasuredPassBand, ph1,
			      restframe_lambda_measuredpassband);
      double fl2 = FluxWithinPhaseRange(MeasuredPassBand, ph2,
			      restframe_lambda_measuredpassband);
      if (fl2 < 0 || fl1 <0)
	throw(SnfitException(" SaltModel::Flux  : negative flux"));
      if (fl2 == 0 || fl1 == 0)
	return 0;
      
      expectedflux = (phase-ph1)*(log(fl2)-log(fl1))/(ph2-ph1)+log(fl1);
      expectedflux = exp(expectedflux);
    }
  else
    {
      expectedflux = FluxWithinPhaseRange(MeasuredPassBand, phase,
					  restframe_lambda_measuredpassband);
    }
      
  return expectedflux; 
}



static double DayWithMaxFlux(const MeasuredPointList &List) {
  
  // evaluate median error
  double *errors = new double[List.size()];
  int i=0;
  for (MeasuredPointCIterator iti=List.begin(); iti!= List.end(); ++iti, ++i) {
    errors[i] = (*iti)->FluxErr();
  }
  double median_error = DArrayMedian(errors, List.size());
  delete [] errors;
  
  
  double maxDay = 0;
  double flux =0;
  for (MeasuredPointCIterator iti=List.begin(); iti!= List.end(); ++iti) {
    const MeasuredPoint *data= *iti;
    if( (data->FluxErr()<3.*median_error) && (data->Flux()>flux) )  {
      maxDay = data->Day();
      flux = data->Flux();
    }
  }
  return maxDay;
}

void SaltModel::InitParams(const MeasuredData* data) {
  
#ifdef DEBUG
  cout << " > SaltModel::InitParams" << endl;
#endif

  
 
  // compute first guess of maxDay;
  
  //select band with most data
  size_t nmax=0;
  size_t bandwithbestsampling=0;
  for (size_t i=0; i < data->NBands(); ++i) {
    if(data->PointList(i).size()>nmax) {
      nmax=data->PointList(i).size();
      bandwithbestsampling=i;
    }
  }
    
  double maxDay = DayWithMaxFlux(data->PointList(bandwithbestsampling));
  
  cout << "maxDay = " << maxDay << endl;

  // 
  if(!pDayMax->isfixed) {
    pDayMax->val = maxDay;
    pDayMax->step = 1;
    pDayMax->minv = maxDay-30;
    pDayMax->maxv = maxDay+30;
  }
  
  // 
  if(!pStretch->isfixed) {
    pStretch->val = 1;
    pStretch->step = 0.1;
    pStretch->minv = 0.4;
    pStretch->maxv = 1.4;
  }
  
  

  // 
  if(!pColor->isfixed) {
    pColor->val = 0.;
    pColor->step = 0.05;
    pColor->minv = -1;
    pColor->maxv = 1;
  }
  
  // 
#ifdef WITH_HOST_EXTINCTION
  if(!pRv->isfixed) {
    pRv->val = 3.1;
    pRv->step = 0.1;
    pRv->minv = 0.5;
    pRv->maxv = 6.;
  }
  if(!pTau->isfixed) { 
    pTau->val = 0;
    pTau->step = 0.1;
    pTau->minv = -2;
    pTau->maxv = 2;
    if(forcepositiveEBV)
      pTau->minv = 0;
  }
#endif
  //
  if(!pDflux->isfixed) {
    pDflux->val = 1;
    
    // compute first guess for dflux 
    const MeasuredPointList &datapoints = data->PointList(bandwithbestsampling);
    const Filter* observerfilter = &(datapoints.GetEffectiveFilterWithMilkyWay());
    
    double sumModelFluxes = 0;
    double sumDataFluxes = 0;
    double mf;
    for (MeasuredPointCIterator iti = datapoints.begin(); iti!= datapoints.end(); ++iti)
      {
      const MeasuredPoint *data= *iti;
      sumDataFluxes += data->Flux();
      mf = Flux(observerfilter,data->Day());
      if(isnan(mf)) {
	cerr << "flux=" << mf << " in SaltModel::InitParams, quit" << endl;
	cerr << "for day = " << data->Day() << " (phase = " << GetPhase(data->Day()) << ") and filter " << observerfilter->Band() << endl;
	exit(-12);
      }
      sumModelFluxes += mf;
      }
    if(sumModelFluxes<=0 || isnan(sumModelFluxes)) {
      cerr << "sumModelFluxes=" << sumModelFluxes << " in SaltModel::InitParams, quit" << endl;
      exit(-12);
    }
    pDflux->val = sumDataFluxes/sumModelFluxes; 
    pDflux->step = 0.2 * pDflux->val;
    pDflux->minv = 0.;
    pDflux->maxv = 100. * pDflux->val;
  }

  //
  if(!pRedshift->isfixed) {
    if(pRedshift->val<0)
      pRedshift->val = 0.6;
    pRedshift->step = 0.2;
    pRedshift->minv = 0.001;
    pRedshift->maxv = 2.;
  }
  
}


#ifdef STORAGE
void SaltModel::KeepOnlyNBands(int maxnumber, MeasuredData* data) {
  
  list<int> restwavelengthdiff;
  map<int,int> bandmap;
  map<int,double> restwavelengthmap;
  
  // fill list
  for (size_t i=0; i < data->NBands(); ++i) {
    if (data->PointList(i).isDiscarded()) continue;
    const Filter& observerfilter = data->PointList(i).GetEffectiveFilterWithMilkyWay();
    double wl = observerfilter.CentralWaveLength()/(1+pRedshift->val);
    int iwl   = abs(int(wl-(ReferenceWavelength+250))); // make sure to keep B+V instread of U+B at z=0
    restwavelengthdiff.push_back(iwl);
    bandmap[iwl]=i;
    restwavelengthmap[iwl]=wl;
  }
  restwavelengthdiff.sort();
  
  int count = 0;
  cout << "Bands kept close to reference wavelength  (" <<  ReferenceWavelength << "):"<< endl;
  for(list<int>::iterator wl = restwavelengthdiff.begin();  wl != restwavelengthdiff.end(); ++wl) {
    count++;
    int keptband = bandmap.find(*wl)->second;
    cout << keptband << endl;
    if(count>maxnumber) {
      cout << "Discarding band " << data->PointList(keptband).GetEffectiveFilterWithMilkyWay().Band() << " (" << restwavelengthmap.find(*wl)->second << ")" << endl;
      data->PointList(keptband).Discard();
    }else{
      cout << "Keeping band " << data->PointList(keptband).GetEffectiveFilterWithMilkyWay().Band() << " (" << restwavelengthmap.find(*wl)->second << ")" << endl;
    }
  }
}
#endif

double SaltModel::GetPhase(double observerday) const {
  
  return (observerday - pDayMax->val)/(1+pRedshift->val)/pStretch->val;
}



#ifdef STORAGE
double  SaltModel::GetRestFrameMagAtPhase(const Filter& filter, double phase, const string &magsys) {
#ifdef DEBUG 
  cout << " SaltModel::GetRestFrameMagAtPhase " << endl;
#endif
  
  
  double day = phase*(1.+pRedshift->val)+pDayMax->val;
  Filter filter_red = filter;
  filter_red.RescaleX(1+pRedshift->val);
  double zp = GetMagSys(magsys)->ZeroPoint(filter);
  double flux_sn = Flux(&filter_red,day)/(1+pRedshift->val);
  if(flux_sn<=0)
    return 99;
  else
    return (-2.5*log10(flux_sn)+zp);
}
  
LcParam  SaltModel::GetMagAtPhase(const Filter& filter, double phase, const string &magsys) {
#ifdef DEBUG 
  cout << " SaltModel::GetMagAtPhase " << endl;
#endif
  
  LcParam mag;
  char name[100];
  sprintf(name,"Mag_%d_%s",(int)phase,filter.Band().c_str());
  mag.name = name;
  mag.val = 99;
  mag.err = 99;
  
  if(! CanComputeFluxFor(filter))
    return mag;
  
  
  double day = phase*(1.+pRedshift->val)+pDayMax->val;
  double zp = GetMagSys(magsys)->ZeroPoint(filter);
  double flux_sn = Flux(&filter,day);
  if(flux_sn<=0)
    mag.val = 99;
  else
    mag.val = -2.5*log10(flux_sn)+zp;
  
  // error is just that of dflux
  mag.err = 2.5/log(10.)*pDflux->err/pDflux->val;
  
  
  return mag;
}
#endif /* STORAGE */

/********************************** SnSpectra ********************/

bool SaltModel::SnSpectra::CheckRange4Filter(const Filter& filter,double z) const
{
  Grid2DXSlice snSpectrum(*this, 0);

  if(filter.Xmin()<snSpectrum.Xmin()*(1.+z)) {
    cout << "in SnSpectra::CheckRange4Filter: Filter " << filter.Band() 
	 << " lambda_min (" << filter.Xmin() << ")" 
	 << " < Spectrum lambda_min*(1+z) (" << snSpectrum.Xmin()*(1.+z) << ")"  << endl;
    return false;
  } 
  if(filter.Xmax()>snSpectrum.Xmax()*(1.+z)) {
    cout << "in SnSpectra::CheckRange4Filter: Filter " << filter.Band() 
	 << " lambda_max  (" << filter.Xmax() << ")" 
	 << " > Spectrum lambda_max*(1+z) ("  << snSpectrum.Xmax()*(1.+z) << ")"   << endl;
    return false;
  } 
  return true;
}

