#include <string>
#include <vector>
#include <fstream>

#include "fileutils.h"
#include "fullfit.h"
#include "measureddata.h"
#include "pathutils.h"
#include "instrument.h"
#include "saltmodel.h"
#include "salt2model.h"
#include "snfitexception.h"

using namespace std;

double phasemin=-15;
double phasemax=+45;
double salt1_default_wmin = 3460;
double salt1_default_wmax = 6600;
double salt2_default_wmin = 3000;
double salt2_default_wmax = 7000;
double uband_calibration_error = 0.1;
map< string,vector<string> > fixableparameters;


#define WITHEXCEPTIONS

//#define WITH_SPECTRA

static void usage() {
  cerr << "snfit lightcurve1 lightcurve2 ..." << endl;
  cerr << "    -o filename              : write output in filename instead of result_salt2.dat"<< endl;
  cerr << "    -m {salt,salt2}          : model (default is salt2)"<< endl;
  cerr << endl;
  cerr << "options for data selection :"<< endl;
  cerr << "    -r phasemin phasemax     : range of phases  (default is " << phasemin << "," << phasemax << ")"<< endl;
  cerr << "    -w wmin wmax             : wavelength range (default SALT1:"
       << salt1_default_wmin << "," << salt1_default_wmax
       << " SALT2:" 
       << salt2_default_wmin << "," << salt2_default_wmax << ")" << endl;
  cerr << endl;
  cerr << "options for model parameters (need to define the model before): "<< endl;
  cerr << "(for salt :";
  for(size_t i=0;i<fixableparameters["salt"].size();i++) 
    cerr << " " << fixableparameters["salt"][i];
  cerr << ", for salt2 : ";
  for(size_t i=0;i<fixableparameters["salt2"].size();i++) 
    cerr << " " << fixableparameters["salt2"][i];
  cerr << ")" << endl;
  
  cerr << "    -f name value            : fix the value of parameter 'name'"<< endl;
  cerr << "    -p name mean sigma       : add a gaussian prior for 'name'"<< endl;
  //cerr << "    -z                       : fit redshift"<< endl;
  cerr << "    -i name value            : initial value of parameter 'name'"<< endl;
  cerr << endl;
  cerr << "options for assumed uncertainties : "<< endl;
  cerr << "    -e (0 or 1)              : (diagonal) model uncertainties in chi2 (0=no,1=yes,default=1, for salt2)"<< endl;
  cerr << "    -k (0 or 1)              : k-correction uncertainties in chi2  (0=no,1=yes,default=1,for salt2)"<< endl;
  cerr << "    -U value                 : add observed U-band calibration error (default is " << uband_calibration_error << ")"<< endl;
  cerr << "    -E  value                : error floor to apply for nearby SNe (frac. of max flux)"<< endl;
  //cerr << "    -M                       : run minos at the end"<< endl;
  //cerr << "    -P value                 : phase weighting value=fraction of max flux at +20 days"<< endl;
  //cerr << "    -u                       : update weights in fit"<< endl;
  
#ifdef ALLOW_WARPING
  cerr << "    -W                       : allow warping (for salt2 only)"<< endl;
#endif


#ifdef WITH_SPECTRA
  cerr << "    -s spectrum_file_name    : fit this spectrum as well"<< endl;
  cerr << "    -c #                     : number of calib. params. for spectra (default=1, for salt2)"<< endl;  
#endif

  cerr << endl;
  cerr << "See http://supernovae.in2p3.fr/~guy/salt/" << endl;
  cerr << endl;
  

  //cerr << "    -{1,2,3}                 : 1,2 or 3 observed band fitted only"<< endl;
  
  //cerr << "    -g                       : fit a galaxy+SN spectrum (default is false)" << endl;
  //cerr << "Debugging options (expert mode): " << endl;
  //cerr << "    -S <band> <wavelenghtshift> : shift the effective filter of a given band by <wavelenghtshift> " << endl;
  exit(EXIT_FAILURE); 
}


void check_param(const string& name , const vector<string>& allowed) {
 
  for(size_t p=0;p<allowed.size();++p) {
    if (name==allowed[p]) return;
  }
  cerr << "cannot deal with unknown parameter " << name << endl;
  usage();
}


#define NBANDS_FOR_SINGLE_FIT 1


#define _GNU_SOURCE 1
#ifndef __USE_GNU
#define __USE_GNU
#endif
#include <fenv.h>



int main(int nargs, char **args)
{
  
  // defaults
  // ------------------------------------------------------
  vector<string> lc;
  vector<string> spectra;
  string modelname = "salt2";
  int nobservedband = 0;
  int n_calibration_parameters = 0;
  float forced_wmin = -100;
  float forced_wmax = -200;
  bool with_galaxy = false;
  bool use_model_errors = true;
  bool use_kcorr_errors = true;
  float error_floor_to_apply = 0;
  bool minos = false;
  vector<FullFitPrior> priors;
  double phase_weight = 0;
  bool update_weight_in_fit = false;
#ifdef ALLOW_WARPING
  bool warping = false;
#endif  
  string resname = "";
  // to crash when NaN
  //feenableexcept (FE_INVALID|FE_DIVBYZERO|FE_OVERFLOW);
  
  {
    vector<string> fixableparameters_salt1;
    fixableparameters_salt1.push_back("DayMax");
    fixableparameters_salt1.push_back("Dflux");
    fixableparameters_salt1.push_back("Stretch");
    fixableparameters_salt1.push_back("Color");
    fixableparameters_salt1.push_back("Redshift");
    fixableparameters["salt"]=fixableparameters_salt1;
    
    vector<string> fixableparameters_salt2;
    fixableparameters_salt2.push_back("DayMax");
    fixableparameters_salt2.push_back("Color");
#ifdef WITH_HOST_EXTINCTION    
    fixableparameters_salt2.push_back("Rv");
    fixableparameters_salt2.push_back("Tau");
#endif
    fixableparameters_salt2.push_back("X0");
    fixableparameters_salt2.push_back("X1");
    fixableparameters_salt2.push_back("Redshift");
    fixableparameters_salt2.push_back("GalaxyType");
    fixableparameters["salt2"]=fixableparameters_salt2;
  }
  
  string band2shift = "none";
  double wavelenghtshift = 0;
  
  map<string,double> fixed_parameter_values;
  map<string,double> initial_parameter_values;
  
  vector<string> known_models;  
  known_models.push_back("salt");
  known_models.push_back("salt2");
  

  // read arguments
  // ------------------------------------------------------
  if (nargs <2) usage();
  for (int i=1; i<nargs; ++i)
    {
      char *arg = args[i];
      if (arg[0] != '-')
	{
	  lc.push_back(arg);continue;
	}
      switch (arg[1])
	{
	case 's' : spectra.push_back(args[++i]); break;
	case 'm' : 
	  {
	    modelname = args[++i]; 
	    bool ok = false;
	    for(size_t p=0;p<known_models.size();p++) {
	      ok = known_models[p]==modelname;
	      if(ok) break;
	    }
	    if(!ok) usage();
	    break;
	  }
	case 'p' : 
	  {
	    string parname = args[++i];
	    double val   = atof(args[++i]);
	    double err   = atof(args[++i]);
	    check_param(parname,fixableparameters[modelname]);
	    cout << "setting prior for parameter \"" <<  parname << "\" : " << val << " +- " << err << endl;
	    priors.push_back(FullFitPrior(parname,val,err));
	    break;
	  }
	case 'r' : phasemin=atof(args[++i]); phasemax=atof(args[++i]); break; 
	case 'w' : forced_wmin = atof(args[++i]); forced_wmax = atof(args[++i]); break;   
	case 'P' : phase_weight = atof(args[++i]); break;  
	case 'u' : update_weight_in_fit = true; break;
	case 'o' : resname = args[++i]; break;
#ifdef ALLOW_WARPING
	case 'W' : warping = true; break;
#endif
        case 'f' :
	  { 
	    string parname = args[++i];
	    double value   = atof(args[++i]);
	    check_param(parname,fixableparameters[modelname]);
	    cout << "fixing parameter \"" <<  parname << "\"" << endl;
	    fixed_parameter_values[parname]=value;
	    break;
	  }
	case 'i' :
	  {
	    string parname = args[++i];
	    double value   = atof(args[++i]);
	    check_param(parname,fixableparameters[modelname]);
	    cout << "initial value for parameter \"" <<  parname << "\"" << endl;
	    initial_parameter_values[parname]=value;
	    break;
	  }
	case 'g': with_galaxy = true; break;
	case 'c': n_calibration_parameters = atoi(args[++i]); break;
	case 'e': use_model_errors = atoi(args[++i]); break;
	case 'k': use_kcorr_errors = atoi(args[++i]); break;
	case 'U': uband_calibration_error = atof(args[++i]); break;
	case 'E': error_floor_to_apply = atof(args[++i]); break;
	  
	case 'M': minos = true; break;
	default : usage();
	}
    }
#ifdef WITHEXCEPTIONS
  try
    {
   
#endif // WITHEXCEPTIONS
  // Load data
  // ------------------------------------------------------
  MeasuredData data(lc,true,band2shift,wavelenghtshift);
  for(size_t s=0;s<spectra.size();s++) {
    data.LoadSpectrum(spectra[s]);
  }

  if(error_floor_to_apply>0) {
    cout << "INFO snfit : applying error floor " << error_floor_to_apply << endl;
    for(size_t b=0;b<data.NBands();b++) {
      data.PointList(b).ApplyErrorFloor(error_floor_to_apply);
    }
  }

    // Dump effective wavelength of filters
  // ------------------------------------------------------
    for(size_t b=0;b<data.NBands();b++) {
    cout << "INFO snfit :  band " << data.PointList(b).BandName() << " mean wavelength=" << data.PointList(b).GetEffectiveFilter().Mean() << endl;
  }
  
  // Check arguments
  // ------------------------------------------------------
  if(modelname != "salt2" && data.NSpectra()>0) {
    cerr << "ERROR snfit : spectra can be fitted only with salt2" << endl;
    usage();
  }
  if(with_galaxy) {
    if(modelname!="salt2") {
      cout << "ERROR snfit : Can fit galaxy only with salt2 model" << endl;
      usage();
    }else{
      if(data.NSpectra()==0) {
	cerr << "WARNING snfit : No galaxy fitted cause no spectra" << endl;
      }
    }
  }
  
  // Load model
  // ------------------------------------------------------
  Model2Fit* model=0;
  if(modelname == "salt") {
    model=new SaltModel();
    model->SetWavelengthRange(salt1_default_wmin,salt1_default_wmax);
  }else{
    if(modelname == "salt2") {
      Salt2Model *tmodel=new Salt2Model();
            if(with_galaxy && data.NSpectra()>0) {
	tmodel->FitGalaxySpectrum(& data.Spectrum(0));
      }
      model = tmodel;
      model->SetWavelengthRange(salt2_default_wmin,salt2_default_wmax);
    
      

    }else{
      cerr << "ERROR snfit : unknown model " << modelname << endl;
      exit(-12);
    }
  }
  
  
  // Setup of the fitter
  // ------------------------------------------------------
  FullFit lcf(data,*model);
  
  if(n_calibration_parameters>0) {
    lcf.RecalibrateSpectra(true);
    for(size_t s = 0; s<data.NSpectra();s++) {
      data.Spectrum(s).SetNumberOfCalibrationParameters(n_calibration_parameters);
    }
    cout << "INFO snfit : number of calibration parameters : " << n_calibration_parameters << endl;
  }
  
  if(nobservedband>0) {
    lcf.SetMaxNumberOfLightcurvesInFit(nobservedband);
    cout << "INFO snfit : maximum number of LCs in fit = " << nobservedband << endl;
  }
  
  // Set priors
  // ------------------------------------------------------
  for(size_t p = 0; p<priors.size() ; p++)
    lcf.priors.push_back(priors[p]);
  
  
  // Fix some parameters
  // ------------------------------------------------------
  if (model->params.LocateParam("X2")) // not used so far 
    model->FixParameter("X2",0.0);
  
  if (lc.size()==0 && spectra.size()>0 && model->params.LocateParam("X0") && n_calibration_parameters>0)  {// fix X0, calib params will do the job
    if ( ! model->params.LocateParam("X0")->isfixed) {
      cout << "INFO snfit : Fixing X0=1 cause no LCs" << endl;
      model->FixParameter("X0",1.0);
    }
  }
  
  for(map<string,double>::iterator it = fixed_parameter_values.begin(); 
      it!=fixed_parameter_values.end(); ++it) 
    model->FixParameter(it->first,it->second);
  
  // Init some parameters
  // ------------------------------------------------------
  for(map<string,double>::iterator it = initial_parameter_values.begin(); 
      it!=initial_parameter_values.end(); ++it) 
    model->SetParameterValue(it->first,it->second);
  
  
  if( ! model->params.LocateParam("Redshift")->isfixed)
    model->FixParameter("Redshift",data.Zhelio()); // heliocentric redshift of SN
  

  if(data.NBands()==1 && data.NSpectra()==0) {
    if( ! model->params.LocateParam("Color")->isfixed) {
      cout << "WARNING snfit : Fixing Color to zero because single lightcurve fit" << endl;
      model->FixParameter("Color",0.0);
    }
  }
#ifdef WITH_HOST_EXTINCTION
  if(data.NBands()==2) {
    if( ! model->params.LocateParam("Tau")->isfixed) {
      cout << "WARNING snfit : Fixing Tau to zero because 2-band lightcurve fit" << endl;
      model->FixParameter("Tau",0.0);
    }
  }
  if( ! model->params.LocateParam("Rv")->isfixed) {
    cout << "WARNING snfit : Fixing Rv to 3.1" << endl;
    model->FixParameter("Rv",3.1);
  }
#endif

  // Data range used for fit
  // ------------------------------------------------------
  cout << "INFO snfit : model = " << modelname << endl;

  lcf.SetPhaseRange(phasemin,phasemax);
  cout << "INFO snfit : phase range = " << phasemin << " " << phasemax << endl;

  if(forced_wmax>0) {
    model->SetWavelengthRange(forced_wmin,forced_wmax);
  }
  {
    double wmin,wmax;
    model->GetWavelengthRange(wmin,wmax);
    cout << "INFO snfit : wavelength range = " << wmin << " " << wmax << endl;
  }


#ifdef ALLOW_WARPING 
  if(warping) {
    if(modelname=="salt2") {
      Salt2Model *tmodel = (Salt2Model*)model;
      cout << "INFO snfit : setting number of warping coefficients " << endl;
      // number of warping coefficients is number of usable lighcurves - 2
      // (2 = X0 and color)
      int nlc=0;
      for(size_t il=0; il<data.NBands(); il++) {
	if(tmodel->CanComputeFluxFor(data.PointList(il).GetEffectiveFilterWithMilkyWay())) nlc++;
	}
      cout << "INFO snfit : nlc= " << nlc << endl;
      if(true) {
	int n = nlc-2;
	if(n>2) n=2;
	if(n>0) 
	  tmodel->SetNumberOfWarpingParams(n);
      }
    }else{
	cout << "WARNING snfit : Warping only usable with salt2" << endl;
    }
  }
#endif
  

  // Very First fit without model error or phase range and X1 forced to 0
  // ------------------------------------------------------  
  
  bool firstfit = true;
  if (modelname == "salt2" && (! model->params.LocateParam("X1")->isfixed)) {
    LcParam *x1 = model->params.LocateOrAddNonConstParam("X1");
    x1->isfixed = true; 
    x1->val = 0;
    lcf.Minimize(true);
    x1->isfixed = false;      
    x1->step = +1;
    x1->minv = -5;
    x1->maxv = +5; 
    firstfit = false;
  }
  
  
  

  // Then refit without model error or phase range
  // ------------------------------------------------------  
  lcf.Minimize(firstfit);

  

  // Error model
  // ------------------------------------------------------  
  if(use_model_errors) 
    cout << "INFO snfit : Using model errors" << endl;
  else
    cout << "INFO snfit : Not using model errors" << endl;
  if(use_kcorr_errors) 
    cout << "INFO snfit : Using k-correction errors" << endl;
  else
    cout << "INFO snfit : Not using k-correction errors" << endl;
  
  if(update_weight_in_fit) 
    cout << "INFO snfit : Update model weight in fit" << endl;
  else
    cout << "INFO snfit : Do not update model weight in fit (only interation)" << endl;

  
  lcf.UseModelErrors(use_model_errors); 
  lcf.UseKcorrErrors(use_kcorr_errors);
  if(uband_calibration_error>0) {
    lcf.SetUbandSystematicError(uband_calibration_error);
    cout << "INFO snfit : Assumed U-band systematic error = " << uband_calibration_error << endl;
  }
  if(phase_weight!=0) cout << "INFO snfit : Using phase weight " << phase_weight << endl;
  lcf.SetPhaseWeight(phase_weight);
  lcf.SetUpdateWeightInFit(update_weight_in_fit);
  
  // Phase range
  // ------------------------------------------------------  
  lcf.SetPhaseRange(phasemin,phasemax);
  
  
  // --------------------------------------------------
  // Main loop
  // this is a loop cause we have to recompute the model errors
  // according to the previous best fit value
  // as many times as it is necessary
  // --------------------------------------------------
  
  bool continue_loop = true;
  int iteration = -1;
  while(continue_loop) {
    
    
    cout << "INFO snfit : ITERATION NUMBER " << ++iteration << endl;
    FitParams input_params = model->params;
    lcf.Minimize(false); // 
      
    continue_loop = false;
    
    // continue to loop if at least one parameter
    // has been modified by more than 1% of its statistical error 
    for(size_t p =0;p<input_params.size();p++) {
      if ( ! input_params[p]->isfixed ) {
	continue_loop |= ( fabs( input_params[p]->val - model->params[p]->val ) > 0.1*model->params[p]->err );
      }
    }
    
    continue_loop &= (iteration<20);
  }
  
  

  lcf.Dump();

  // minos
  // ------------------------------------------------------
  if(minos) {
    lcf.Minos();
  }
  
  // write result
  // ------------------------------------------------------
  if(resname=="") resname = "result_"+modelname+".dat";
  ofstream res(resname.c_str());
  res << lcf << endl;

  //compute UBV restframe mags and their covariances 
  string mu_magsys = "VEGA";
  {
    Instrument standard("STANDARD");
    vector<Filter> rfStuff;
    Vect rfMags;
    Mat rfCov;
    //rfStuff.push_back(standard.EffectiveFilterByBand("U02"));
    rfStuff.push_back(standard.EffectiveFilterByBand("U"));
    rfStuff.push_back(standard.EffectiveFilterByBand("B"));
    rfStuff.push_back(standard.EffectiveFilterByBand("V"));
    FitParams ubv("UBV");
    model->GetRestFrameMagsAtPhase(rfStuff, 0, ubv, mu_magsys);
    res << ubv << endl;
    
    // compute variance and covariance of MB with the other parameters.
    FitParams mb("MB_and_pars");
    model->GetRestFrameMagAndCov(standard.EffectiveFilterByBand("B"),
				 0, mb, mu_magsys);

    res << mb <<endl;
  }

#ifdef  ALLOW_WARPING
  // compute variance and covariance of observed magnitudes offsets
  // with respect to SN with null color
  {
    
    Vect rfMags;
    Mat rfCov;
    //rfStuff.push_back(standard.EffectiveFilterByBand("U02"));
    
    Salt2Model *tmodel = (Salt2Model*)model;

    vector<Filter> filters;
    for(size_t il=0; il<data.NBands(); il++) {
      Filter f = data.PointList(il).GetEffectiveFilter(); //
      if(tmodel->CanComputeFluxFor(f)) {
	filters.push_back(f);
      }
    }
    FitParams mags("delta_mags_due_to_color"); 
    model->GetMagsAtPhase(filters,0,mags, "VEGA");

    model->SetParameterValue("Color",0);
    char pname[100];
    for(int i=0;i<tmodel->NumberOfWarpingParams();i++) {
      sprintf(pname,"W%d",i);
      model->SetParameterValue(pname,0);
    }
    FitParams mags_null_color("mags_no_color"); 
    model->GetMagsAtPhase(filters,0,mags_null_color, "VEGA");
    
    // no we want to subtract magnitudes
    for(size_t i=0;i<mags.size();i++) {
      mags[i]->val = mags[i]->val - mags_null_color[i]->val;
    }
    res << mags <<endl;
  }
#endif 
  

  res.close();

  
  //lcf.Scan("W0",-0.6,0.6,100,true); // TEST
  
  delete model;

#ifdef WITHEXCEPTIONS
    }   catch (SnfitException e)
    {
      e.PrintMessage(cout);
      return EXIT_FAILURE;
    }
#endif // WITHEXCEPTIONS

  return EXIT_SUCCESS;


}

