
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "measureddata.h"
#include "saltmodel.h"
#include "salt2model.h"
#include "fileutils.h"
#include "dictfile.h"



static void usage(const char *pg)
{ 
  std::cerr << pg << " writes an ASCII file with the lightcurve of the model" << endl;
  std::cerr << "usage : " << pg << " <lightcurve1> <lightcurve2> ... -p <filename>" << std::endl;
  std::cerr << " <filename> is the path of the output file of snfit: (i.e. result_salt.dat or result_salt2.dat)" << endl; 
  exit(EXIT_FAILURE);  
}

int main(int nargs, char **args) {
  if (nargs <2) usage(args[0]);
  
  std::vector<std::string> lc;
  string parameters_filename = "";
  double step = 0.5; // days
  
  for (int i=1; i<nargs; ++i)
    {
      char *arg = args[i];
      if (arg[0] != '-') {
	if ( ! FileExists(arg)) {
	  cerr << "cannot read file " << arg << endl;
	  usage(args[0]);
	}
	lc.push_back(arg);
	continue;
      }
      switch(arg[1]) {
      case 'p': parameters_filename = args[++i]; break;// parameters
      case 's': step = atof(args[++i]); break;// parameters
      default:
	cerr << "error: unknown option " << arg << endl;
	usage(args[0]);
      } 
    }
  
  if(lc.size()==0) {
    cerr << "error: no input lightcurve" << endl;
    usage(args[0]);
  }
  if(parameters_filename=="") { 
    cerr << "error: no input model" << endl;
    usage(args[0]);
  }

  MeasuredData md(lc,false);
  // read again magsys and instrument cause not available in measureddata
  vector<string> instruments;
  vector<string> magsys;
  for(size_t i=0;i<lc.size();i++) {
    DictFile file(lc[i]);
    instruments.push_back(file.GlobalValue("INSTRUMENT"));
    magsys.push_back(file.GlobalValue("MAGSYS"));
  }
  
  // read model name
  string model_name = "";
  {
    std::ifstream stream(parameters_filename.c_str());
    if(!stream) {
      cerr << "error: cannot read paramete file " << parameters_filename << endl;
      usage(args[0]);
    }
    
    if(!(stream >> model_name)) {
      cerr << "error: cannot read model name in " << parameters_filename << endl;
      exit(EXIT_FAILURE);
    }
    stream.close();
  }
  Model2Fit *model=0;
  if(model_name=="SaltModel")
    model = new SaltModel();
  if(model_name=="Salt2Model")
    model = new Salt2Model();
  if(!model) {
    cerr << "error: cannot deal with model '" << model_name << "'" << endl;
    exit(-1);
  }

  model->ReadParams(parameters_filename);
  
  // for salt1 backward compatibility
  if(model->params.LocateParam("Redshift")->err<0) {
    cout << "warning : redshift parameter is not set in file, set it to " << md.Zhelio() << endl;
    model->SetParameterValue("Redshift",md.Zhelio());
  }
  
  const LcParam *p_daymax = model->params.LocateParam("DayMax");
  if ( ! p_daymax) {
    cerr << "error: no parameter DayMax in " << parameters_filename << endl; 
    exit(EXIT_FAILURE);
  }
  
  //===============================================
  double first_date = p_daymax->val-15.*(1.+md.Zhelio());
  double last_date = p_daymax->val+45.*(1.+md.Zhelio());

  for(size_t i=0;i<md.NBands();++i) {
    const MeasuredPointList& pointlist = md.PointList(i);
    const Filter& observerfilter = pointlist.GetEffectiveFilterWithMilkyWay();

    // for SALT1 model discard bands that are outside valid wavelength range
    if (model_name=="SaltModel") {
      if( ! model->CanComputeFluxFor(observerfilter,true)) {
	cout << "Band " << observerfilter.Band() 
	     << " at z=" << model->params.LocateParam("Redshift")->val
	     << " is outside the valid wavelength range of SALT" << endl; 
	continue;
      }
    }

    double zp_simulation = pointlist.GetSimulationZP();
    double zp_data       = (*(pointlist.begin()))->Zp();
    double flux_scale_simu2data = pow(10.,0.4*(zp_data-zp_simulation));
    
    string lcfilename = string("lc_"+model_name+"_"+observerfilter.Band()+".dat");
    cout << "Writing " <<  lcfilename << endl;
    ofstream stream(lcfilename.c_str());
    stream << "@MODEL SALT" << endl;
    stream << "@INSTRUMENT " << instruments[i] << endl;
    stream << "@MAGSYS " << magsys[i] << endl;
    stream << "@BAND " << observerfilter.Band() << endl;
    stream << "#Date :" << endl;
    stream << "#Flux :" << endl;
    stream << "#Fluxerr : 'diagonal' model error (null with SALT1)" << endl;
    stream << "#ZP :" << endl;
    stream << "#end" << endl;
    
    for (double date = first_date; date<last_date+step/2.; date += step) {
      double flux = model->Flux(&observerfilter,date)*flux_scale_simu2data;
      double relative_error = model->ModelRelativeError(&observerfilter,date,true);
      stream << date << " " 
	     << flux << " " 
	     << fabs(relative_error*flux) << " " 
	     << zp_data << endl;
    }
    stream.close();
  }
  cout << "done" << endl;
  return EXIT_SUCCESS;
}
