#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector> 

#include "measureddata.h"
#include "model2fit.h"
//#include "lctemplate.h"
//#include "spectrumtemplate.h"
#include "saltmodel.h"
#include "salt2model.h"
#include "fileutils.h"



double daymin,daymax,ymin,ymax;
bool show_phases;
string time_axis_label;
string label;

/*
  Use this code to say whether this SN is well sampled enough for cosmo usage or not



SiFTO culs (from Mark 2009-07-02) :

1) There must be some data in any of U02UBVR after +3 effective days
2) There must be some data in any of U02UBVR before +12 effective days
3) There must be B data between -8<t<+13
4) There must be one of U02UVR between -8<t<+13
5) At least two rest-frame mags (including B) measured (5 satisfied by 4)
-) 0.7<s<1.3
-) -0.25<c<0.25

SiFTO culs (from Mark 2009-09-07) :
-) data fit from 2700A -> 7200A and -25 to +30 days
-) Must be some data after +3 days
-) Must be some data before +5 days
-) At least 1 B point between -8 and +13 (here B corresponds to the
closest k-corrected observed filter)
-) At least 1 U02/U/V between -8 and +13
-) Minimum s of 0.7, maximum of 1.3
-) Minimum c of -0.25, maximum of 0.25
-) Maximum E(B-V)_mw of 0.2
-) Exclude 03D1fl, 03D1fb, 03D4cn, 03D1bf, 03D3bh
-) Exclude peculiar: 03D3bb, 05D1by (may not be necessary - normal
culls may suffice).
-) Exclude outliers: 05D2ei, 04D4gz, 03D4au, 04D3dd, sn2007bz,
sn2002jy, sn2006cj, SDSS5635

 */


static void usage(const char *pg)
{
  std::cerr << pg << " <lightcurve1> <lighcurve2> -p <filename>" << endl; 
  std::cerr << "option : -r pmin pmax" << endl;
  
  exit(EXIT_FAILURE); 
}


vector<int> days;
bool is_new(vector<int>& days, const double& day) {
  int iday = int(day);
  for(size_t i=0;i<days.size();i++)
    if(iday==days[i]) return false;
  days.push_back(iday);
  return true;
}

class interval {

public :

  double pmin; // min phase
  double pmax; // max phase
  int n_lcs;   // number of LCs that satisfy this
  int n_pts;   // number of photometric points that satisfy this

  interval(double ipmin,double ipmax ) : pmin(ipmin), pmax(ipmax), n_lcs(0), n_pts(0) {};
  bool in(const double& phase) {return phase>=pmin && phase<=pmax;}
};


int main(int nargs, char **args) {
  if (nargs <2) usage(args[0]);
  
  std::vector<std::string> lc;
  
  string parameters_filename = "";
  
  double min_phase_in_fit = -15;
  double max_phase_in_fit = 45;
  
  for (int i=1; i<nargs; ++i)
    {
      char *arg = args[i];
      if (arg[0] != '-') {
	lc.push_back(arg);
	continue;
      }
      switch(arg[1]) {
      case 'p': parameters_filename = args[++i]; break;// parameters
      case 'r': min_phase_in_fit = atof(args[++i]); max_phase_in_fit = atof(args[++i]); break;// parameters
      default:
	cerr << "error: unknown option " << arg << endl;
	usage(args[0]);
      } 
    }
  
  


  // PARAMETERS /////////////////
  interval imain(-10,+35); // to count number of epochs in fit >3
  interval irise(-10,+5); // to count number of epochs in tail for stretch and daymax >=1
  interval itail(+5,+20); // to count number of epochs in tail for stretch and daymax >=1
  interval imax(-8,+10); // to count number of LC for mags >=2
  
  // tight selection
  interval irise_tight(-10,-1); // to count number of epochs in rise
  interval imax_tight(-2,+2); // to count number of LC for mags
  
  vector<interval*> intervals;
  intervals.push_back(&imain);
  intervals.push_back(&irise);
  intervals.push_back(&itail);
  intervals.push_back(&imax);
  intervals.push_back(&irise_tight);
  intervals.push_back(&imax_tight);
  
  
  
  
  



  
  if(lc.size()==0) {
    cerr << "error: no input lightcurve" << endl;
    usage(args[0]);
  }
  if(parameters_filename=="") {
    cerr << "error: no parameter filename" << endl;
    usage(args[0]);
  }
  
  
  MeasuredData md(lc,true); // 
  
  Model2Fit *model=0;
  
  std::ifstream stream(parameters_filename.c_str());
  if(!stream) {
    cerr << "error: cannot read paramete file " << parameters_filename << endl;
    usage(args[0]);
  }
  string fitname;
  if(!(stream >> fitname)) {
    cerr << "error: cannot read fit name in " << parameters_filename << endl;
    exit(EXIT_FAILURE);
  }
  stream.close();
  
  if(fitname=="SaltModel"){
    SaltModel* tmodel = new SaltModel();
    tmodel->ReadParams(parameters_filename);
    if(tmodel->params.LocateParam("Redshift")->err<0) {
      cout << "redshift parameter was not defined, set it to " << md.Zhelio() << endl;
      tmodel->SetParameterValue("Redshift",md.Zhelio());
    }
    model = tmodel;
  }else{
    if(fitname=="Salt2Model") {
      Salt2Model * tmodel = new Salt2Model();
      model = tmodel;
      model->ReadParams(parameters_filename);
    }else{
      cerr << "error, unknown fitting model " << fitname << endl; 
      exit(12);
    }
  }
 
  bool debug = true;
  double first_phase = 1000;
 

  for(size_t i=0;i<md.NBands();++i) {
    const MeasuredPointList& pointlist = md.PointList(i);
    const Filter& f = pointlist.GetEffectiveFilterWithMilkyWay();
    if(! model->CanComputeFluxFor(f)) {
      if(debug) cout << "DEBUG snstat : ignore band " << f.Name() << endl;
      continue;
    }
    
    Vect checked_intervals(intervals.size());
    
    // record phases and compute max_flux
    
    
    for(MeasuredPointCIterator ip = pointlist.begin(); ip!=pointlist.end(); ++ip) {
      double phase = model->GetPhase( (*ip)->Date() );
      
      if(phase<min_phase_in_fit || phase>max_phase_in_fit) continue;
      
      bool new_date = is_new(days,(*ip)->Date());
      
      if(phase<first_phase) first_phase = phase;
      
      // fill intervals
      for(size_t i=0;i<intervals.size();i++) {
	if(intervals[i]->in(phase)) { 
	  
	  // count only phases at different days 
	  if(new_date)
	    intervals[i]->n_pts ++;
	  
	  if(checked_intervals(i)==0) {
	    intervals[i]->n_lcs ++;
	    checked_intervals(i)=1;
	  }
	}
      }
      
    }
  }
    
  
  cout << "SNSTAT:FPHASE " << first_phase << endl;
  cout << "SNSTAT:NLCMAX " << imax.n_lcs << endl;
  cout << "SNSTAT:NMAIN " <<  imain.n_pts << endl;
  cout << "SNSTAT:NRISE "  << irise.n_pts << endl;
  cout << "SNSTAT:NTAIL "  << itail.n_pts << endl;
  cout << "SNSTAT:NLCMAXTIGHT " << imax_tight.n_lcs << endl;
  cout << "SNSTAT:NRISETIGHT " <<  irise_tight.n_lcs << endl;
  
  bool ok = imax.n_lcs>=2 
    && imain.n_pts>=4
    && irise.n_pts>=1
    && itail.n_pts>=1;
 
  cout << "SNSTAT:SCORE " << ok << endl;
  
  ok = imax_tight.n_lcs>=2 
    && imain.n_pts>=4
    && irise_tight.n_pts>=1
    && itail.n_pts>=1;
  
  cout << "SNSTAT:TIGHTSCORE " << ok << endl;
  return 0;
}
