//#include "cosmology_prior.h"

#include "wmap5cosmo.h"
#include "wmap5cosmoprior.h"


#include "matvect.h"
#include "fatalerror.h"

#include <cmath>

using namespace std;

/* Five-Year Wilkinson Microwave Anisotropy Probe (WMAP) Observations: 
   Cosmological Interpretation
Komatzu et al, section 5.4

Accepted paper but no journal ref yet (dec 2008)

http://cdsads.u-strasbg.fr/abs/2008arXiv0803.0547K
http://lambda.gsfc.nasa.gov/product/map/dr3/pub_papers/fiveyear/cosmology/wmap_5yr_cosmo.pdf

P.A.

*/




static void offset_params(Wmap5Cosmo &C, 
			  const int Indices[],
			  const Vect &Offsets)
{
  for (unsigned int k=0; k<Offsets.size(); ++k)
    {
      C.SetParam(Indices[k],C.GetParam(Indices[k])+Offsets(k));
    }
}

Vect Wmap5CosmoPrior::Residual(const Cosmology *Cosmo) const { // Cosmo must be a OmegamW0Cosmo
  // WARNING:  not optimal call Wmap5CosmoPrior::Residual
  Vect R;
  Chi2(Cosmo,&R);
  return R;
}


// does marginalisation over obh2 and h0
// provides Weight and Residuals at end of marginalisation
double Wmap5CosmoPrior::Chi2(const Cosmology * Cosmo, Vect* output_residuals) const // Cosmo must be a OmegamW0Cosmo
{
  

  double om,ol,w;
  double obh2=0.02268, h0=72.4;
  const OmegamW0Cosmo *cosmo = dynamic_cast<const OmegamW0Cosmo*>(Cosmo);
  if (cosmo != 0)
    {
      om = cosmo->Omegam();
      ol = 1-om;
      w = cosmo->W0();
    }
  // else try other casts
  else {
    //FatalError(" Wmap5CosmoPrior::Chi2 : unknown cosmo type " );
    cerr << "FATAL  Wmap5CosmoPrior::Chi2 : unknown cosmo type " << endl;
    abort();
  }
  // we are going to marginalize over obh2 and h0
  /* it does not work in regions where the fitting functions for z*
     just don't work, or where the model is just crazy (e.g. OmegaB>>Omegam)
     In sems accpetable in regions where the "observational" constraints 
     are satisfied.
  */
  Wmap5Cosmo w5(om,ol,w,obh2,h0);
  int indices[2];
  indices[0] = w5.GetParamRank("OmegabH2");
  indices[1] = w5.GetParamRank("H0");
  double eps[2] = {0.0001, 0.1};

  double chi2;
  double oldchi2 = Chi2ForWmap5Cosmo(&w5);
  Mat weight(Weight());

  Vect res=ResidualForWmap5Cosmo(&w5);

  for (int iter=0; iter<10; ++iter)
    {
      // derivatives of prior residuals w.r.t obh2 and h0
      Mat H(2,3);
      for (int k=0; k<2; ++k)
	{
	  Wmap5Cosmo tmp(w5);
	  int index = indices[k];
	  // numerical derivatives
	  tmp.SetParam(index,tmp.GetParam(index)+eps[k]);
	  Vect tmpRes = ResidualForWmap5Cosmo(&tmp);
	  for (int j=0; j<3; ++j)
	    H(k,j) = (res(j)-tmpRes(j))/eps[k];
	}
      // second derivatives of chi2 w.r.t the parameters 
      // we are marginalizing over :
      Mat A(2,2);
      A = H.transposed()*weight*H;
      // gradient of the chi2 w.r.t to the same ones
      Vect B(2);
      B = H.transposed()*weight*res;
      Vect grad(B);
      if (cholesky_solve(A,B,"U") != 0) // try the gradient
	  B = grad;

      // pas de brent pour l'instant
      offset_params(w5,indices,B);
      res =   ResidualForWmap5Cosmo(&w5);
      chi2 = scalar_product(res,weight*res);
      if (fabs(chi2-oldchi2)<1e-3) break;
      if (isnan(chi2) || chi2>oldchi2)
	{
	  // poor man's "brent":
	  // undo last move
	  offset_params(w5,indices,-1.*B);
	  bool found = false;
	  for (double s=1; s>-2; s-=2)
	    {
	      Vect tmpB(s*B);
	      for (int trial=0; trial<10; ++trial)
		{
		  Wmap5Cosmo tc(w5);
		  offset_params(tc,indices,0.5*tmpB);
		  double tmpchi2 = Chi2ForWmap5Cosmo(&tc);
		  if (tmpchi2<oldchi2)
		    {
		      chi2 = tmpchi2;
		      w5 = tc;
		      found = true;
		      break;
		    }
		  tmpB *= 0.5;
		}
	      if (found) break;
	    }
	}
      if (isnan(chi2) || chi2>oldchi2) return 1e6;	  
      oldchi2 = chi2;
      // cout << " oldchi2, chi2 " << oldchi2 << ' ' << chi2 << ' ' 
      // << oldchi2-chi2 << endl;
    }
  if(output_residuals) *output_residuals = res;

  return (isnan(chi2) ? 1e6 : chi2);
}


Vect Wmap5CosmoPrior::ResidualForWmap5Cosmo(const Wmap5Cosmo *cosmo) const
{// data from table 10 of Komatsu et al
  
  /*
  const Wmap5Cosmo *cosmo = dynamic_cast<const Wmap5Cosmo*>(Cosmo);
  
  if(!cosmo)
    FatalError("Wmap5CosmoPrior::Residual(const Cosmology *Cosmo) Cosmo must be a Wmap5Cosmo");
  */
  
  

  Vect r(3);
  double zstar = cosmo->zstar();
  r(0) = cosmo->l_a(zstar)- 302.10;
  r(1) = cosmo->R(zstar) - 1.710;
  r(2) = zstar - 1090.04;
  return r;
}

Mat Wmap5CosmoPrior::Weight(const Cosmology *cosmo) const
{// table 11 of Komatsu et al
  Mat W(3,3);
  W(0,0) = 1.800;
  W(1,0) = W(0,1) = 27.968;
  W(2,0) = W(0,2) = -1.103;
  W(1,1) = 5667.577;
  W(2,1) = W(1,2) = -92.263;
  W(2,2) = 2.923;
  return W;
}

double Wmap5CosmoPrior::Chi2ForWmap5Cosmo(const Wmap5Cosmo* Cosmo) const
{
  Vect r(ResidualForWmap5Cosmo(Cosmo));
  return scalar_product(r,Weight()*r);
}
  


